/*
	$Id: Annotation.java 7456 2018-03-06 16:15:04Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.Transactional.Action;
import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.core.data.AnnotationTypeData;
import net.sf.basedb.core.data.UnitData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.util.EqualsHelper;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.PrefixSuffixFormatter;
import net.sf.basedb.util.units.LinearUnitConverter;
import net.sf.basedb.util.units.UnitConverter;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Collections;
import java.util.Set;

/**
	This class represents an annotation. Annotations are user-defined
	properties that can be attached to {@link Annotatable}
	items. They can be very useful in the filtering and analysis
	process.
	<p>
	An annotation must have an {@link AnnotationType},
	which controls the type of and how many values that can be attached 
	to an annotation.
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2018-03-06 17:15:04 +0100 (ti, 06 mar 2018) $
*/
public class Annotation
	extends BasicItem
{
	/**
		The type of item represented by this class.
		@see Item#ANNOTATION
		@see #getType()
	*/
	public static final Item TYPE = Item.ANNOTATION;
	
	/**
		Create a new primary annotation.
		
		@param dc The DbControl object
		@param annotationType The annotation type of the new annotation
		@param annotationSet The annotation set of the new annotation
		@see AnnotationSet#getAnnotation(AnnotationType)
		@since 3.6
	*/
	static Annotation getNewPrimary(DbControl dc, AnnotationSet annotationSet, AnnotationTypeData annotationType)
		throws BaseException
	{
		assert annotationSet != null : "annotationSet == null";
		assert annotationType != null : "annotationType == null";
		Annotation a = dc.newItem(Annotation.class, annotationSet);
		AnnotationData data = a.getData();
		data.setAnnotationSet(annotationSet.getData());
		data.setAnnotationType(annotationType);
		data.setSource(Source.PRIMARY.ordinal());
		data.setProjectId(annotationType.getProjectAnnotations() ? dc.getSessionControl().getActiveProjectId() : 0);
		data.setUnit(annotationType.getDefaultUnit());
		data.setValues(Type.fromValue(annotationType.getValueType()).newParameterValueData());
		annotationSet.getData().getAnnotations().add(data);
		return a;
	}
	
	/**
		Create a new inherited or cloned annotation.
		@param dc The DbControl object
		@param inheritedFrom The annotation to inherit
		@param annotationSet The annotation set of the new annotation
		@param clone If the annotation values should be cloned or not
		@see AnnotationSet#getAnnotation(AnnotationType)
		@since 3.6
	*/
	static Annotation getNewInherited(DbControl dc, AnnotationSet annotationSet, AnnotationData inheritedFrom, boolean clone)
		throws BaseException
	{
		assert annotationSet != null : "annotationSet == null";
		assert inheritedFrom != null : "inheritedFrom == null";
		Annotation a = dc.newItem(Annotation.class, annotationSet);
		AnnotationData data = a.getData();
		data.setAnnotationSet(annotationSet.getData());
		data.setAnnotationType(inheritedFrom.getAnnotationType());
		data.setInheritedFrom(inheritedFrom);
		data.setProjectId(inheritedFrom.getProjectId());
		if (clone)
		{
			data.setSource(Source.CLONED.ordinal());
			data.setValues(Type.fromValue(inheritedFrom.getAnnotationType().getValueType()).newParameterValueData());
			data.getValues().replaceValuesIfDifferent(inheritedFrom.getValues().getValues(), false);
			data.setUnit(inheritedFrom.getUnit());
			data.setLastUpdate(inheritedFrom.getLastUpdate());
		}
		else
		{
			data.setSource(Source.INHERITED.ordinal());
		}
		annotationSet.getData().getAnnotations().add(a.getData());
		return a;
	}
	
	/**
		Get an <code>Annotation</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Annotation</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static Annotation getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Annotation a = dc.loadItem(Annotation.class, id);
		if (a == null) throw new ItemNotFoundException("Annotation[id="+id+"]");
		return a;
	}
	
	/**
		Get a query configured to retrieve annotations. This query may return items 
		which the logged in user doesn't have 
		read permission to. At the moment there is no way to solve this problem.
		@return An {@link ItemQuery} object
	*/
	static ItemQuery<Annotation> getQuery()
	{
		return new ItemQuery<Annotation>(Annotation.class, new QueryRuntimeFilterImpl());
	}
	
	private AnnotationSet annotationSet;
	private boolean disableATPermissionCheck;
	
	Annotation(AnnotationData annotationData)
	{
		super(annotationData);
	}
	
	Annotation(AnnotationData annotationData, AnnotationSet annotationSet)
	{
		super(annotationData);	
		this.annotationSet = annotationSet;
	}
	
	@Override
	AnnotationData getData()
	{
		return (AnnotationData)getBasicData();
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		READ permission is granted if the logged in user has READ permission on 
		the associated item and annotation type. CREATE, WRITE and DELETE permissions
		are granted if the logged in user has WRITE permission on the associated item
		and USE permission on the annotation type. USE permission (needed for inheriting)
		is granted if the logged in user has USE permission on the associated item.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		AnnotationSet as = getAnnotationSet();
		AnnotationType at = null;
		if (!disableATPermissionCheck)
		{
			at = getAnnotationType();
		}
		if (as.hasPermission(Permission.WRITE) && (at == null || at.hasPermission(Permission.USE)))
		{
			granted |= Permission.grant(Permission.CREATE, Permission.READ, Permission.WRITE, Permission.DELETE);
		}
		else if (as.hasPermission(Permission.USE))
		{
			granted |= Permission.grant(Permission.USE);
		}
		else
		{
			granted |= Permission.grant(Permission.READ);
		}
		super.initPermissions(granted, denied);
	}
	
	@Override
	void onAfterInsert() 
		throws BaseException
	{
		super.onAfterInsert();
		AnnotationData data = getData();
		if (data.getAnnotationType().getProjectAnnotations())
		{
			boolean isPrimary = getSource() == Source.PRIMARY;
			ProjectSpecificAnnotationsManager pspManager = getDbControl().getProjectSpecificAnnotationsManager();
			
			if (isPrimary)
			{
				pspManager.primaryCreated(data.getId(), data.getAnnotationSet().getId(), data.getAnnotationType().getId(), data.getProjectId());
			}
			else
			{
				AnnotationData inheritedFrom = data.getInheritedFrom();
				pspManager.inheritedCreated(data.getId(), data.getAnnotationSet().getId(), 
					inheritedFrom.getId(), inheritedFrom.getAnnotationSet().getId(), inheritedFrom.getOverrideId(), data.getProjectId());
			}
		}
	}

	/**
		Nullify links to this annotation from cloned annotations when this annotation is deleted
		(since the cloned annotation should not be deleted).
	*/
	@Override
	void onBeforeCommit(Action action) 
		throws BaseException 
	{
		super.onBeforeCommit(action);
		boolean isPrimary = getSource() == Annotation.Source.PRIMARY;
		if (action == Action.CREATE)
		{
			if (isPrimary && getData().getValues().getValues().isEmpty())
			{
				String msg = "An annotation was created with not values: '";
				msg += getData().getAnnotationType().getName() + "'";
				if (annotationSet != null)
				{
					msg += " on item " + annotationSet.getItemType().name();
					msg += "[ID="+annotationSet.getItemId() + "]";
				}
				throw new InvalidDataException(msg);
			}
		}
		else if (action == Action.DELETE)
		{
			if (getData().getAnnotationType().getProjectAnnotations() && !isProjectSpecific())
			{
				getDbControl().getProjectSpecificAnnotationsManager().defaultDeleted(getId(), isPrimary);
			}
			if (isPrimary)
			{
				// Cloned annotations should remain
				org.hibernate.Session session = getDbControl().getHibernateSession();
				org.hibernate.query.Query<?> query = HibernateUtil.getPredefinedQuery(session, 
					"NULLIFY_LINK_FROM_CLONES");
				/*
					UPDATE AnnotationData ad
					SET ad.inheritedFrom = null
					WHERE ad.inheritedFrom = :targetId
					AND ad.source = CLONED
				*/
				query.setParameter("targetId", getId(), TypeWrapper.H_INTEGER);
				HibernateUtil.executeUpdate(query);
			}
		}
	}

	@Override
	PluginPermission getPluginPermissions()
	{
		return PluginPermission.NO_PLUGIN;
	}
	// -------------------------------------------

	void disableATPermissionCheck()
	{
		this.disableATPermissionCheck = true;
	}
	
	private boolean ignoreActiveProject;
	/**
		Set the project that this annotation value should 
		belong to. This will disable the automatic check
		that creates a new project-specific value when 
		trying to change the default value and a project
		is active.
		@since 3.11
	*/
	void setProjectId(int projectId)
	{
		getData().setProjectId(projectId);
		ignoreActiveProject = true;
	}
	
	/**
		Get the annotation set this annotation belongs to.
		@return An <code>AnnotationSet</code>
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the annotation set (should never happen)
		@throws BaseException If anything else goes wrong when getting the annotation set.
	*/
	public AnnotationSet getAnnotationSet()
		throws PermissionDeniedException, BaseException
	{
		if (annotationSet == null)
		{
			return getDbControl().getItem(AnnotationSet.class, getData().getAnnotationSet());
		}
		else
		{
			return annotationSet;
		}		
	}

	/**
		Get the date and time the values in this annotation was last
		updated.
		@since 2.14
	*/
	public Date getLastUpdate()
	{
		return DateUtil.copy(getData().getLastUpdate());
	}
	
	/**
		Check if this annotation belongs to the primaray annotations for a specified annotation set.
		This method is useful to check if an annotation has been inherited or not.
		@param as The annotation set to check
		@return TRUE if this annotation is a primary annotation of the specified annotation set,
			FALSE otherwise
		@since 2.10
	*/
	public boolean belongsTo(AnnotationSet as)
	{
		return getData().getAnnotationSet().equals(as.getData());
	}
	
	/**
		Get the annotation type of this annotation.
		@return An <code>AnnotationType</code>
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the annotation type
	 * @throws BaseException If anything else goes wrong when getting the annotation type.
	*/
	public AnnotationType getAnnotationType()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(AnnotationType.class, getData().getAnnotationType());
	}
	
	/**
		Get the annotation source.
		@since 3.6
	*/
	public Source getSource()
	{
		return Source.values()[getData().getSource()];
	}
	
	/**
		Get the parent annotation if this is an inherited or cloned 
		annotation. May return null for a cloned annotation if the
		parent annotation has been deleted.
		@since 3.6
	*/
	public Annotation getInheritedFrom()
	{
		return getDbControl().getItem(Annotation.class, getData().getInheritedFrom());
	}
	
	/**
		Is this annotation a project-specific annotation?
		@since 3.10
	*/
	public boolean isProjectSpecific()
	{
		return getData().getProjectId() != 0;
	}
	
	/**
		Checks if this annotation is ignoring the currently active
		project. This affects how the {@link #setValuesIfDifferent(List, Unit)}
		methods treats annotations that represents the default value
		for a project-specific annotation.

		If this method returns FALSE the default value is not modified.
		Instead, a new project-specific value for the currently active project
		is created (or the existing one is updated)

		If this method returns TRUE it is always this annotation that is
		modified.
		
		@return TRUE if ignoring the active project
		@since 3.11
	*/
	public boolean isIgnoringActiveProject()
	{
		return ignoreActiveProject;
	}
	
	/**
		Get the ID of the project that this project-specicif annotation belongs to.
		@return A project ID or 0 if this annotation is not project-specific
		@since 3.10
	*/
	public int getProjectId()
	{
		return getData().getProjectId();
	}
	
	/**
		Get the project that this project-specific annotation belongs to.
		@return A project or null if this annotation is not project-specific
		@since 3.10
	*/
	public Project getProject()
	{
		int projectId = getProjectId();
		return projectId == 0 ? null : Project.getById(getDbControl(), projectId);
	}
	
	/**
		Is this (project-specific) annotation overriding a default value?
		@since 3.10
	*/
	public boolean isOverridingDefault()
	{
		return getData().getOverrideId() != 0;
	}
	
	/**
		Get the default annotation that this project-specific annotation is overriding.
		@return The default annotation or null if this annotation is not overriding a default value
		@since 3.10
	*/
	public Annotation getDefaultAnnotation()
	{
		int overrideId = getData().getOverrideId();
		return overrideId == 0 ? null : getById(getDbControl(), overrideId);
	}
	
	/**
		Checks if the values in this annotation has a unit.
		@return TRUE if a unit is used, FALSE otherwise
		@since 2.9
	*/
	public boolean hasUnit()
	{
		return getData().getUnit() != null;
	}
	
	/**
		Get the unit used for the values in this annotation.
		@return A unit, or null if this annotation doesn't use units
		@since 2.9
	*/
	public Unit getUnit()
	{
		return getDbControl().getItem(Unit.class, getData().getUnit());
	}
	
	/**
		Get a converter for converting the values in this annotation to
		a specific unit. 
		
		@param unit The unit to convert the values to, or null
			to convert the values to the unit specified by {@link #getUnit()}
		@return A unit converter, or null if this annotation doesn't use units, or
			if the specified unit is the same as the default unit of the annotation
			type (see {@link AnnotationType#getDefaultUnit()}
		@since 2.9
	*/
	public UnitConverter getUnitConverter(Unit unit)
	{
		UnitConverter converter = null;
		UnitData unitData = unit == null ? getData().getUnit() : unit.getData();
		if (unitData != null)
		{
			UnitData reference = getData().getAnnotationType().getDefaultUnit();
			if (!unitData.equals(reference))
			{
				converter = new LinearUnitConverter(
						unitData.getReferenceFactor(), unitData.getReferenceOffset(),
						reference.getReferenceFactor(), reference.getReferenceOffset());
			}
		}
		return converter;
	}

	/**
		Wrap a parent formatter with a formatter for units. NOTE!
		Values should be retrived with the same unit using 
		{@link #getValues(Unit)} to be correctly formatted with the
		returned formatter.
		
		@param parent The parent formatter, which should be of the
			correct type for the annotation values
		@param unit The unit to use for formatting, or null to
			use the annotation's unit
		@return A formatter object
		@since 2.9
	*/
	public <T> Formatter<T> getUnitFormatter(Formatter<T> parent, Unit unit)
	{
		Formatter<T> formatter = parent;
		UnitData unitData = unit == null ? getData().getUnit() : unit.getData();
		if (unitData == null) 
		{
			unitData = getData().getAnnotationType().getDefaultUnit();
		}
		if (unitData != null)
		{
			formatter = new PrefixSuffixFormatter<T>("", parent, 
					PrefixSuffixFormatter.NBSP + unitData.getDisplaySymbol());
		}
		return formatter;
	}
	
	/**
		Shortcut for <code>Unit.getDisplaySymbol()</code>, but also
		checks the default unit of the annotation type.
		@param unit The unit to get the symbol for, or null to use
			the annotation's unit
		@return The unit symbol, or null if this annotation doesn't use
			units
		@since 2.9
	*/
	public String getUnitSymbol(Unit unit)
	{
		UnitData unitData = unit == null ? getData().getUnit() : unit.getData();
		if (unitData == null) 
		{
			unitData = getData().getAnnotationType().getDefaultUnit();
		}
		return unitData == null ? null : unitData.getDisplaySymbol();
	}
	
	/**
		Get the value type for this annotation type. It can't be change once
		the object has been created.
	  	@return Type value type as an enum. 
		@since 2.2
	*/
	public Type getValueType()
	{
		return Type.fromValue(getData().getAnnotationType().getValueType());
	}
	
	/**
		Get the values this annotation contains. The values are of a
		{@link Type} appropriate for the given annotation type, ie. Integer object
		for an integer annotation type. All objects in the list are of the same type.
		If this is an annotation with units, the values are returned in the unit
		of the annotation type ({@link AnnotationType#getDefaultUnit()}). Use
		{@link #getValues(Unit)} to get the values in a different unit, including the
		specific unit used on this annotation.
		@see AnnotationType#getValueType()
		@return A list of objects
		@throws BaseException If there is an error
	*/
	public List<?> getValues()
		throws BaseException
	{
		if (getSource() == Source.INHERITED) return Collections.emptyList();
		List<?> values = Values.getItemValues(getDbControl(), getData().getValues().getValues());
		return Collections.unmodifiableList(values);
	}
	
	/**
		Get the values for this annotation in a specific unit. The values are of a
		{@link Type} appropriate for the given annotation type, ie. Integer object
		for an integer annotation type. All objects in the list are of the same type.
		
		@param unit The unit to convert the values to, or null to get the values
			in the unit used on this annotation
		@see AnnotationType#getValueType()
		@return A list of objects
		@throws BaseException If there is an error
		@since 2.9
	*/
	public List<?> getValues(Unit unit)
		throws BaseException
	{
		if (getSource() == Source.INHERITED) return Collections.emptyList();
		List<?> values = Values.getItemValues(getDbControl(), getData().getValues().getValues());
		UnitConverter converter = getUnitConverter(unit);
		if (converter != null)
		{
			List<Object> convertedValues = new ArrayList<Object>(values.size());
			Type valueType = getValueType();
			for (Object value : values)
			{
				if (value instanceof Number)
				{
					value = valueType.convertNumber(converter.convertToSpecificUnit(((Number)value).doubleValue()));
				}
				convertedValues.add(value);
			}
			values = convertedValues;
		}
		return Collections.unmodifiableList(values);
	}
	
	
	/**
		Get the version number of the item keeping the values. This is useful when you need
		to know if the annotation values has been modified by another transaction since
		they were read from the database. Unlike other items the version number of the
		annotation item doesn't change when the values are modified.
	  	@return The version number as an int. 
	*/
	public int getValuesVersion()
	{
		return getData().getValues().getVersion();
	}

	@Deprecated
	public void setValue(Object value)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		setValueIfDifferent(value, null);
	}
	
	/**
		Set the value of the annotation, replacing any previous values.
		@param value The new value
		@param unit The unit of the value, or null to use the default unit of
			the annotation type
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission for the annotation or read permission for the annotation
			type
		@throws InvalidDataException If the value isn't a valid value for the 
			annotation type, see {@link AnnotationType#validateAnnotationValue(Object)}
		@throws BaseException If there is another error
		@since 2.9
		@deprecated In 3.8, use {@link #setValueIfDifferent(Object, Unit)} instead
	*/
	@Deprecated
	public void setValue(Object value, Unit unit)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		setValueIfDifferent(value, unit);
	}

	/**
		Set the value of the annotation, replacing any previous values if they
		are different. No change is made if the new value is the same
		as the existing value.
		<p>
		If this annotation type uses project-specific annotations this method
		behaves differently. If a project is active and this annotation is the
		default annotation, this method will not update the values. Instead
		it will load the existing project-specific annotation or create a new
		one and update that instead (but only if the values are different). Thus,
		this method may return TRUE even if the valuse in this annotation hasn't 
		changed. The project-specific annotation can be retrieved by calling
		{@link AnnotationSet#getAnnotation(AnnotationType)}.
		
		@param value The new value
		@param unit The unit of the value, or null to use the default unit of
			the annotation type
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission for the annotation or read permission for the annotation
			type
		@throws InvalidDataException If the value isn't a valid value for the 
			annotation type, see {@link AnnotationType#validateAnnotationValue(Object)}
		@throws BaseException If there is another error
		@since 3.8
	*/
	public boolean setValueIfDifferent(Object value, Unit unit)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		return setValuesIfDifferent(Arrays.asList(value), unit);
	}

	
	@Deprecated
	public void setValues(List<?> values)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		setValuesIfDifferent(values, null);
	}
	
	/**
		Set the values of the annotation, replacing any previous values. NOTE!
		As of BASE 3.8 the values are only replaced if the new values are
		different from the existing values. 
		
		@deprecated In 3.8, use {@link #setValuesIfDifferent(List, Unit)} instead
	*/
	@Deprecated
	public void setValues(List<?> values, Unit unit)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		setValuesIfDifferent(values, unit);
	}

	/**
		Set the values of the annotation, replacing any previous values if they
		are different. No change is made if the new list of values is the same
		as the existing values.
		<p>
		If this annotation type uses project-specific annotations this method
		behaves differently. If a project is active and this annotation is the
		default annotation, this method will not update the values. Instead
		it will load the existing project-specific annotation or create a new
		one and update that instead (but only if the values are different). Thus,
		this method may return TRUE even if the valuse in this annotation hasn't 
		changed. The project-specific annotation can be retrieved by calling
		{@link AnnotationSet#getAnnotation(AnnotationType)}.
		<p>
		An exception to the above (since BASE 3.11) is if the annotation was loaded with
		{@link AnnotationSet#getProjectAnnotation(AnnotationType, Project)} in 
		which case currently active project is ignored and the values are
		always changed on this annotation (eg. the same as if this annotation
		was not using project-specific annotations).
		Use the {@link #isIgnoringActiveProject()} to find out if the active 
		project is ignored or not.
		
		@param values A list containing the new values
		@param unit The unit of the value, or null to use the default unit of
			the annotation type
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission for the annotation or read permission for the annotation
			type
		@throws InvalidDataException If the number of values are too many according
			to the {@link AnnotationType#getMultiplicity()} setting any of the values 
			isn't a valid value for the annotation type, see 
			{@link AnnotationType#validateAnnotationValue(Object)}
		@throws BaseException If there is another error
		@since 3.8
	*/
	public boolean setValuesIfDifferent(List<?> values, Unit unit)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (values == null || values.size() == 0)
		{
			throw new InvalidDataException("The 'values' parameter is null or empty");
		}
		AnnotationType annotationType = getAnnotationType();
		int multiplicity = annotationType.getMultiplicity();
		if (multiplicity > 0 && values.size() > multiplicity)
		{
			throw new InvalidDataException("Too many values. Max "+multiplicity+" value(s) are allowed for this annotation type: "+annotationType.getName());
		}
		
		int projectId = getSessionControl().getActiveProjectId();
		boolean useProjectAnnotations = !ignoreActiveProject && isInDatabase() && 
			annotationType.getProjectAnnotations() && projectId != 0 && getData().getProjectId() == 0;
		if (useProjectAnnotations)
		{
			// We have a default annotation, but are setting project-specific values
			// We need to create a new project-specific annotation unless one already 
			// exists in which case we should update that instead
			Annotation projectSpecific = getDbControl().getProjectSpecificAnnotationsManager().findProjectSpecificAnnotation(this, projectId);
			if (projectSpecific != null) 
			{
				return projectSpecific.setValuesIfDifferent(values, unit);
			}
			// There is no existing project-specific annotation, we will create one below
			// if the values are different from the the default values
		}
		
		List<Object> convertedValues = convertValues(annotationType, values, unit);
		
		AnnotationData data = getData();
		boolean different = data.getValues().replaceValuesIfDifferent(Values.getDataValues(convertedValues), useProjectAnnotations);
		if (different)
		{
			if (useProjectAnnotations)
			{
				// Now, we need to create a new annotation instead of changing this default annotation
				DbControl dc = getDbControl();
				Annotation psAnnotation = getNewPrimary(dc, getAnnotationSet(), annotationType.getData());
				data = psAnnotation.getData();
				data.getValues().replaceValuesIfDifferent(convertedValues, false);
				dc.saveItem(psAnnotation);
			}

			if (annotationType.supportUnits())
			{
				data.setUnit(unit == null ? annotationType.getData().getDefaultUnit() : unit.getData());
			}
			data.setLastUpdate(new Date());
			getAnnotationSet().setSnapshotInvalid();
		};
		return different;
	}

	/**
		Check if the given value is different from the values already stored in this 
		annotation.
		@since 3.11
	*/
	public boolean checkIfDifferent(Object value, Unit unit)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		return checkIfDifferent(Arrays.asList(value), unit);
	}

	
	/**
		Check if the given values are different from the values already stored in this 
		annotation.
		@since 3.11
	*/
	public boolean checkIfDifferent(List<?> values, Unit unit)
	{
		if (values == null || values.size() == 0)
		{
			throw new InvalidDataException("The 'values' parameter is null or empty");
		}

		List<Object> convertedValues = convertValues(getAnnotationType(), values, unit);
		
		AnnotationData data = getData();
		boolean different = data.getValues().replaceValuesIfDifferent(Values.getDataValues(convertedValues), true);
		return different;
	}
	
	private List<Object> convertValues(AnnotationType annotationType, List<?> values, Unit unit)
	{
		UnitConverter converter = null;
		if (annotationType.supportUnits() && unit != null)
		{
			converter = getUnitConverter(unit);
		}
		List<Object> convertedValues = new ArrayList<Object>(values.size());
		Type valueType = getValueType();
		for (Object value : values)
		{
			if (value instanceof Number)
			{
				Number numericValue = (Number)value;
				if (converter != null) 
				{
					numericValue = converter.convertToReferenceUnit(numericValue.doubleValue());
				}
				value = valueType.convertNumber(numericValue);
			}
			annotationType.validateAnnotationValue(value);
			convertedValues.add(value);
		}
		return convertedValues;
	}
	
	/**
		If this is a cloned annotation, resync the cloned values. This method
		does nothing on PRIMARY or INHERITED annotations.
		@since 3.6
	*/
	public void resyncCloned()
	{
		if (getSource() != Source.CLONED) return;
		
		AnnotationData data = getData();
		AnnotationData inherited = data.getInheritedFrom();
		if (inherited == null) return;
		
		if (EqualsHelper.equals(data.getLastUpdate(), inherited.getLastUpdate())) 
		{
			return;
		}
		
		if (data.getValues().replaceValuesIfDifferent(inherited.getValues().getValues(), false))
		{
			data.setLastUpdate(inherited.getLastUpdate());
			data.setUnit(inherited.getUnit());
			getAnnotationSet().setSnapshotInvalid();
		}
	}
	
	/**
		The annotation source.
		@since 3.6
	*/
	public enum Source
	{
		/**
			The annotation is a primary annotation.
		*/
		PRIMARY,
		
		/**
			The annotation is an inherited annotation 
			without a local copy of the values.
		*/
		INHERITED,
		
		/**
			The annotation is an inherited annotation
			with a local copy of the values at the
			time the copy was made.
		*/
		CLONED
	}

	/*
		Filter implementation that make sure that only annotations with annotation types
		that the logged in user has permission to read are loaded.
	 */
	static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		
		private Collection<Integer> annotationTypeIds;
		
		QueryRuntimeFilterImpl() 
		{}
		
		@Override
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			SessionControl sc = dc.getSessionControl();
			if (sc.hasPermission(Permission.READ, Item.ANNOTATIONTYPE))
			{
				// Do not add any filters.
			}
			else
			{
				// The user doesn't have generic permission to read annotation types
				// Load all that annotation types that the user owns or has been shared
				if (annotationTypeIds == null)
				{
					org.hibernate.query.Query<Integer> atQuery = HibernateUtil.getPredefinedSQLQuery(dc.getHibernateSession(), 
						"GET_ALL_ANNOTATIONTYPE_IDS_FOR_USER", Integer.class);
					
					// Owner
					atQuery.setParameter("owner", sc.getLoggedInUserId(), TypeWrapper.H_INTEGER);
					
					// All items keys for the logged in user
					Set<Integer> itemKeys = sc.getItemKeys();
					if (itemKeys == null || itemKeys.size() == 0) itemKeys = QueryRuntimeFilterFactory.ZERO_SET;
					atQuery.setParameterList("itemKeys", itemKeys);
					
					// All project keys for the active project and logged in user
					Set<Integer> projectKeys = sc.getProjectKeys();
					if (projectKeys == null || projectKeys.size() == 0) projectKeys = QueryRuntimeFilterFactory.ZERO_SET;
					atQuery.setParameterList("projectKeys", projectKeys);
					
					annotationTypeIds = HibernateUtil.loadList(atQuery, null);
					if (annotationTypeIds.size() == 0)
					{
						annotationTypeIds = QueryRuntimeFilterFactory.ZERO_SET;
					}
				}
				org.hibernate.Filter filter = manager.enableFilter("annotationTypes");
				filter.setParameterList("annotationTypes", annotationTypeIds);
			}
		}
	}
}

