/*
	$Id: AnnotationBetweenRestriction.java 7456 2018-03-06 16:15:04Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Query;
import net.sf.basedb.util.EqualsHelper;

import org.hibernate.query.NativeQuery;

/**
	Restricts a query using annotation values between a low and high
	value.
	
	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.query.Annotations
	@base.modified $Date: 2018-03-06 17:15:04 +0100 (ti, 06 mar 2018) $
*/
public class AnnotationBetweenRestriction
	extends AnnotationRestriction
{
	private final Object lowValue;
	private final Object highValue;

	/**
		@since 2.4
		@deprecated In 3.5; Use {@link #AnnotationBetweenRestriction(String, AnnotationType, Object, Object, boolean, boolean)} instead
	*/
	@Deprecated
	public AnnotationBetweenRestriction(String alias, AnnotationType annotationType, Object lowValue, Object highValue, boolean includeInherited)
		throws InvalidDataException
	{
		this(alias, annotationType, lowValue, highValue, new Options().setIncludeInherited(includeInherited));
	}


	/**
		@since 3.0
		@deprecated In 3.5; Use {@link #AnnotationBetweenRestriction(String, int, Type, Object, Object, boolean, boolean)} instead
	*/
	@Deprecated
	public AnnotationBetweenRestriction(String alias, int annotationTypeId, Type valueType, Object lowValue, Object highValue, 
		boolean includeInheriting)
		throws InvalidDataException
	{
		this(alias, annotationTypeId, valueType, lowValue, highValue, true, includeInheriting);
	}

	/**
		@since 3.5
		@deprecated In 3.11, use {@link #AnnotationBetweenRestriction(String, AnnotationType, Object, Object, AnnotationRestriction.Options)} instead
	*/
	@Deprecated
	public AnnotationBetweenRestriction(String alias, AnnotationType annotationType, Object lowValue, Object highValue, 
		boolean includePrimary, boolean includeInherited)
		throws InvalidDataException
	{
		this(alias, annotationType, lowValue, highValue, new Options().setIncludePrimary(includePrimary).setIncludeInherited(includeInherited));
	}
	
	
	/**
		@since 3.5
		@deprecated In 3.11, use {@link #AnnotationBetweenRestriction(String, AnnotationType, Object, Object, AnnotationRestriction.Options)} instead
	*/
	@Deprecated
	public AnnotationBetweenRestriction(String alias, int annotationTypeId, Type valueType, Object lowValue, Object highValue, 
		boolean includePrimary, boolean includeInherited)
		throws InvalidDataException
	{
		super(alias, annotationTypeId, valueType, new Options().setIncludePrimary(includePrimary).setIncludeInherited(includeInherited));
		if (lowValue == null) throw new InvalidUseOfNullException("lowValue");
		if (highValue == null) throw new InvalidUseOfNullException("highValue");
		if (!valueType.isCorrectType(lowValue) && !(valueType.isNumerical() && lowValue instanceof Number))
		{
			throw new InvalidDataException("Low value '"+lowValue+"' is a "+
				lowValue.getClass()+", not a "+valueType);
		}
		if (!valueType.isCorrectType(highValue) && !(valueType.isNumerical() && highValue instanceof Number))
		{
			throw new InvalidDataException("Hig value '"+highValue+"' is a "+
				highValue.getClass()+", not a "+valueType);
		}
		this.lowValue = lowValue;
		this.highValue = highValue;
	}
	
	/**
		Create a new annotation restriction.
		
		@param alias The alias of a joined item where the annotations are
			located or null to use the root entity of the query
		@param annotationType The annotation type to use in the query
		@param lowValue The low value to use in the query, it must be of the 
			correct value type for the annotation as defined by the 
			valueType property
		@param highValue The high value to use in the query, it must be of the 
			correct value type for the annotation as defined by the 
			valueType property
		@param options Options that specify which annotations to consider when searching
		@throws InvalidDataException If any of the parameters are null
			or not follow the rules above.
		@since 3.11
	*/
	public AnnotationBetweenRestriction(String alias, AnnotationType annotationType, Object lowValue, Object highValue, 
		Options options)
		throws InvalidDataException
	{
		super(alias, annotationType, options);
		if (lowValue == null) throw new InvalidUseOfNullException("lowValue");
		if (highValue == null) throw new InvalidUseOfNullException("highValue");
		if (!valueType.isCorrectType(lowValue) && !(valueType.isNumerical() && lowValue instanceof Number))
		{
			throw new InvalidDataException("Low value '"+lowValue+"' is a "+
				lowValue.getClass()+", not a "+valueType);
		}
		if (!valueType.isCorrectType(highValue) && !(valueType.isNumerical() && highValue instanceof Number))
		{
			throw new InvalidDataException("Hig value '"+highValue+"' is a "+
				highValue.getClass()+", not a "+valueType);
		}
		this.lowValue = lowValue;
		this.highValue = highValue;
	}

	/*
		From the AnnotationRestriction class
		-------------------------------------------
	*/
	@Override
	String getRestrictionSql(Query query, DbControl dc)
		throws BaseException
	{
		org.hibernate.dialect.Dialect dialect = HibernateUtil.getDialect();
		String valueCol = "v."+dialect.openQuote()+"value"+dialect.closeQuote();
		if (valueType == Type.TIMESTAMP)
		{
			valueCol = HibernateUtil.getDbEngine().castToDate(valueCol);
		}

		return ":lowValue <= " + valueCol + " AND " + valueCol + " <= :highValue";
	}
	
	@Override
	void setRestrictionParameters(NativeQuery<?> query)
	{
		Type parameterType = valueType;
		if (lowValue instanceof Double) parameterType = Type.DOUBLE;
		if (parameterType == Type.TIMESTAMP) parameterType = Type.DATE;
		org.hibernate.type.Type hType = parameterType.getTypeWrapper().getHibernateType();
		query.setParameter("lowValue", lowValue, hType);
		query.setParameter("highValue", highValue, hType);
	}
	// -----------------------------------------

	/*
		From the Object class
		-------------------------------------------
	*/
	/**
		This restriction is equal to another AnnotationBetweenRestriction if they 
		have the same annotation type, upper and lower bound and inheritance setting.
	*/
	@Override
	public boolean equals(Object other)
	{
		Boolean eq = maybeEquals(other);
		if (eq != null) return eq;
		AnnotationBetweenRestriction o = (AnnotationBetweenRestriction)other;
		return EqualsHelper.equals(lowValue, o.lowValue) && 
			EqualsHelper.equals(highValue, o.highValue);
	}
	@Override
	public int hashCode()
	{
		return super.hashCode() + EqualsHelper.hashCode(lowValue, highValue);
	}
	@Override
	public String toString()
	{
		return "#" + annotationTypeId + " BETWEEN " + lowValue + " AND " + highValue;
	}
	// -------------------------------------------

	

}

