/*
	$Id$

	Copyright (C) 2005 Johan Enell
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Set;

import net.sf.basedb.core.data.ArrayDesignPlateData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
 
	@base.modified $Date$
	@author Enell
	@version 2.0
*/
public class ArrayDesignPlate
	extends ChildItem
{
	/**
		The type of item represented by this class.
		@see Item#ARRAYDESIGNPLATE
		@see #getType()
	*/
	public static final Item TYPE = Item.ARRAYDESIGNPLATE;
	
	/**
		This filter will only return items if the logged in user has
		generic read permission to array designs.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = 
		new QueryRuntimeFilterFactory.ChildFilter(TYPE, Item.CLIENT);

	/**
		Get a query that returns ArrayDesignPlates connections for the
		specified array design.
		@param design The array design to retreive plate connections for, null is allowed if
			the logged in user has generic READ permission for array designs in which case
			all plate connections will be retrieved
		@return An {@link ItemQuery} object
		@see ArrayDesign#getArrayDesignPlates()
	*/
	public static ItemQuery<ArrayDesignPlate> getQuery(ArrayDesign design)
	{
		ItemQuery<ArrayDesignPlate> query = null;
		if (design != null)
		{
			query = new ItemQuery<ArrayDesignPlate>(ArrayDesignPlate.class, null);
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("arrayDesign"), 
					Hql.entity(design)
				)
			);
		}
		else
		{
			query = new ItemQuery<ArrayDesignPlate>(ArrayDesignPlate.class, RUNTIME_FILTER);
		}
		return query;
	}
	
	/**
		Get a <code>ArrayDesignPlate</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>ArrayDesignPlate</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	static ArrayDesignPlate getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException,
			BaseException
	{
		ArrayDesignPlate adp = dc.loadItem(ArrayDesignPlate.class, id);
		if (adp == null) throw new ItemNotFoundException("ArrayDesignPlate[id=" + id + "]");
		return adp;
	}
	
	ArrayDesignPlate(ArrayDesignPlateData arrayDesignPlateData)
	{
		super(arrayDesignPlateData);
	}
	
	@Override
	ArrayDesignPlateData getData()
	{
		return (ArrayDesignPlateData)getBasicData();
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}

	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Always TRUE. It is used by the array design. 
	*/
	@Override
	public boolean isUsed() 
		throws BaseException
	{
		return true;
	}
	/**
		Get the array design this plate is used on.
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
		throws BaseException
	{
		Set<ItemProxy> using = super.getUsingItems();
		ArrayDesign item = getArrayDesign();
		using.add(new ItemProxy(item.getId(), item.getType()));
		return using;
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getArrayDesign();
	}
	// -------------------------------------------

	/**
	  	Gets the <code>Plate</code> that belongs to this item.
	  	@return A {@link Plate} object
	  	@throws PermissionDeniedException If the logged in user doesn't 
	  		have {@link Permission#READ} permission. 
	  	@throws BaseException If there is another error.
	*/
	public Plate getPlate()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Plate.class, getData().getPlate());
	}
	
	/**
	  	Gets this item's <code>ArrayDesign</code>. 
	  	@return An {@link ArrayDesign} object.
	  	@throws PermissionDeniedException If the logged in user doesn't 
	  		have {@link Permission#READ} permission.
	  	@throws BaseException If there is another error.
	*/
	public ArrayDesign getArrayDesign()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(ArrayDesign.class, getData().getArrayDesign());
	}
	
	/** 
	  	Gets the position of this item.
	*/
	public int getPosition()
	{
		return getData().getPosition();
	}
}
