/**
	$Id: DynamicSpotQueryJoiners.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.AutoJoiner;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.ReporterListExpression;

/**
	This class collects <code>AutoJoiner</code>:s for <code>DynamicSpotQuery</code>
	queries. The implementations are thread-safe and can be re-used between threads.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public final class DynamicSpotQueryJoiners
{

	/**
		This instance joins extra values with a left join.
	*/
	public static final ExtraValueJoiner EXTRA_VALUE_LEFT = 
		new ExtraValueJoiner(JoinType.LEFT);

	/**
		This instance joins extra values with an inner join.
	*/
	public static final ExtraValueJoiner EXTRA_VALUE_INNER = 
		new ExtraValueJoiner(JoinType.INNER);

	/**
		Get a joiner instance for extra values for the specified join type.
		@param joinType The type of join
		@return {@link #EXTRA_VALUE_INNER} or {@link #EXTRA_VALUE_LEFT}
	 */
	public static ExtraValueJoiner getExtraValueJoiner(JoinType joinType)
	{
		return joinType == JoinType.LEFT ? EXTRA_VALUE_LEFT : EXTRA_VALUE_INNER;
	}
	
	/**
		This instance joins reporters with a left join.
	*/
	public static final ReporterJoiner REPORTERS_LEFT = 
		new ReporterJoiner(JoinType.LEFT);
	
	/**
		This instance joins reporters with an inner join.
	*/
	public static final ReporterJoiner REPORTERS_INNER = 
		new ReporterJoiner(JoinType.INNER);

	/**
		Get a joiner instance for reporters for the specified join type.
		@param joinType The type of join
		@return {@link #REPORTERS_LEFT} or {@link #REPORTERS_INNER}
	 */
	public static ReporterJoiner getReporterJoiner(JoinType joinType)
	{
		return joinType == JoinType.LEFT ? REPORTERS_LEFT : REPORTERS_INNER;
	}

	/**
		This instance joins raw data with a left join.
	*/
	public static final RawDataJoiner RAWDATA_LEFT = 
		new RawDataJoiner(JoinType.LEFT);
	
	/**
		This instance joins raw data with an inner join.
	*/
	public static final RawDataJoiner RAWDATA_INNER = 
		new RawDataJoiner(JoinType.INNER);

	/**
		Get a joiner instance for raw data for the specified join type.
		@param joinType The type of join
		@return {@link #RAWDATA_LEFT} or {@link #RAWDATA_INNER}
	 */
	public static RawDataJoiner getRawDataJoiner(JoinType joinType)
	{
		return joinType == JoinType.LEFT ? RAWDATA_LEFT : RAWDATA_INNER;
	}
	
	/**
		This instance joins reporter lists with a left join, unless the
		{@link ReporterListExpression#getAutoJoinType()} method overrides 
		the default.
	*/
	public static final ReporterListJoiner REPORTER_LIST_LEFT = 
		new ReporterListJoiner(JoinType.LEFT);
	
	/**
		This instance joins reporter lists with an inner join, unless the
		{@link ReporterListExpression#getAutoJoinType()} method overrides 
		the default.
	*/
	public static final ReporterListJoiner REPORTER_LIST_INNER = 
		new ReporterListJoiner(JoinType.INNER);
	
	/**
		Get a joiner instance for raw data for the specified join type.
		@param joinType The type of join
		@return {@link #REPORTER_LIST_LEFT} or {@link #REPORTER_LIST_INNER}
	 */
	public static ReporterListJoiner getReporterListJoiner(JoinType joinType)
	{
		return joinType == JoinType.LEFT ? REPORTER_LIST_LEFT : REPORTER_LIST_INNER;
	}

	
	/**
		Joiner for extra values.
		@author nicklas
		@version 2.0
		@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
	*/
	private static class ExtraValueJoiner
		implements AutoJoiner<DynamicSpotQuery, ExtraValueExpression>
	{
		private final JoinType joinType;

		private ExtraValueJoiner(JoinType joinType)
		{
			this.joinType = joinType;
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		@Override
		public Class<DynamicSpotQuery> getQueryClass()
		{
			return DynamicSpotQuery.class;
		}
		@Override
		public Class<ExtraValueExpression> getElementClass()
		{
			return ExtraValueExpression.class;
		}
		@Override
		public void joinIfNeeded(DynamicSpotQuery query, ExtraValueExpression element)
		{
			query.joinExtraValue(element.getExtraValue(), joinType);
		}
		@Override
		public void reset(DynamicSpotQuery query)
		{}
		// -------------------------------------------
	}

	/**
		Joiner for reporters
		@author nicklas
		@version 2.0
		@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
	*/
	private static class ReporterJoiner
		implements AutoJoiner<DynamicSpotQuery, ReporterPropertyExpression>
	{
		private final JoinType joinType;
	
		private ReporterJoiner(JoinType joinType)
		{
			this.joinType = joinType;
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		@Override
		public Class<DynamicSpotQuery> getQueryClass()
		{
			return DynamicSpotQuery.class;
		}
		@Override
		public Class<ReporterPropertyExpression> getElementClass()
		{
			return ReporterPropertyExpression.class;
		}
		@Override
		public void joinIfNeeded(DynamicSpotQuery query, ReporterPropertyExpression element)
		{
			query.joinReporters(joinType, element.useCloned(query));
		}
		@Override
		public void reset(DynamicSpotQuery query)
		{}
		// -------------------------------------------
	}

	/**
		Joiner for raw data.
		@author nicklas
		@version 2.0
		@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
	*/
	private static class RawDataJoiner
		implements AutoJoiner<DynamicSpotQuery, RawDataPropertyExpression>
	{
		private final JoinType joinType;
	
		private RawDataJoiner(JoinType joinType)
		{
			this.joinType = joinType;
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		@Override
		public Class<DynamicSpotQuery> getQueryClass()
		{
			return DynamicSpotQuery.class;
		}
		@Override
		public Class<RawDataPropertyExpression> getElementClass()
		{
			return RawDataPropertyExpression.class;
		}
		@Override
		public void joinIfNeeded(DynamicSpotQuery query, RawDataPropertyExpression element)
		{
			query.joinRawData(joinType);
		}
		@Override
		public void reset(DynamicSpotQuery query)
		{}
		// -------------------------------------------
	}

	/**
		Joiner for reporter lists.
		@author nicklas
		@version 2.0
		@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
	*/
	private static class ReporterListJoiner
		implements AutoJoiner<DynamicSpotQuery, ReporterListExpression>
	{
		private final JoinType joinType;
	
		private ReporterListJoiner(JoinType joinType)
		{
			this.joinType = joinType;
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		@Override
		public Class<DynamicSpotQuery> getQueryClass()
		{
			return DynamicSpotQuery.class;
		}
		@Override
		public Class<ReporterListExpression> getElementClass()
		{
			return ReporterListExpression.class;
		}
		@Override
		public void joinIfNeeded(DynamicSpotQuery query, ReporterListExpression element)
		{
			JoinType jt = element.getAutoJoinType();
			query.joinReporterList(element.getReporterList(), jt == null ? joinType : jt);
		}
		@Override
		public void reset(DynamicSpotQuery query)
		{}
		// -------------------------------------------
	}

}
