/*
	$Id $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ItemSubtypeFileTypeData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	This class represents a specific {@link DataFileType}
	that is associated with an item subtype.

	@author Nicklas
	@since 3.0
	@base.modified $Date $
*/
public class ItemSubtypeFileType
	extends BasicChildItem
	implements UsableDataFileType
{
	
	/**
		The type of item represented by this class.
		@see Item#ITEMSUBTYPEFILETYPE
		@see #getType()
	*/
	public static final Item TYPE = Item.ITEMSUBTYPEFILETYPE;

	/**
		This filter will only return items if the logged in user has
		generic read permission to item subtypes.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = 
		new QueryRuntimeFilterFactory.ChildFilter(TYPE, Item.ITEMSUBTYPE);

	/**
		Create a new item linking a file type to an item subtype.
		@param subtype The item subtype
		@param fileType The file type
	*/
	static ItemSubtypeFileType getNew(DbControl dc, ItemSubtype subtype,
		DataFileType fileType)
	{
		ItemSubtypeFileType type = dc.newItem(ItemSubtypeFileType.class);
		ItemSubtypeFileTypeData data = type.getData();
		data.setItemSubtype(subtype.getData());
		data.setDataFileType(fileType.getData());
		return type;
	}
	
	/**
		Get a <code>PlatformFileType</code> object when you know the ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>PlatformFileType</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static ItemSubtypeFileType getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ItemSubtypeFileType p = dc.loadItem(ItemSubtypeFileType.class, id);
		if (p == null) throw new ItemNotFoundException("ItemSubtypeFileType[id="+id+"]");
		return p;
	}
	
	/**
		Get the file types that have been registered for a specific item
		subtype.
		
		@param subtype An item subtype object or null to return items for
			all platforms
		@return A query returning <code>ItemSubtypeFileType</code> items
	*/
	public static ItemQuery<ItemSubtypeFileType> getQuery(ItemSubtype subtype)
	{
		ItemQuery<ItemSubtypeFileType> query = 
			new ItemQuery<ItemSubtypeFileType>(ItemSubtypeFileType.class, RUNTIME_FILTER);
		if (subtype != null)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("itemSubtype"),
					Hql.entity(subtype)
				)
			);
		}
		return query;
	}
	
	/**
		Get the item subtype types that have been registered for a specific data
		file type.
		
		@param fileType An data file type or null to return items for
			all file types
		@return A query returning <code>ItemSubtypeFileType</code> items
		@since 3.11
	*/
	public static ItemQuery<ItemSubtypeFileType> getQuery(DataFileType fileType)
	{
		ItemQuery<ItemSubtypeFileType> query = 
			new ItemQuery<ItemSubtypeFileType>(ItemSubtypeFileType.class, RUNTIME_FILTER);
		if (fileType != null)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("dataFileType"),
					Hql.entity(fileType)
				)
			);
		}
		return query;
	}

	/**
		Creates a new member item from the given data.
		@param data the data.
	*/
	ItemSubtypeFileType(ItemSubtypeFileTypeData data)
	{
		super(data);
	}
	
	@Override
	ItemSubtypeFileTypeData getData()
	{
		return (ItemSubtypeFileTypeData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicChildItem class
		-------------------------------------------
	*/
	@Override
	Item getParentType()
	{
		return Item.ITEMSUBTYPE;
	}
	// -------------------------------------------
	
	/**
		Get the item subtype.
		@return A subtype item
	*/
	public ItemSubtype getItemSubtype()
	{
		return getDbControl().getItem(ItemSubtype.class, getData().getItemSubtype());		
	}


	/**
		Get the type of the file.
		@return A file set member type item
	*/
	@Override
	public DataFileType getDataFileType()
	{
		return getDbControl().getItem(DataFileType.class, getData().getDataFileType());
	}
	
	/**
		Check if a file of this type is required or not. 
		Note! The requirement is not enforced by the core. It should be used as a
		hint to client applications so they can create a proper GUI. To check
		if a {@link FileSet} contains all files that have been marked as required
		use {@link FileSet#hasAllRequiredFiles()}.
	*/
	@Override
	public boolean isRequired()
	{
		return getData().isRequired();
	}
	/**
		Set the required flag for this file type.
	*/
	public void setRequired(boolean required)
	{
		checkPermission(Permission.WRITE);
		getData().setRequired(required);
	}
	
	/**
		Check if more than one file of this type is allowed in a file set or not.
		Note! The requirement is not enforced by the core. It should be used as a
		hint to client applications so they can create a proper GUI. 
	*/
	@Override
	public boolean getAllowMultiple()
	{
		return getData().getAllowMultiple();
	}
	
	/**
		Allow or disallow multiple files of this file type in a file set.
	*/
	public void setAllowMultiple(boolean multiple)
	{
		checkPermission(Permission.WRITE);
		getData().setAllowMultiple(multiple);
	}

}
