/*
	$Id$

	Copyright (C) 2011 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ArraySlideData;
import net.sf.basedb.core.data.BioMaterialEventData;
import net.sf.basedb.core.data.PhysicalBioAssayData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.util.collections.BasicItemToIdTransformer;

import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Set;
import java.util.HashSet;

/**
	A physical bioassay is a representation of the process of 
	performing an experiment one or more {@link Extract}:s.

	@author Nicklas
	@since 3.0
	@base.modified $Date$
*/
public class PhysicalBioAssay 
	extends AnnotatedItem
	implements Registered, Subtypable, Listable
{
	/**
		The type of item represented by this class.
		@see Item#PHYSICALBIOASSAY
		@see #getType()
	*/
	public static final Item TYPE = Item.PHYSICALBIOASSAY;

	/**
		The id for the <code>PhysicalBioAssayType</code> item representing a
		hybridization.
	*/
	public static final String HYBRIDIZATION = "net.sf.basedb.core.PhysicalBioAssayType.HYBRIDIZATION";

	/**
		The id for the <code>PhysicalBioAssayType</code> item representing a
		flow cell.
	*/
	public static final String CLONAL_AMPLIFICATION = "net.sf.basedb.core.PhysicalBioAssayType.CLONAL_AMPLIFICATION";

	
	/**
		Create a new <code>PhysicalBioAssay</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>PhysicalBioAssay</code> item
		@throws BaseException If there is an error
	*/
	public static PhysicalBioAssay getNew(DbControl dc)
		throws BaseException
	{
		return getNew(dc, null);
	}

	/**
		Create a new <code>PhysicalBioAssay</code> item as part of a
		master bioplate event.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param master The master bioplate event participant or null
		@return The new <code>PhysicalBioAssay</code> item
		@throws BaseException If there is an error
		@since 2.17
	*/
	public static PhysicalBioAssay getNew(DbControl dc, BioPlateEventParticipant master)
		throws BaseException
	{
		PhysicalBioAssay h = dc.newItem(PhysicalBioAssay.class);
		h.setName("New physical bioassay");
		h.setSize(1);
		h.setBioPlateEventParticipant(master);
		return h;
	}

	/**
		Get a <code>Hybridization</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Hybridization</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static PhysicalBioAssay getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		PhysicalBioAssay pba = dc.loadItem(PhysicalBioAssay.class, id);
		if (pba == null) throw new ItemNotFoundException("PhysicalBioAssay[id="+id+"]");
		return pba;
	}

	/**
		Get a query that returns physical bioassays.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<PhysicalBioAssay> getQuery()
	{
		return new ItemQuery<PhysicalBioAssay>(PhysicalBioAssay.class);
	}

	private BioMaterialEvent creationEvent;
	private BioPlateEventParticipant bioPlateEventParticipant;

	PhysicalBioAssay(PhysicalBioAssayData physicalBioAssayData)
	{
		super(physicalBioAssayData);
	}
	
	@Override
	PhysicalBioAssayData getData()
	{
		return (PhysicalBioAssayData)getBasicData();
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		Get the extracts and array slide. Note that this method will
		include all extracts, even on bioassays with multiple positions.
		@see #getAnnotatableParents(int, Extract)
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		return getAnnotatableParents(0, (Collection<Extract>)null);
	}
	
	/**
		Get the protocol from the creation event.
		@since 2.2
	*/
	@Override
	public Protocol getProtocol()
	{
		return getCreationEvent().getProtocol();
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return getCreationEvent().getEntryDate();
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		getCreationEvent().setEntryDate(entryDate);
	}
	// -------------------------------------------
	/*
		From the Subtypable interface
		-----------------------------
	*/
	@Override
	@SupportsPushAnnotations
	@SubtypableRelatedItems({Item.HARDWARE, Item.PROTOCOL, Item.KIT, Item.EXTRACT})
	public ItemSubtype getItemSubtype()
	{
		return getDbControl().getItem(ItemSubtype.class, getData().getItemSubtype());
	}
	@Override
	public void setItemSubtype(ItemSubtype subtype)
	{
		checkPermission(Permission.WRITE);
		if (subtype != null)
		{
			subtype.setOnItem(this);
		}
		else
		{
			getData().setItemSubtype(null);
		}
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check that:
		<ul>
		<li>No {@link DerivedBioAssay}:s use the physical bioassay
		</ul>
	*/
	@Override
	public boolean isUsed() 
		throws BaseException
	{
		return countRootDerivedBioAssays() > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>(root) {@link DerivedBioAssay}:s created from this physical bioassay
		</ul>
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_ROOT_DERIVEDBIOASSAYS_FOR_PHYSICALBIOASSAY", Integer.class, "dba.id");
			/*
				SELECT {1}
				FROM DerivedBioAssayData dba
				INNER JOIN dba.physicalBioAssays pba
				WHERE pba = :bioAssay
				AND dba.root = true
			*/
		query.setParameter("bioAssay", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.DERIVEDBIOASSAY, query);
		return using;
	}
	/**
		Make sure the creation event is created at the same time this item
		is created and that source biomaterials get their used quantities back
		if this item is deleted. All events are deleted by cascade mapping in Hibernate.
	*/
	@Override
	void onBeforeCommit(Transactional.Action action)
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (action == Transactional.Action.CREATE)
		{
			getDbControl().saveItem(getCreationEvent());
		}
		else if (action == Transactional.Action.DELETE)
		{
			getCreationEvent().clearSources();
		}
	}
	// -------------------------------------------

	/**
		Get the extracts, possible on a specific assay position or only a specific
		extract, and the array slide.
		@param position If &gt; 0 only include the extracts on that
			position in the bioassay (see {@link BioMaterialEventSource#getPosition()}.
			If &lt;= 0 include all extracts.
		@param extract If not null, only include that extract (if present on the
			bioassay)
		@since 3.0
	*/
	public Set<Annotatable> getAnnotatableParents(int position, Extract extract)
		throws BaseException
	{
		return getAnnotatableParents(position, extract == null ? null : Collections.singleton(extract));
	}

	/**
		Get the extracts, possible on a specific assay position or only a given set of specific
		extracts, and the array slide.
		@param position If &gt; 0 only include the extracts on that
			position in the bioassay (see {@link BioMaterialEventSource#getPosition()}.
			If &lt;= 0 include all extracts.
		@param extracts If not null or empty, only include the given extracts (if present on the
			bioassay)
		@since 3.2
	*/
	@SuppressWarnings("unchecked")
	public Set<Annotatable> getAnnotatableParents(int position, Collection<Extract> extracts)
		throws BaseException
	{
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		PhysicalBioAssayData data = getData();
		DbControl dc = getDbControl();

		BioMaterialEvent event = getCreationEvent();
		BioMaterialEventData eventData = event.getData();
		
		ItemQuery<Extract> query = (ItemQuery<Extract>)event.getSources();
		query.include(Include.ALL);
		if (position > 0)
		{
			query.restrict(
				Restrictions.eq(
					Hql.property("srcevt", "position"), 
					Expressions.integer(position)
				)
			);
		}
		if (extracts != null && extracts.size() > 0)
		{
			query.restrict(Restrictions.in(Hql.property("id"), Expressions.parameter("extracts")));
			Set<Integer> extractIds = new HashSet<Integer>();
			new BasicItemToIdTransformer().transform(extracts, extractIds);
			query.setParameter("extracts", extractIds, Type.INT);
		}
		annotatable.addAll(query.list(dc));

		addAnnotatableParents(dc, annotatable, 
			data.getArraySlide(),
			eventData.getHardware(),
			eventData.getProtocol(),
			eventData.getKit()
		);

		return annotatable;
	}

	
	/**
		Get the event that represents the creation of this hybridization.
		Use this event to add source labeled extracts, a protocol, etc.
		@return A {@link BioMaterialEvent} data.
	 	@throws PermissionDeniedException If the logged in user don't have read permissions 
	 	@throws BaseException If there is another error.
	*/
	public BioMaterialEvent getCreationEvent()
		throws PermissionDeniedException, BaseException
	{
		if (creationEvent == null)
		{
			DbControl dc = getDbControl();
			if (!isInDatabase())
			{
				creationEvent = BioMaterialEvent.getNewBioAssayEvent(dc, this, bioPlateEventParticipant);
			}
			else
			{
				creationEvent = getDbControl().getItem(BioMaterialEvent.class, getData().getCreationEvent());
			}
		}
		return creationEvent;
	}

	void setBioPlateEventParticipant(BioPlateEventParticipant bioPlateEventParticipant)
	{
		this.bioPlateEventParticipant = bioPlateEventParticipant;
	}

	
	/**
		Get the {@link ArraySlide} this hyridization is using. A single
		slide can only be used on one hybridization.

		@return The <code>ArraySlide</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the array slide
		@throws BaseException If there is another error.
	*/
	public ArraySlide getArraySlide()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(ArraySlide.class, getData().getArraySlide());
	}
	
	/**
		Set the {@link ArraySlide} this hybridization is using.

		@param arraySlide The array slide, or null
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission for the hybridization or use permission for the array slide
		@throws BaseException If there is another error.
	*/
	public void setArraySlide(ArraySlide arraySlide)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		ArraySlideData data = null;
		if(arraySlide != null) 
		{
			arraySlide.checkPermission(Permission.USE);
			data = arraySlide.getData();
			data.setPhysicalBioAssay(getData());
		}
		getData().setArraySlide(data);
	}
	
	/**
		Get the size of the bioassay. The size is the number of
		independent positions for biomaterial.
		The default value is 1.
	*/
	public int getSize()
	{
		return getData().getSize();
	}
	/**
		Set the size of the bioassay.
		@param size A value &gt; 0
		@throws NumberOutOfRangeException If the value is &lt;=0
	*/
	public void setSize(int size)
	{
		checkPermission(Permission.WRITE);
		if (size <= 0) throw new NumberOutOfRangeException("size", size, 0, false);
		getData().setSize(size);
	}
	
	/**
		Get a query that returns extracts connected to
		this bioassay.
		@param position If &gt; 0 only extracts on the specific position
			are returned by the query, if &lt;= 0, all extracts
			are returned
		@return An {@link ItemQuery} object
	*/
	@SuppressWarnings("unchecked")
	public ItemQuery<Extract> getExtracts(int position)
	{
		ItemQuery<Extract> query = (ItemQuery<Extract>)getCreationEvent().getSources();
		if (position > 0)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("srcevt", "position"), 
					Expressions.integer(position)
				)
			);
		}
		return query;
	}
	
	/**
		Get a query that returns all derived bioassay created from this 
		physical bioassay.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<DerivedBioAssay> getDerivedBioAssays()
	{
		ItemQuery<DerivedBioAssay> query = DerivedBioAssay.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("physicalBioAssay"), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Get a query that returns all root derived bioassay created from this 
		physical bioassay.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<DerivedBioAssay> getRootDerivedBioAssays()
	{
		ItemQuery<DerivedBioAssay> query = DerivedBioAssay.getQuery();
		query.joinPermanent(Hql.innerJoin("physicalBioAssays", "pba"));
		query.restrictPermanent(Restrictions.eq(Hql.alias("pba"), Hql.entity(this)));
		query.restrictPermanent(Restrictions.eq(Hql.property("root"), Expressions.bool(true)));
		return query;
	}
	
	
	/**
		Counts all root derived bioassays that have been created from this physical bioassay.
		The count includes bioassays that the logged in user doesn't have access to.
		@return The number of root bioassay
	 	@throws BaseException If the bioassay could not be counted. 
	 */
	public long countRootDerivedBioAssays()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_ROOT_DERIVEDBIOASSAYS_FOR_PHYSICALBIOASSAY", Long.class, "count(*)");
		/*
			SELECT {1}
			FROM DerivedBioAssayData dba
			INNER JOIN dba.physicalBioAssays pba
			WHERE pba = :bioAssay
			AND dba.root = true
		*/
		query.setParameter("bioAssay", getId(), TypeWrapper.H_INTEGER);
		return HibernateUtil.loadData(query);
	}


}
