/*
	$Id: PluginExecutionRequest.java 7329 2017-04-07 07:21:57Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.Application.Pinger;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.signal.DelegatingSignalHandler;
import net.sf.basedb.core.signal.Signal;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalReceiver;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.SignalTransporter;

/**
	Executes a job. You will get an instance of
	this class when:
	
	<ol>
	<li>You want to execute a job: {@link Job#execute(ProgressReporter, String)}
	</ol>
	
	
	You should use the following protocol:
	<ol>
	<li>Close your currently opened {@link DbControl} if you doesn't need it
		and you expect that the job takes a long time to execute. The job will 
		open it's own connection(s) to the database.
		
	<li>Call {@link #invoke()} to start executing of the job.

	<li>Check the {@link PluginResponse} and display a user-friendly message.
	</ol>
	
	Note! You do not have to call the {@link #done()} method, it is called 
	automatically by the core once the job has finished executing.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2017-04-07 09:21:57 +0200 (fr, 07 apr 2017) $
*/
public final class PluginExecutionRequest
	extends PluginRequest<Plugin>
	implements Runnable
{
	private Job.ProgressReporterImpl progress;
	private SignalReceiver signalReceiver = null;
	
	PluginExecutionRequest(SessionControl sc, Plugin plugin, String command,
		PluginConfiguration pluginConfiguration, ParameterValuesImpl configurationParameters,
		Job job, ParameterValuesImpl jobParameters, PluginDefinition pluginDefinition)
	{
		super(sc, plugin, command, pluginConfiguration, configurationParameters, job, jobParameters, pluginDefinition);
		job.removeState();
	}
	
	PluginExecutionRequest(PluginRequest<? extends Plugin> copyFrom, String command)
	{
		super(copyFrom, command);
	}

	/*
		From the PluginRequest class
		-------------------------------------------
	*/
	@Override
	public PluginResponse invoke()
	{
		PluginResponse pluginResponse = new PluginResponse(null, false, false, getJob().isDryRun());
		Request request = new RequestImpl(false);
		Response response = pluginResponse.getResponseImpl();
		DelegatingSignalHandler signalHandler = null;
		boolean isShutdown = false;
		Pinger pinger = null;
		try
		{
			Plugin plugin = getPlugin();
			String signalId = null;
			Class<? extends SignalTransporter> signalTransporter = null;
			if (plugin instanceof SignalTarget)
			{
				// Register a signal handler, receiver and transporter
				// We wrap the signal handler in a DelegatingSignalHandler
				// since we need to know if the SHUTDOWN signal has been sent
				SignalHandler sh = ((SignalTarget)plugin).getSignalHandler();
				if (signalReceiver != null && sh != null)
				{
					signalHandler = new DelegatingSignalHandler(sh);
					signalId = signalReceiver.registerSignalHandler(signalHandler);
					signalTransporter = signalReceiver.getSignalTransporterClass();
				}
			}
			if (progress != null) progress.setStarted(signalTransporter, signalId);
			// The pinger keeps the session alive while the plug-in is executing
			pinger = Application.newPinger(getSessionControl());
			plugin.run(request, response, progress);
		}
		catch (Throwable t)
		{
			pluginResponse.setError("Error invoking plugin: " + t.getMessage() , t);
		}
		finally
		{
			if (signalReceiver != null && signalHandler != null)
			{
				signalReceiver.unregisterSignalHandler(signalHandler);
			}
			if (pinger != null) pinger.stop();
		}
		done();
		if (progress != null) progress.setEnded(pluginResponse, signalHandler != null && signalHandler.hasReceived(Signal.SHUTDOWN));
		return pluginResponse;
	}
	// -------------------------------------------
	/*
		From the Runnable interface
		-------------------------------------------
	*/
	@Override
	public void run()
	{
		invoke();
	}
	// -------------------------------------------

	/**
		Initialise a progress reporter that the plugin can use to store
		progress information in the database. This object is only
		available while executing the plugin.
	*/
	void setProgressReporter(Job.ProgressReporterImpl progress)
	{
		this.progress = progress;
	}
	
	/**
		Register a signal receiver that will receive signals for this job.
		The plug-in must implement the {@link SignalTarget} interface
		and provide a {@link SignalHandler}. If not, the job will not be
		registered with the signal receiver and it will not be possible to
		send signals to the job.
		
		@param signalReceiver The signal recevier to use
	*/
	public void registerSignalReceiver(SignalReceiver signalReceiver)
	{
		this.signalReceiver = signalReceiver;
	}
	
}
