/**
	$Id: StatelessSessionWrapper.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import java.io.Serializable;
import java.sql.Connection;

import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.LockMode;
import org.hibernate.Session;
import org.hibernate.StatelessSession;
import org.hibernate.Transaction;
import org.hibernate.internal.StatelessSessionImpl;
import org.hibernate.procedure.ProcedureCall;
import org.hibernate.query.Query;
import org.hibernate.query.NativeQuery;

import net.sf.basedb.core.HibernateUtil;

/**
	Wrapper around stateless Hibernate sessions that includes a query cache.
	This wrapper forwards all calls to the underlying real Hibernate 
	session.

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (må, 22 maj 2017) $
*/
public class StatelessSessionWrapper
	extends AbstractQueryCache
	implements StatelessSession
{

	private static final long serialVersionUID = -771179312268431157L;

	private final StatelessSession session;
	/**
		Creates a new session wrapper
		@param session The real Hibernate session
	*/
	public StatelessSessionWrapper(StatelessSession session)
	{
		this.session = session;
	}

	/**
		Get the underlying parent session from Hibernate.
	*/
	public StatelessSession getParentSession()
	{
		return session;
	}

	
	/**
		Cancels the currently executing query in the same manner as {@link Session#cancelQuery()}.
		Since this method is not part of the StatelessSession interface this only works if
		the underlying parent session is the Hibernate implementation.
	*/
	public void cancelQuery() 
		throws HibernateException
	{
		if (session instanceof StatelessSessionImpl)
		{
			((StatelessSessionImpl)session).getJdbcCoordinator().cancelLastQuery();
		}
	}
	
	/*
		From the StatelessSession interface
		-----------------------------------
	*/
	@Override
	public Transaction beginTransaction()
	{
		return session.beginTransaction();
	}

	@Override
	public void close()
	{
		session.close();
	}

	@Override
	@Deprecated
	public Connection connection()
	{
		return ((StatelessSessionImpl)session).connection();
	}

	@SuppressWarnings("rawtypes")
	@Override
	@Deprecated
	public Criteria createCriteria(Class persistentClass, String alias)
	{
		return session.createCriteria(persistentClass, alias);
	}

	@Override
	@SuppressWarnings("rawtypes")
	@Deprecated
	public Criteria createCriteria(Class persistentClass)
	{
		return session.createCriteria(persistentClass);
	}

	@Override
	@Deprecated
	public Criteria createCriteria(String entityName, String alias)
	{
		return session.createCriteria(entityName, alias);
	}

	@Override
	@Deprecated
	public Criteria createCriteria(String entityName)
	{
		return session.createCriteria(entityName);
	}

	/**
		If the same query has been executed before, return the cached
		instance, otherwise create a new query from the parent session
		and cache it.
	*/
	@Override
	public Query<?> createQuery(String queryString)
		throws HibernateException
	{
		Query<?> q = getCachedQuery(queryString);
		if (q == null)
		{
			q = session.createQuery(queryString);
			q.setFetchSize(HibernateUtil.getJdbcFetchSize());
			cacheQuery(queryString, q);
		}
		return q;
	}

	/**
		If the same query has been executed before, return the cached
		instance, otherwise create a new query from the parent session
		and cache it.
	*/
	@Override
	@Deprecated
	public NativeQuery<?> createSQLQuery(String queryString) 
		throws HibernateException
	{
		Query<?> q = getCachedQuery(queryString);
		if (q == null || !(q instanceof NativeQuery))
		{
			q = session.createSQLQuery(queryString);
			q.setFetchSize(HibernateUtil.getJdbcFetchSize());
		}
		return (NativeQuery<?>)q;
	}

	@Override
	public void delete(Object entity)
	{
		session.delete(entity);
	}

	@Override
	public void delete(String entityName, Object entity)
	{
		session.delete(entityName, entity);
	}

	@Override
	@SuppressWarnings("rawtypes")
	public Object get(Class entityClass, Serializable id, LockMode lockMode)
	{
		return session.get(entityClass, id, lockMode);
	}

	@Override
	@SuppressWarnings("rawtypes")
	public Object get(Class entityClass, Serializable id)
	{
		return session.get(entityClass, id);
	}

	@Override
	public Object get(String entityName, Serializable id, LockMode lockMode)
	{
		return session.get(entityName, id, lockMode);
	}

	@Override
	public Object get(String entityName, Serializable id)
	{
		return session.get(entityName, id);
	}

	/**
		If the same query has been executed before, return the cached
		instance, otherwise create a new query from the parent session
		and cache it.
	*/
	@Override
	public Query<?> getNamedQuery(String queryName) 
		throws HibernateException
	{
		Query<?> q = getCachedQuery(queryName);
		if (q == null)
		{
			q = session.getNamedQuery(queryName);
			cacheQuery(queryName, q);
		}
		return q;
	}

	@Override
	public Transaction getTransaction()
	{
		return session.getTransaction();
	}

	@Override
	public Serializable insert(Object entity)
	{
		return session.insert(entity);
	}

	@Override
	public Serializable insert(String entityName, Object entity)
	{
		return session.insert(entityName, entity);
	}

	@Override
	public void refresh(Object entity, LockMode lockMode)
	{
		session.refresh(entity, lockMode);
	}

	@Override
	public void refresh(Object entity)
	{
		session.refresh(entity);
	}

	@Override
	public void refresh(String entityName, Object entity, LockMode lockMode)
	{
		session.refresh(entityName, entity, lockMode);
	}

	@Override
	public void refresh(String entityName, Object entity)
	{
		session.refresh(entityName, entity);
	}

	@Override
	public void update(Object entity)
	{
		session.update(entity);
	}

	@Override
	public void update(String entityName, Object entity)
	{
		session.update(entityName, entity);
	}
	// -------------------------------------------

	@Override
	public String getTenantIdentifier() 
	{
		return session.getTenantIdentifier();
	}

	@Override
	public ProcedureCall getNamedProcedureCall(String name) 
	{
		return session.getNamedProcedureCall(name);
	}

	@Override
	public ProcedureCall createStoredProcedureCall(String name) 
	{
		return session.createStoredProcedureCall(name);
	}

	@Override
	@SuppressWarnings("rawtypes")
	public ProcedureCall createStoredProcedureCall(String paramString,
			Class... paramArrayOfClass) 
	{
		return session.createStoredProcedureCall(paramString, paramArrayOfClass);
	}

	@Override
	public ProcedureCall createStoredProcedureCall(String paramString,
			String... paramArrayOfString) 
	{
		return session.createStoredProcedureCall(paramString, paramArrayOfString);
	}

	@Override
	public Integer getJdbcBatchSize() 
	{
		return session.getJdbcBatchSize();
	}

	@Override
	public boolean isConnected() 
	{
		return session.isConnected();
	}

	@Override
	public boolean isOpen()
	{
		return session.isOpen();
	}

	@Override
	public void setJdbcBatchSize(Integer batchSize) 
	{
		session.setJdbcBatchSize(batchSize);
	}

	@SuppressWarnings("unchecked")
	@Override
	public <R> Query<R> createNamedQuery(String queryName, Class<R> returnType)
	{
		Query<R> q = (Query<R>)getCachedQuery(queryName);
		if (q == null)
		{
			q = session.createNamedQuery(queryName, returnType);
			cacheQuery(queryName, q);
		}
		return q;
	}

	@Override
	public Query<?> createNamedQuery(String queryName) 
	{
		Query<?> q = getCachedQuery(queryName);
		if (q == null)
		{
			q = session.createNamedQuery(queryName);
			cacheQuery(queryName, q);
		}
		return q;
	}

	@SuppressWarnings("unchecked")
	@Override
	public <R> NativeQuery<R> createNativeQuery(String queryString, Class<R> resultClass) 
	{
		NativeQuery<R> q = (NativeQuery<R>)getCachedQuery(queryString);
		if (q == null)
		{
			q = session.createNativeQuery(queryString, resultClass);
			cacheQuery(queryString, q);
		}
		return q;
	}

	@Override
	public NativeQuery<?> createNativeQuery(String queryString, String resultSetMapping) 
	{
		NativeQuery<?> q = (NativeQuery<?>)getCachedQuery(queryString);
		if (q == null)
		{
			q = session.createNativeQuery(queryString, resultSetMapping);
			cacheQuery(queryString, q);
		}
		return q;
	}

	@Override
	public NativeQuery<?> createNativeQuery(String queryString) 
	{
		NativeQuery<?> q = (NativeQuery<?>)getCachedQuery(queryString);
		if (q == null)
		{
			q = session.createNativeQuery(queryString);
			cacheQuery(queryString, q);
		}
		return q;
	}

	@SuppressWarnings("unchecked")
	@Override
	public <R> Query<R> createQuery(String queryString, Class<R> resultClass) 
	{
		Query<R> q = (Query<R>)getCachedQuery(queryString);
		if (q == null)
		{
			q = session.createQuery(queryString, resultClass);
			cacheQuery(queryString, q);
		}
		return q;
	}

	@Override
	public NativeQuery<?> getNamedNativeQuery(String queryName) 
	{
		NativeQuery<?> q = (NativeQuery<?>)getCachedQuery(queryName);
		if (q == null)
		{
			q = session.getNamedNativeQuery(queryName);
			cacheQuery(queryName, q);
		}
		return q;
	}
	
	
}
