/**
	$Id $

	Copyright (C) 2016 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.data.LoggableData;

/**
	Represents a manual entry into the change history log.
	After creating an instance of this class, call 
	{@link DbControl#logEntry(ManualLogEntry)} to register the
	entry. If the entry is actually logged or not depends on if
	logging has been enabled and also on the extensions that handle
	logging. Note that the built-in logging will only log entries
	for items that implements the {@link LoggableData} interface.
	<p>
	
	The manual log entries are mainly intended for logging
	{@link ChangeType#VIEW} and {@link ChangeType#DOWNLOAD} but can be
	used for the other change types as well.
	
	@author nicklas
	@since 3.9
*/
public class ManualLogEntry 
	implements LoggableData
{
	
	private final ChangeType changeType;
	private final BasicItem item;
	private final String logInfo;
	private final String oldValue;
	private final String newValue;
	
	/**
		Create a new log entry. An item and change type is required.
		
		@param item The item the entry is related to
		@param changeType The type of log entry
		@param logInfo Other information about the log entry
	*/
	public ManualLogEntry(BasicItem item, ChangeType changeType, String logInfo)
	{
		this(item, changeType, logInfo, null, null);
	}
	
	/**
		Create a new log entry with optional old and new values.
		
		@param item The item the entry is related to
		@param changeType The type of log entry
		@param logInfo Other information about the log entry 
		@param oldValue Old value
		@param newValue New value
	*/
	public ManualLogEntry(BasicItem item, ChangeType changeType, String logInfo, String oldValue, String newValue)
	{
		if (item == null) throw new InvalidUseOfNullException("item");
		if (changeType == null) throw new InvalidUseOfNullException("changeType");
		this.item = item;
		this.changeType = changeType;
		this.logInfo = logInfo;
		this.oldValue = oldValue;
		this.newValue = newValue;
	}
	
	/**
		Get the change type.
	*/
	public ChangeType getChangeType()
	{
		return changeType;
	}
	
	/**
		Get the item this log entry is related to.
	*/
	public BasicItem getItem()
	{
		return item;
	}
	
	/**
		Get information about the log entry.
	*/
	public String getLogInfo()
	{
		return logInfo;
	}
	
	/**
		Old value before the change.
	*/
	public String getOldValue()
	{
		return oldValue;
	}
	
	/**
		New value after the change.
	*/
	public String getNewValue()
	{
		return newValue;
	}
	
}
