/**
	$Id $

	Copyright (C) 2016 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.io.IOException;
import java.io.InputStream;

import net.sf.basedb.core.AbsoluteProgressReporter;

/**
	Input stream wrapper that calls a progress reporter at regular
	intervals. The default interval is 1MB of processed data. The
	default progress message is simply to write the number of bytes
	processed. If a different message is required the {@link #progressReport(String)}
	method should be overridden.
	
	@author nicklas
	@since 3.9
*/
public class ProgressInputStream 
	extends InputStreamTracker
{

	protected final AbsoluteProgressReporter progress;
	protected final long interval;
	protected long nextProgress;
	
	/**
		Create a new input stream using the default 1MB interval.
	*/
	public ProgressInputStream(InputStream in, AbsoluteProgressReporter progress)
	{
		this(in, progress, Values.MB);
	}

	/**
		Create a new input stream using the specified interval.
	*/
	public ProgressInputStream(InputStream in, AbsoluteProgressReporter progress, long interval)
	{
		super(in);
		this.progress = progress;
		this.interval = interval;
		this.nextProgress = interval;
	}

	
	@Override
	public int read(byte[] b, int off, int len) 
		throws IOException 
	{
		int n = super.read(b, off, len);
		if (getNumRead() > nextProgress)
		{
			long bytes = getNumRead();
			progressReport(Values.formatBytes(bytes, bytes > Values.GB ? 1 : 0));
			nextProgress = getNumRead()+interval;
		}
		return n;
	}

	/**
		Send a progress report with the given message. This method is called
		automatically at the given intervals. The default message is the number
		of bytes read so far (formatted as a string with 
		{@link Values#formatBytes(Long, int)}). Subclasses may override this
		method to generate a different message.
	*/
	protected void progressReport(String message)
	{
		progress.displayAbsolute(getNumRead(), message);
	}
	
}
