/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.formatter;

import java.util.Date;
import java.util.List;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.Operator;
import net.sf.basedb.core.PropertyFilter;
import net.sf.basedb.core.Type;
import net.sf.basedb.util.Values;

/**
 	Formatter for PropertyFilter items. This will create a more readable representation
 	of the property filter than the {@link PropertyFilter#toString()} method. It knows
 	about some special properties and load the names of annotation types, item subtypes, etc
 	instead of just displaying their ID values.
 	
  	@author Nicklas
  	@since 3.5
 */
public class PropertyFilterFormatter 
	implements Formatter<PropertyFilter>
{

	private final DbControl dc;
	private final Formatter<Date> dateFormatter;
	
	/**
		Create a new formatter.
	*/
	public PropertyFilterFormatter(DbControl dc, Formatter<Date> dateFormatter)
	{
		this.dc = dc;
		this.dateFormatter = dateFormatter;
	}
	
	
	/*
		From the Formatter interface
		-------------------------------------------
	*/
	@Override
	public String format(PropertyFilter filter)
	{		
		String property = filter.getProperty();
		Operator operator = filter.getOperator();
		String value = filter.getValue();
		String quote = filter.getValueType().isNumerical() || value == null ? "" : "'";
		
		StringBuilder sb = new StringBuilder();
		if (property.startsWith("#"))
		{
			// Annotation type filter
			int startIndex = property.startsWith("##") ? 2 : 1;
			int annotationTypeId = Values.getInt(property.substring(startIndex));
			sb.append("{").append(nameOfItem(dc, Item.ANNOTATIONTYPE, annotationTypeId, false)).append("}");
			sb.append(" ").append(operator).append(" ");
			if (operator.isListOperator() && value != null)
			{
				appendListOfValues(sb, value, quote);
			}
			else
			{
				sb.append(quote).append(value).append(quote);
			}
		}
		else if (property.startsWith("¤"))
		{
			// Data file type filter
			int i = property.indexOf('(');
			int fileTypeId = Values.getInt(property.substring(1, i));
			String fileFilter = property.substring(i+1, property.length()-1);
			sb.append("{").append(nameOfItem(dc, Item.DATAFILETYPE, fileTypeId, false)).append("#").append(fileFilter).append("}");
			sb.append(" ").append(operator).append(" ");
			if (operator.isListOperator() && value != null)
			{
				appendListOfValues(sb, value, quote);
			}
			else
			{
				sb.append(quote).append(value).append(quote);
			}
		}
		else if (property.startsWith("§"))
		{
			// Item list filter
			sb.append(operator == Operator.EQ ? "memberOfList(" : "NOT memberOfList(");
			sb.append(nameOfItem(dc, Item.ITEMLIST, Values.getInt(value), true));
			sb.append(")");
		}
		else if (property.equals("bioWell.row") || property.equals("bioWell.column"))
		{
			WellCoordinateFormatter formatter = new WellCoordinateFormatter(property.equals("bioWell.row"));
			sb.append(property).append(" ").append(operator).append(" ");
			List<Object> selected = filter.getValuesAsObjects();
			if (selected.size() > 1) sb.append("(");
			if (selected.size() == 0) sb.append("null");
			String sep = "";
			for (Object index : selected)
			{
				sb.append(sep);
				sep = ", ";
				sb.append(formatter.format((Integer)index));
			}
			if (selected.size() > 1) sb.append(")");
		}
		else if (property.equals("parentType"))
		{
			sb.append(property).append(" ").append(operator).append(" ");
			List<Object> selected = filter.getValuesAsObjects();
			if (selected.size() > 1) sb.append("(");
			if (selected.size() == 0) sb.append("null");
			String sep = "";
			for (Object index : selected)
			{
				sb.append(sep);
				sep = ", ";
				sb.append(Item.fromValue((Integer)index).toString());
			}
			if (selected.size() > 1) sb.append(")");
		}
		else
		{
			Item nameOfItemType = null;
			if (property.equals("itemSubtype") || property.endsWith(".itemSubtype"))
			{
				// Special handling to display name of subtype
				nameOfItemType = Item.ITEMSUBTYPE;
			}
			else if (property.equals("platform"))
			{
				// Special handling to display name of platform
				nameOfItemType = Item.PLATFORM;
			}
			if (nameOfItemType != null)
			{
				sb.append(property).append(" ").append(operator).append(" ");
				List<Object> itemIds = filter.getValuesAsObjects();
				if (itemIds.size() > 1) sb.append("(");
				if (itemIds.size() == 0) sb.append("null");
				String sep = "";
				for (Object id : itemIds)
				{
					sb.append(sep);
					sep = ", ";
					sb.append(nameOfItem(dc, nameOfItemType, (Integer)id, true));
				}
				if (itemIds.size() > 1) sb.append(")");
			}
			else
			{
				if (property.startsWith("&") || property.startsWith("@") || property.startsWith("!"))
				{
					property = property.substring(1);
				}
				sb.append(property).append(" ").append(operator).append(" ");
				if (operator.isListOperator() && value != null)
				{
					appendListOfValues(sb, value, quote);
				}
				else
				{
					if (filter.getValueType().isTemporal())
					{
						value = dateFormatter.format((Date)Type.DATE.parseString(value));
					}
					sb.append(quote).append(value).append(quote);
				}
			}
		}
		return sb.toString();
	}

	/**
		@throws UnsupportedOperationException Always
	*/
	@Override
	public PropertyFilter parseString(String value)
	{
		throw new UnsupportedOperationException("parseString: " + value);
	}
	// -------------------------------------------

	
	private String nameOfItem(DbControl dc, Item itemType, int id, boolean quote)
	{
		String q = quote ? "'" : "";
		try
		{
			return q+((Nameable)itemType.getById(dc, id)).getName()+q;
		}
		catch (RuntimeException ex)
		{
			return Integer.toString(id);
		}
	}
	
	private void appendListOfValues(StringBuilder sb, String value, String quote)
	{
		String[] values = value.split("\\|");
		if (values.length > 1) sb.append("(");
		String sep = "";
		for (Object v : values)
		{
			sb.append(sep);
			sep = ", ";
			sb.append(quote).append(v).append(quote);
		}
		if (values.length > 1) sb.append(")");
		
	}
}
