/**
	$Id$

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileServer;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Location;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

/**
	Plug-in for importing file items in a batch. The plug-in can create new 
	items (offline or external) and updated existing items (all).

	@author nicklas
	@since 3.3
*/
public class FileImporter
	extends AbstractItemImporter<File>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.FILE, GuiContext.Type.LIST));

	protected static final PluginParameter<String> pathColumnMapping = new PluginParameter<String>(
		"pathColumnMapping",
		"Path",
		"Mapping that picks the path of the file from the data columns. Paths " +
			"can absolute or relative to the current directory. " +
			"Absolute paths must start with a forward slash (/). Example: \\Path\\",
		optionalColumnMapping
		);

	protected static final PluginParameter<String> fileServerColumnMapping = new PluginParameter<String>(
		"fileServerColumnMapping",
		"File server",
		"Mapping that picks the file server for external files from the data columns. " +
			"Example: \\File server\\",
		optionalColumnMapping
		);
	
	protected static final PluginParameter<String> urlColumnMapping = new PluginParameter<String>(
		"urlColumnMapping",
		"URL",
		"Mapping that picks the URL for external files from the data columns. " +
			"Files with no URL are registered as OFFLINE files. Example: \\URL\\",
		optionalColumnMapping
		);
	
	protected static final PluginParameter<String> mimeTypeColumnMapping = new PluginParameter<String>(
		"mimeTypeColumnMapping",
		"Mime type",
		"Mapping that picks the MIME type for the file from the data columns. Example: \\Mime type\\",
		optionalColumnMapping
		);

	protected static final PluginParameter<String> characterSetColumnMapping = new PluginParameter<String>(
		"characterSetColumnMapping",
		"Character set",
		"Mapping that picks the character set (for text files) for the file from the data columns. Example: \\Character set\\",
		optionalColumnMapping
		);
	
	private String rootPath;
	private PathIdMethod pathIdMethod;
	private Map<String, Directory> pathToDirectory;
	
	public FileImporter()
	{
		this.pathIdMethod = new PathIdMethod();
		this.pathToDirectory = new HashMap<String, Directory>();
	}

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper pathMapper;
	private Mapper fileServerMapper;
	private Mapper urlMapper;
	private Mapper mimeTypeMapper;
	private Mapper characterSetMapper;
	private Mapper descriptionMapper;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PathIdMethod}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { pathIdMethod, InternalIdMethod.INTERNAL_ID};
	}
	
	@Override
	protected Item getItemForSubtypes()
	{
		return Item.FILE;
	}
	
	@Override
	protected void addMoreJobParameters(List<PluginParameter<?>> parameters) 
	{
		// Root directory
		parameters.add(new PluginParameter<Directory>(
			"root",
			"Root directory",
			"The current directory is used for relative file paths.",
			new ItemParameterType<Directory>(Directory.class, null, true, 1, null))
		);
	}

	/**
		Adds column mappings for name, externalId and description.
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(pathColumnMapping);
		parameters.add(subtypeColumnMapping);
		parameters.add(urlColumnMapping);
		parameters.add(fileServerColumnMapping);
		parameters.add(mimeTypeColumnMapping);
		parameters.add(characterSetColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(permissionTemplateColumnMapping);
	}
	
	/**
		Calls {@link File#getQuery()}.
	*/
	@Override
	protected ItemQuery<File> createItemQuery()
	{
		return File.getQuery();
	}
	
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		pathMapper = getMapper(ffp, (String)job.getValue("pathColumnMapping"), null, null);
		urlMapper = getMapper(ffp, (String)job.getValue("urlColumnMapping"), 
				cropStrings ? File.MAX_URL_LENGTH : null, null);
		fileServerMapper = getMapper(ffp, (String)job.getValue("fileServerColumnMapping"), null, null);
		mimeTypeMapper = getMapper(ffp, (String)job.getValue("mimeTypeColumnMapping"), 
				cropStrings ? File.MAX_MIMETYPE_LENGTH : null, null);
		characterSetMapper = getMapper(ffp, (String)job.getValue("characterSetColumnMapping"), 
				cropStrings ? File.MAX_CHARSET_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? File.MAX_DESCRIPTION_LENGTH : null, null);
		
	}
	
	@Override
	protected File createItem(DbControl dc, FlatFileParser.Data data)
	{
		String path = pathMapper.getValue(data);
		Path filePath = new Path(resolvePath(path, getRootPath(dc)), Path.Type.FILE);
		Directory d = getDirectory(dc, filePath);
		File file = File.getNew(dc, d);
		file.setName(filePath.getFilename());
		updateItem(dc, file, data);
		return file;
	}
	
	@Override
	protected void updateItem(DbControl dc, File file, FlatFileParser.Data data)
	{
		if (pathMapper != null && file.isInDatabase()) 
		{
			String path = pathMapper.getValue(data);
			Path filePath = new Path(resolvePath(path, getRootPath(dc)), Path.Type.FILE);
			file.setDirectory(getDirectory(dc, filePath));
			file.setName(filePath.getFilename());
		}
		if (fileServerMapper != null) 
		{
			String nameOrId = fileServerMapper.getValue(data);
			FileServer server = findFileServer(dc, FallbackIdMethod.NAME_OR_ID, nameOrId);
			if (nameOrId == null || server != null) file.setFileServer(server);
		}
		if (urlMapper != null)
		{
			String url = urlMapper.getValue(data);
			if (url == null)
			{
				file.setLocation(Location.OFFLINE);
			}
			else
			{
				file.setUrl(url, false);
				try
				{
					file.updateURLMetadata();
				}
				catch (Exception ex)
				{}
			}
		}
		if (mimeTypeMapper != null) file.setMimeType(mimeTypeMapper.getValue(data));
		if (file.getMimeType() == null) file.setMimeTypeAuto(null, null);
		if (characterSetMapper != null) file.setMimeType(characterSetMapper.getValue(data));
		if (descriptionMapper != null) file.setDescription(descriptionMapper.getValue(data));
		updateItemSubtype(dc, file, data);
	}
	// ------------------------------------
	
	
	private String getRootPath(DbControl dc)
	{
		if (rootPath == null)
		{
			Directory d = Directory.getById(dc, ((Directory)job.getValue("root")).getId());
			rootPath = d.getPath().toString();
			pathIdMethod.rootPath = rootPath;
		}
		return rootPath;
	}
	
	private Directory getDirectory(DbControl dc, Path path)
	{
		String p = path.getType() == Path.Type.DIRECTORY ? path.toString() : path.getParent().toString();
		Directory d = pathToDirectory.get(p);
		if (d == null)
		{
			d = Directory.getNew(dc, path);
			pathToDirectory.put(p, d);
		}
		return d;
	}
	
	private static String resolvePath(String path, String rootPath)
	{
		if (path.startsWith("/"))
		{
			return path;
		}
		else
		{
			return rootPath + "/" + path;
		}		
	}
	
	public class PathIdMethod
		implements IdMethod
	{

		private String rootPath;
		
		@Override
		public String getMethod() 
		{
			return "path";
		}

		@Override
		public String getTitle() 
		{
			return "Path";
		}

		@Override
		public boolean isTrulyUnique() 
		{
			return true;
		}

		@Override
		public String getColumnMappingParameterName() 
		{
			return pathColumnMapping.getName();
		}

		@Override
		public <I extends BasicItem> ItemQuery<I> prepareQuery(DbControl dc, ItemQuery<I> query) 
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("name"),
					Expressions.parameter("name")
			));
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("directory"), 
					Expressions.parameter("directory")
			));
			return query;

		}

		@Override
		public <I extends BasicItem> List<I> find(DbControl dc, ItemQuery<I> query, String identifier)
		{
			Path filePath = new Path(resolvePath(identifier, getRootPath(dc)), Path.Type.FILE);
			try
			{
				int directoryId = Directory.getIdFromPath(dc, filePath);				
				query.setParameter("name", filePath.getFilename(), Type.STRING);
				query.setParameter("directory", directoryId, Type.INT);
				return query.list(dc);
			}
			catch (ItemNotFoundException ex)
			{
				return Collections.emptyList();
			}
		}
		
	}
}
