import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.ItemProxy;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Kit;
import net.sf.basedb.core.Trashcan;
import net.sf.basedb.core.ItemResultList;

/*
 $Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

 BASE is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 3
 of the License, or (at your option) any later version.

 BASE is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with BASE. If not, see <http://www.gnu.org/licenses/>.
 */

public class TestKit
{
	static boolean ok = true;

	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing kits");
		write_header();
		// Standard tests: create, load, list
		int id1 = test_create(0, null, true);
		int id2 = test_create(0, null, false);
		test_load(id1);
		test_list(-1);

		int extractId = TestExtract.test_create(0, "Extract with kit", 0, 0, false);
		test_use_kit(id1, extractId);
		
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();

		TestExtract.test_delete(extractId);
		
		// Standard test: Delete
		test_delete(id1);
		test_delete(id2);
		write("++Testing kits "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}
	
	static int test_create(int kitTypeId, String name, boolean setAll)
	{
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Kit k = Kit.getNew(dc);
			if (kitTypeId > 0)
			{
				k.setItemSubtype(ItemSubtype.getById(dc, kitTypeId));
			}
			if(setAll)
			{
				k.setName("Test kit");
				k.setDescription("Added at "+new Date());
				k.setInactive(true);
				k.setExpirationDate(new Date(System.currentTimeMillis()+48*3600*1000)); // Expire in two days
			}
			if (name != null) 
			{
				k.setName(name);
			}
			dc.saveItem(k);
			dc.commit();
			id = k.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(k, false);
			write_item(0, k);
			write("--Create kit OK");
		}
		catch (Throwable ex)
		{
			write("--Create kit FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Kit k = Kit.getById(dc, id);
			dc.deleteItem(k);
			Set<ItemProxy> using = k.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(k + " is used by " + using);
			}
			dc.commit();
			write("--Delete kit OK");
		}
		catch (Throwable ex)
		{
			write("--Delete kit FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Kit k = Kit.getById(dc, id);
			write_item(0, k);
			write("--Load kit OK");
		}
		catch (Throwable ex)
		{
			write("--Load kit FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static boolean test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Kit> l = Kit.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >=0 && expectedResults != l.size())
			{
				throw new BaseException("Expected " + expectedResults + " results, " +
						"not " + l.size());
			}
			write("--List kits OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List kits FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return ok;
	}
	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tType\tInactive\tExpiration date");
			write("-- \t-- \t--------- \t-----------\t----\t--------\t---------------");
		}
	}
	static void write_item(int i, Kit k)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
			write(i+":\t"+k.getId()+"\t"+k.getName()+"\t"+k.getDescription()+"\t"+k.getItemSubtype()+
				"\t"+k.isInactive()+"\t"+k.getExpirationDate()
				);
	}
	static void write_item(int i, Extract e)
			throws BaseException
		{
			if (!TestUtil.getSilent()) 
				write(i+":\t"+e.getId()+"\t"+e.getName()+"\t"+e.getCreationEvent().getKit());
		}
	static void write(String message)
	{
		System.out.println(message);
	}
		
	static void test_use_kit(int kitId, int extractId)
	{
		if (extractId == 0 || kitId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Kit k = Kit.getById(dc, kitId);
			Extract e = Extract.getById(dc, extractId);
			e.getCreationEvent().setKit(k);
			dc.commit();
			dc = TestUtil.getDbControl();
			e = Extract.getById(dc, extractId);
			write_item(0, e);
			write("--Use kit OK");
		}
		catch (Throwable ex)
		{
			write("--Use kit FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_delete_all()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Kit> l = Kit.getQuery().list(dc);
			int numDeleted = Trashcan.delete(dc.getSessionControl(), l, true, null);
			if (numDeleted != l.size()) throw new BaseException("Could not delete all items: " + numDeleted + " of " + l.size());
			dc.commit();
			write("--Delete all kits OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--Delete all kits FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

}
