/* $Id $
	------------------------------------------------------------------
	Copyright (C) 2012 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
	------------------------------------------------------------------

	@author Nicklas
	@since 3.3
*/
'use strict';

var App = function()
{
	var topWin = null;
	var app = {};
	var internal = {};

	var root;
	var sessionId;
	var storage;
	var rememberDialogPositions;

	/**
		Get the root URL path of this BASE installation. Always ends with a '/' (eg. /base/)
	*/
	app.getRoot = function()
	{
		return root;
	}
	
	/**
		Get the session-id of the current session or an empty string if no session has been started yet
		(should only happen on the top frameset of the login page)
	*/
	app.getSessionId = function()
	{
		return sessionId;
	}

	/**
		Get the id of the currently logged in user. Return null if the user has not 
		logged in yet.
	*/
	app.getLoggedInUserId = function()
	{
		return Data.int(document.body, 'user-id');
	}

	/**
		Get the id of the currently active project. Return null if no project is active.
	*/
	app.getActiveProjectId = function()
	{
		return Data.int(document.body, 'project-id');
	}

	/**
		Get the scaling factor to use in the gui for popup windows (and other things that may
		require scaling to fit).
	*/
	app.getScale = function()
	{
		return Data.float(document.body, 'gui-scale');
	}
	
	/**
		Get the max length of an URL that is allowed in a GET request. URLs that
		are longer than this should be rewritten to a POST request instead. A
		value of '0' disable the rewriting.
	*/
	app.getMaxUrlLength = function()
	{
		return Data.int(document.body, 'max-url-length', 0);
	}
	
	/**
		Check if popup dialog positions should be remembered or not.
	 */
	app.rememberDialogPositions = function()
	{
		return rememberDialogPositions;
	}
	
	/*
		Check if the URL is longer than the configured max URL length.
	*/
	app.isTooLongUrl = function(url)
	{
		var max = app.getMaxUrlLength();
		return max > 0 && url.length > max;
	}
	
	/*
		Safely set the location of a document as if
		win.document.location.href = url or win.document.replace(url)
		had been used. If the url is longer than specified by max 
		length the url will be re-written to a <form> using hidden 
		fields to submit the query values.
		@param url The url to set
		@param win The window to set the location on, if not given,
			the current window is assumed
		@param replace If given, use location.replace() instead of 
			location.href (only if url is not too long)
	*/
	app.safeSetLocation = function(url, win, replace)
	{
		if (!url) return;
		if (!win) win = window;
		var postform;
		if (app.isTooLongUrl(url))
		{
			postform = Forms.convertUrlToForm(url, win);
		}
		if (postform)
		{
			document.body.appendChild(postform);
			postform.submit();
			document.body.removeChild(postform);
		}
		else if (replace)
		{
			win.document.location.replace(url);
		}
		else
		{
			win.document.location.href = url;
		}
	}
	
	/**
		Get the top window in the BASE application. 
	*/
	app.topWindow = function()
	{
		if (!topWin)
		{
			var topWin = window.top;
			while (topWin.opener)
			{
				topWin = topWin.opener.top;
			}
		}
		return topWin;
	}

	/**
		Close the current popup window.
	*/
	app.closeWindow = function(event)
	{
		window.top.close();
	}
	
	/**
		Reload the current window.
	*/
	app.reloadWindow = function()
	{
		window.location.reload();
	}
	
	/**
		Get the screen coordinates, width and height of
		the given window. The result is returned as an
		object with 'top', 'left', 'width' and 'height'
		properties.
		@param win The window to get the position for or
			null to use the current window
	*/
	app.getWindowPosition = function(win)
	{
		if (!win) win = window;
		var pos = {};
		pos.top = win.screenY;
		pos.left = win.screenX;
		pos.width = win.innerWidth;
		pos.height = win.innerHeight;
		return pos;
	}
		
	/**
		Get a reference to the localStorage implementation.
		Return null if no storage is available.
	*/
	app.localStorage = function()
	{
		return storage;
	}
	
	app.setLocal = function(key, value)
	{
		if (storage == null) return;
		storage.setItem(root+':'+key, value);
	}
	
	app.getLocal = function(key, defaultValue)
	{
		if (storage == null) return defaultValue || null;
		return storage.getItem(root+':'+key) || defaultValue || null;
	}
	
	app.removeLocal = function(key)
	{
		if (storage == null) return;
		storage.removeItem(root+':'+key);
	}
	
	/**
		Write a message to the browser debug console.
	*/
	app.debug = function(message)
	{
		console.log(message);
	}
	
	/**
		Log usage of a deprecated method. A call to this method
		should be included as the first line in a deprecated method.
		@param method The name of the method (eg. Main.openPopup())
		@param voidVersion The version number of BASE in which the method will be removed
	 */
	app.deprecatedMethod = function(method, voidVersion)
	{
		var msg = 'Deprecated method "' + method + '" will be removed in BASE ' + voidVersion;
		console.error(msg);
	}
	
	internal.initApp = function()
	{
		if (document.body == null) return;
		
		root = Data.get(document.body, 'app-root');
		sessionId = Data.get(document.body, 'session-id', '');

		try
		{
			storage = window.localStorage;
		}
		catch (exception)
		{
			// Firefox throws a SecurityException if the 'ask every time' option is
			// selected for cookies, so we have to catch this situation
		}
		
		rememberDialogPositions = Data.int(document.body, 'remember-dialog-positions', 1) && storage != null;
	}
	document.addEventListener('DOMContentLoaded', internal.initApp, false);
	
	return app;
}();


var Doc = function()
{
	var elementInitializers = [];
	var onLoadFunctions = [];
	var finalizers = [];
	
	var doc = {};
	var internal = {};
	
	/**
		Get the page id of the current page if it has been defined.
		Eg. the id attribute of the root <html> tag.
	*/
	doc.getPageId = function()
	{
		return document.documentElement.id;
	}
	
	/**
		Get a document element reference. If the parameter is already a document node it
		is returned, otherwise try to find the given element by id.
		@param elementOrId A document element or the id of an element
	*/
	doc.element = function(elementOrId)
	{
		var e = elementOrId && (elementOrId.nodeType || elementOrId.window) ? elementOrId : document.getElementById(elementOrId);
		return e;
	}
	
	/**
		Get a document form reference. If the parameter is already a document node it
		is returned, otherwise try to find the given element by id or name.
		@param formNameOrId A form element or the id/name of a form element
	*/
	doc.form = function(formNameOrId)
	{
		var frm = doc.element(formNameOrId);
		if (!frm) frm = document.forms[formNameOrId];
		return frm || null;
	}
	
	/**
		Get the absolute coordinates of the given element.
		Returns an object with the following properties:
		left, top, width, height, right, bottom
	*/
	doc.getElementPosition = function(element)
	{
		element = Doc.element(element);

		var nextNode = element;
		var offsetTrail = element;
		var offsetTop = 0;
		var offsetLeft = 0;
		var offsetWidth = element.offsetWidth;
		var offsetHeight = element.offsetHeight;
		
		while (nextNode)
		{
			if (nextNode == offsetTrail)
			{
				if (offsetTrail.offsetTop) offsetTop += offsetTrail.offsetTop;
				if (offsetTrail.offsetLeft) offsetLeft += offsetTrail.offsetLeft;
				offsetTrail = offsetTrail.offsetParent;
			}
			if (nextNode != element && nextNode.scrollTop)
			{
				offsetTop -= nextNode.scrollTop;
			}
			nextNode = nextNode.parentNode;
		}
		
		var pos = {
			left: offsetLeft, 
			top: offsetTop, 
			width: offsetWidth, 
			height: offsetHeight, 
			right: offsetLeft+offsetWidth,
			bottom: offsetTop+offsetHeight
		};
		return pos;
	}
	
	/**
		Show an element in the document by setting a proper value for
		it's display property.
		@param element A document element or the id of an element
		@param display The value for the display property, if not set
			automatically select a value base on the node type
	*/
	doc.show = function(element, display)
	{
		element = Doc.element(element);
		if (element) 
		{
			if (!display)
			{
				if (element.nodeName == 'TR')
				{
					display='table-row';
				}
				else if (element.nodeName == 'TBODY')
				{
					display='table-row-group';
				}
				else if (element.nodeName == 'TD')
				{
					display = 'table-cell';
				}
				else 
				{
					display = 'block';
				}
			}
			element.style.display = display;
		}
	}

	
	/**
		Hides an element in the document by setting it's display property
		to 'none'.
		@param element A document element or the id of an element
	*/
	doc.hide = function(element)
	{
		element = Doc.element(element);
		if (element) element.style.display='none';
	}
	
	/**
		Show or hide an element in the document.
		@param element A document element or the id of an element
		@param showIfTrue If not set, automatically hide visible elements and show
			hidden elements, otherwise show/hide depending on the setting
	*/
	doc.showHide = function(element, showIfTrue)
	{
		element = Doc.element(element);
		if (element) 
		{
			if (showIfTrue == undefined) 
			{
				showIfTrue = element.style.display == 'none';
			}
			if (showIfTrue)
			{
				doc.show(element);
			}
			else
			{
				doc.hide(element);
			}
		}
	}

	/**
		Add a class to an element. 
		@param element A document element or the id of an element
		@param className The class to add
	*/
	doc.addClass = function(element, className)
	{
		element = Doc.element(element);
		if (element)
		{
			element.classList.add(className);
		}
	}
	
	/**
		Remove a class from an element. 
		@param element A document element or the id of an element
		@param className The class to remove
	*/
	doc.removeClass = function(element, className)
	{
		element = Doc.element(element);
		if (element)
		{
			element.classList.remove(className);
		}
	}
	
	/**
		Add or remove a class from an element depending on
		if it exists or not.
		@param element A document element or the id of an element
		@param className The class to add/remove
		@param addIfTrue If not set, automatically add/remove the class if
			depending on if it exists or not, otherwise add/remove depending 
			on the setting
	 */
	doc.addOrRemoveClass = function(element, className, addIfTrue)
	{
		element = Doc.element(element);
		if (element)
		{
			if (addIfTrue == undefined)
			{
				element.classList.toggle(className);
			}
			else if (addIfTrue)
			{
				element.classList.add(className);
			}
			else
			{
				element.classList.remove(className);
			}
		}
	}
		
	/**
		Check if the element is disabled (eg. it's className contains the class 'disabled').
		@param element A document element or the id of an element
	*/
	doc.isDisabled = function(element)
	{
		element = Doc.element(element);
		return element.classList.contains('disabled');
	}
	
	/**
		Add an element initialization callback to the document. Once the document has finished
		loading, the document is scanned for all elements marked with the class name
		'auto-init'. Then, for each found element, the callbacks are invoked so that
		they can perform their actions on it. Typically, the callbacks are used to add
		(click) event handlers to elements. 
		
		Note that all registered callbacks are invoked for all elements. Each callback should
		check that the 'auto-init' parameter has the expected value before proceeding.
		
		@param f The element initialization callback should be a function that accept two parameters.
			The first is a reference to the element, and the second the string value
			for the 'data-auto-init' attribute on the element (or null if no such attribute exists)
	*/
	doc.addElementInitializer = function(f)
	{
		if (!f || !f.call) return; // 'f' is not a function
		elementInitializers[elementInitializers.length] = f;
	}
	
	/**
		Register a function that should be called whenever the document DOM is
		ready. Eg. when the 'DOMContentLoaded' event is fired. The functions are
		called in the order they have been registered, but after all
		element initializers and before all finalizers.
	*/
	doc.onLoad = function(f)
	{
		if (!f || !f.call) return; // 'f' is not a function
		onLoadFunctions[onLoadFunctions.length] = f;
	}
	
	/**
		Register a function that should be called to finalize the document initialization. 
		Eg. when the 'DOMContentLoaded' event is fired. The functions are
		called in the order they have been registered, but after all
		element initializers and onload functions.
	*/
	doc.addFinalizer = function(f)
	{
		if (!f || !f.call) return; // 'f' is not a function
		finalizers[finalizers.length] = f;
	}
		
	/**
		Invoke all element initializers, onload functions and document finalizers
		in this order.
	*/
	internal.initDocument = function()
	{
		if (document.body == null) return;
		internal.blockFormSubmission();
		internal.invokeElementInitializers();
		internal.invokeOnLoadFunctions();
		internal.invokeFinalizers();
		Events.addEventHandler('page-reload', 'click', function() { location.reload(true) });
		var alertMsg = Data.get(document.body, 'alert-message');
		if (alertMsg) alert(alertMsg);
	}
	
	document.addEventListener('DOMContentLoaded', internal.initDocument, false);
	
	/**
		Initializer function that adds an event handler to all 'form' elements
		that block automatic submission (eg. by pressing ENTER) of the form.
		Forms should be submitted by calling form.submit().
	*/
	internal.blockFormSubmission = function()
	{
		var allForms = document.getElementsByTagName('form');
		for (var i = 0; i < allForms.length; i++)
		{
			allForms[i].addEventListener('submit', Events.preventDefault, false);
		}
	}
		
	/**
		Scan the document for all elements marked with the class name 'auto-init'.
		For each found element, call the registered element initialization callbacks.
	*/
	internal.invokeElementInitializers = function()
	{
		if (elementInitializers.length == 0) return;
		var autoElements = document.getElementsByClassName('auto-init');
		for (var i = 0; i < autoElements.length; i++)
		{
			var element = autoElements[i];
			var autoInit = Data.get(element, 'auto-init');
			for (var j = 0; j < elementInitializers.length; j++)
			{
				try
				{
					elementInitializers[j].call(element, element, autoInit);
				}
				catch (err)
				{
					var msg = err;
					if (err.toString)
					{
						msg = err.toString() + '\n' + err.stack;
					}
					console.error(msg);
				}
			}
		}
		elementInitializers = null;
	}

	/**
		Call all onload functions in the order they were registered.
	*/
	internal.invokeOnLoadFunctions = function()
	{
		for (var i = 0; i < onLoadFunctions.length; i++)
		{
			try
			{
				onLoadFunctions[i].call();
			}
			catch (err)
			{
				var msg = err;
				if (err.toString)
				{
					msg = err.toString() + '\n' + err.stack;
				}
				console.error(msg);
			}
		}
		onLoadFunctions = null;
	}
	
	/**
		Call all document finalizers in the order they were registered.
	*/
	internal.invokeFinalizers = function()
	{
		for (var i = 0; i < finalizers.length; i++)
		{
			try
			{
				finalizers[i].call();
			}
			catch (err)
			{
				var msg = err;
				if (err.toString)
				{
					msg = err.toString() + '\n' + err.stack;
				}
				console.error(msg);
			}
		}
		finalizers = null;
	}
	
	return doc;
}();


var Events = function()
{
	var events = {};
	
	/**
		Add event handler to the given element.
		@param element A document element or the id of an element
		@param eventName The name of the event to handle (eg. 'click')
		@param handler A function reference that is called for the event
		@param attributes Additional data attributes that should be attached to
			the target element (see Data.define)
	*/
	events.addEventHandler = function(element, eventName, handler, attributes)
	{
		element = Doc.element(element);
		if (!element) return;
		if (attributes) Data.define(element, attributes);
		element.addEventListener(eventName, handler, false);
	}
	
	/**
		Add an event handler to the given element that, when the element has the
		focus, reacts to the 'ENTER' key and sends a 'click' event to the same 
		element.
		@param element A document element or the id of an element
	*/
	events.enableClickOnEnter = function(element)
	{
		element = Doc.element(element);
		if (!element) return;
		element.addEventListener('keypress', Events.handleFocusedKeypress, false);
	}
	
	/**
		Add an event handler to the given element that, when the element
		has the focus, reacts to the 'ENTER' key and then call the given
		callback function. 
		@param element A document element or the id of an element
		@param callback A function callback
	*/
	events.doOnEnter = function(element, callback)
	{
		element = Doc.element(element);
		if (!element) return;
		if (!callback || !callback.call) return; // 'callback' is not a function
		
		element.onEnterCallback = callback;
		element.addEventListener('keypress', Events.handleFocusedKeypress, false);
	}
	
	/**
		Handle keypress events on focused elements:
		If the 'ENTER' key is pressed:
		If the element has a property 'onEnterCallback' that is callable, call that function
		with the event as the parameter, otherwise send a 'click' event to the target element
	*/
	events.handleFocusedKeypress = function(event)
	{
		if (event.keyCode == 13)
		{
			// ENTER key pressed
			var target = event.currentTarget;
			if (!target) return;

			if (target.onEnterCallback && target.onEnterCallback.call)
			{
				target.onEnterCallback.call(target, event);
			}
			else
			{
				events.sendClickEvent(target, event);
			}
		}
	}

	/**
		Checks if the target of the element has been disabled. It first try
		to check if the CSS property 'pointer-events' is set to 'none' for the
		target element. If this is not supported it checks the target element
		contains 'disabled' in it's class name.
		If the target element is considered disabled, the event.stopPropagation()
		function is called.
	*/
	events.stopIfDisabled = function(event)
	{
		var pointerEvents = null;
		var disabled = false;
		try
		{
			pointerEvents = window.getComputedStyle(event.currentTarget).getPropertyValue('pointer-events');
		}
		catch (e)
		{}
		if (pointerEvents == null)
		{
			disabled = event.currentTarget.classList.contains('disabled');
		}
		else
		{
			disabled = pointerEvents == 'none';
		}
		if (disabled) event.stopImmediatePropagation();
	}

	
	/**
		Event handler that can be attached to 'keypress' event for 
		form input fields to block all keys except numeric (including minus).
	*/
	events.integerOnly = function(event)
	{
		var charCode = event.charCode;
		// <31 for tab, delete, etc. 48-57=key 0-9, 45= '-'
		var allow = charCode <= 31 || (charCode >= 48 && charCode <= 57) || charCode == 45;
		if (!allow) event.preventDefault();
	}
	
	/*
		Event handler that can be attached to 'keypress' event for 
		form input fields to block all keys except numeric (including minus, plus, decimal dot and E).
	*/
	events.numberOnly = function(event)
	{
		var charCode = event.charCode;
		// <31 for tab, delete, etc. 48-57=key 0-9, 45= '-', 46='.', 69/101=E/e
		var allow = charCode <= 31 || (charCode >= 48 && charCode <= 57) || charCode == 45 || charCode == 46 || charCode == 69 || charCode == 101;
		if (!allow) event.preventDefault();
	}

	
	/**
		Event handler that can be attached to any event/element that
		simply calls event.preventDefault() to stop the default
		action.
	*/
	events.preventDefault = function(event)
	{
		event.preventDefault();
	}
	
	/**
		Send an 'onchange' event to the given element.
		@param element A document element or the id of an element
	*/
	events.sendChangeEvent = function(element)
	{
		element = Doc.element(element);
		var changeEvent = element.ownerDocument.createEvent('HTMLEvents');
		changeEvent.initEvent('change', true, true);
		element.dispatchEvent(changeEvent);
	}
	
	/**
		Send a 'click' event to the specifiec element, pretending that the
		mouse is located in the center of the element. A srcEvent may be given
		to set meta key status (CTRL, ALT, SHIFT).
	*/
	events.sendClickEvent = function(element, srcEvent)
	{
		element = Doc.element(element);

		var pos = Doc.getElementPosition(element);
		var wPos = App.getWindowPosition();
		var clientX = pos.left+pos.width/2;
		var clientY = pos.top+pos.height/2;
		
		if (!srcEvent) 
		{
			srcEvent = {
				'ctrlKey': false, 
				'altKey': false, 
				'shiftKey': false, 
				'metaKey': false
			};
		}
			
		var clickEvent = document.createEvent("MouseEvent");
		clickEvent.initMouseEvent("click", true, true, window, 1, wPos.left + clientX,  wPos.top + clientY, clientX, clientY, srcEvent.ctrlKey, srcEvent.altKey, srcEvent.shiftKey, srcEvent.metaKey, 0, null);
		element.dispatchEvent(clickEvent);
	}
	
	/**
		Send a custom event to the given element.
		@param element A document element or the id of an element
		@param eventType The name of the custom event
		@param details An object with detailed data to send with the event
	*/
	events.sendCustomEvent = function(element, eventType, details)
	{
		element = Doc.element(element);
		var customEvent = element.ownerDocument.createEvent('CustomEvent');
		customEvent.initCustomEvent(eventType, true, true, details);
		element.dispatchEvent(customEvent);
	}
	
	return events;

}();


/**
	Encapsulates data handling functions for handling data-* attributes on elements.
*/
var Data = function()
{
	var data = {};
	
	/**
		Get the value of a 'data-*' attribute for an element. If the attribute
		doesn't exists, the default value is returned.
		@param element A document element or the id of an element
		@param attribute The attribute name (not including the 'data-' prefix)
		@param An optional default value
	*/
	data.get = function(element, attribute, defaultValue)
	{
		element = Doc.element(element);
		var value = element.getAttribute('data-'+attribute) || defaultValue;
		return value === undefined ? null : value;
	}
	
	/**
		Set the value of a 'data-*' attribute for an element.
		Existing values are overwritten.
		@param element A document element or the id of an element
		@param attribute The attribute name (not including the 'data-' prefix)
		@param value The value to set
	*/
	data.set = function(element, attribute, value)
	{
		element = Doc.element(element);
		element.setAttribute('data-'+attribute, value);
	}
	
	/**
		Define missing 'data-*' attributes for an element. Attributes
		that already have a value are not overwritten. 
		@param element A document element or the id of an element
		@param attributes A JSON object with key-value pairs of attributes,
			the keys should not include the 'data-' prefix
	*/
	data.define = function(element, attributes)
	{
		element = Doc.element(element);
		for (var attr in attributes)
		{
			var name = 'data-'+attr;
			if (!element.hasAttribute(name))
			{
				element.setAttribute(name, attributes[attr]);
			}
		}
	}
	
	/**
		Get the value of a 'data-*' attribute as an integer. 
		If the attribute doesn't exists, the default value is used.
	*/
	data.int = function(element, attribute, defaultValue)
	{
		return parseInt(data.get(element, attribute, defaultValue), 10);
	}
	
	/**
		Get the value of a 'data-*' attribute as a float. 
		If the attribute doesn't exists, the default value is used.
	*/
	data.float = function(element, attribute, defaultValue)
	{
		return parseFloat(data.get(element, attribute, defaultValue));
	}
	
	/**
		Get the JSON-formatted value of a 'data-*' attribute for an element. 
		If the attribute doesn't exists, the default value is used.
	*/
	data.json = function(element, attribute, defaultValue)
	{
		var result;
		try
		{
			result = JSON.parse(data.get(element, attribute, defaultValue));
		}
		catch (ex)
		{
			result = {};
			console.error('Could not parse JSON attribute "' + attribute + '" on element <' + element.tagName.toLowerCase() +' id="'+element.id +'">: ' + ex);
		}
		return result;
	}
	
	/**
		Store a key/value pair in session-scoped storage (eg. the information is lost
		when the user closes the browser or navigates away from the BASE site). The
		information is kept by an array in the topmost window (see App.topWindow()).
		
		@param key The key which must be a string
		@param value The value, which can be any type of object
		@param namespace Optional, if not specified 'location.pathname' is used
	*/
	data.setPageValue = function(key, value, namespace)
	{
		var win = App.topWindow();
		if (!namespace) namespace = location.pathname;
		win.pageValues[namespace+'::'+key] = value;
	}
	
	/**
		Get value from session-scoped storage (eg. the information is lost
		when the user closes the browser or navigates away from the BASE site). The
		information is kept by an array in the topmost window (see App.topWindow()).
		If no value is found or if there is an error retrieving it, null is
		returned.
		
		@param key The key which must be a string
		@param namespace Optional, if not specified 'location.pathname' is used
	*/
	data.getPageValue = function(key, namespace)
	{
		var win = App.topWindow();
		if (!namespace) namespace = location.pathname;
		var value = null;
		try
		{
			value = win.pageValues[namespace+'::'+key] || null;
		}
		catch (e)
		{
			value = null;
		}
		return value;
	}
	
	return data;
}();

/**
	Functions that are specific to buttons in the gui (eg. in the toolbar or free-floating).
*/
var Buttons = function()
{
	var buttons = {};
	var internal = {};

	/**
		Add a click event handler to the button.
		If the button is focusable (eg. a tabindex>=0 exists) a
		'keypress' listener that forwards 'ENTER' to the click handler is 
		also added. The btn is either the id or a button <div> reference.
	*/
	buttons.addClickHandler = function(btn, handler, attributes)
	{
		var btn = Doc.element(btn);
		if (!btn) return;
		if (attributes) Data.define(btn, attributes);
		// Event handler for checking if the button has been disabled
		btn.addEventListener('click', Events.stopIfDisabled, true);
		btn.addEventListener('click', handler, false);
		if (btn.tabIndex >= 0)
		{
			Events.enableClickOnEnter(btn);
		}
	}

	/**
		Event handler that open the 'edit item' dialog for the item 
		that is given by the 'data-item-type' and 'data-item-id'
		attributes. Usually attached to an 'Edit' button on a
		single-item page.
	*/
	buttons.editItem = function(event)
	{
		var target = event.currentTarget;
		var itemType = Data.get(target, 'item-type');
		var itemId = Data.get(target, 'item-id');
		var extraUrl = Data.get(target, 'extra-url');
		Items.editItem(itemType, itemId, extraUrl);
	}
	
	/**
		Event handler that open the 'new item' dialog for the item 
		that is given by the 'data-item-type' attribute. 
		Usually attached to a 'New' button on a list page.
	*/
	buttons.newItem = function(event)
	{
		var target = event.currentTarget;
		var itemType = Data.get(target, 'item-type');
		var extraUrl = Data.get(target, 'extra-url');
		Items.newItem(itemType, extraUrl);
	}
	
	/**
		Event handler that submits a request for the item 
		that is given by the 'data-item-type' and 'data-item-id'
		attributes to be marked for removal. Usually attached to 
		an 'Remove' button on a single-item page.
	*/
	buttons.deleteItem = function(event)
	{
		var target = event.currentTarget;
		var itemType = Data.get(target, 'item-type');
		var itemId = Data.int(target, 'item-id');
		var extraUrl = Data.get(target, 'extra-url');
		var confirm = Data.int(target, 'confirm');
		Items.deleteItem(itemType, itemId, confirm, extraUrl);
	}

	/**
		Event handler that submits a request for all selected items
		in table to be marked for removal. Usually attached to 
		an 'Remove' button on a list page. The table id must be specifed
		in the 'data-table-id' attribute.
	*/
	buttons.deleteItems = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var regexp = Data.get(target, 'regexp');
		var confirm = Data.int(target, 'confirm');
		Table.deleteItems(tableId, regexp, confirm);
	}
	
	/**
		Event handler that submits a request for the item 
		that is given by the 'data-item-type' and 'data-item-id'
		attributes to be unmarked for removal. Usually attached to 
		an 'Restore' button on a single-item page.
	*/
	buttons.restoreItem = function(event)
	{
		var target = event.currentTarget;
		var itemType = Data.get(target, 'item-type');
		var itemId = Data.int(target, 'item-id');
		var extraUrl = Data.get(target, 'extra-url');
		Items.restoreItem(itemType, itemId, extraUrl);
	}
	
	/**
		Event handler that submits a request for all selected items
		in table to be unmarked for removal. Usually attached to 
		an 'Restore' button on a list page. The table id must be specifed
		in the 'data-table-id' attribute.
	*/
	buttons.restoreItems = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var regexp = Data.get(target, 'regexp');
		var confirm = Data.int(target, 'confirm', 0);
		Table.restoreItems(tableId, regexp, confirm);
	}

	/**
		Event handler that submits a request for the item 
		that is given by the 'data-item-type' and 'data-item-id'
		attributes to be deleted from the database. Usually attached 
		to an 'Trashcan' icon on a single-item page. Specify a target
		element id in 'data-notify' attribute to send a 'base-notify'
		message after the item has been deleted, otherwise the current
		page is automatically reloaded with the list page instead.
	*/
	buttons.deleteItemPermanently = function(event)
	{
		var target = event.currentTarget;
		var itemType = Data.get(target, 'item-type');
		var itemId = Data.int(target, 'item-id');
		var extraUrl = Data.get(target, 'extra-url');
		var notify = Data.get(target, 'notify');
		Items.deleteItemPermanently(itemType, itemId, true, notify, extraUrl);
	}
	
	/**
		Event handler that opens the 'Share item' dialog
		for the item that is given by the 'data-item-type' and 
		'data-item-id' attributes. Usually attached to a 'Shared' 
		icon on a list page or a button on a single-item page.
	*/
	buttons.shareItem = function(event)
	{
		var target = event.currentTarget;
		var itemType = Data.get(target, 'item-type');
		var itemId = Data.int(target, 'item-id');
		Dialogs.openShareItem(itemType, itemId);
	}
	
	/**
		Event handler that opens the 'Share items' dialog
		for all selected items in a table. Usually attached to 
		an 'Share' button on a list page. The table id must be specifed
		in the 'data-table-id' attribute.
	*/
	buttons.shareItems = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var regexp = Data.get(target, 'regexp');
		Table.shareItems(tableId, regexp);
	}
	
	/**
		Event handler that opens the 'Set owner' dialog
		for the item that is given by the 'data-item-type' and 
		'data-item-id' attributes. Usually attached to a button 
		on a single-item page.
	*/
	buttons.setOwner = function(event)
	{
		var target = event.currentTarget;
		var itemType = Data.get(target, 'item-type');
		var itemId = Data.int(target, 'item-id');
		Dialogs.openSetOwner(itemType, itemId);
	}

	/**
		Event handler that opens the 'Set owner of items' dialog
		for all selected items in a table. Usually attached to 
		an 'Set owner' button on a list page. The table id must be specifed
		in the 'data-table-id' attribute.
	*/
	buttons.setOwnerOfItems = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var regexp = Data.get(target, 'regexp');
		Table.setOwnerOfItems(tableId, regexp);
	}

	/**
		Event handler that opens the 'Batch inherit anntations' dialog
		for all selected items in a table. Usually attached to 
		an 'Inherit annotations' button on a list page. The table id must be specifed
		in the 'data-table-id' attribute.
	*/
	buttons.inheritAnnotations = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var regexp = Data.get(target, 'regexp');
		Table.inheritAnnotations(tableId, regexp);
	}

	
	/**
		Event handler that open the 'using items' (in the trashcan)
		page for the item that is given by the 'data-item-type' and 
		'data-item-id' attributes. Usually attached to an icon on a
		single-item page for items that have been marked for removal
		but can't be deleted.
	*/
	buttons.showUsingItems = function(event)
	{
		var target = event.currentTarget;
		var itemType = Data.get(target, 'item-type');
		var itemId = Data.int(target, 'item-id');
		Items.showUsingItems(itemType, itemId);
	}
	
	/**
		Event handler that opens the 'Run plug-in' dialog in
		a single-item context. The current item must be specified 
		by the 'data-item-type' and 'data-item-id' attributes. The
		plug-in type to select should be specified by the
		'data-plugin-type' attribute. Valid values are 'IMPORT'
		'EXPORT', 'ANALYZE' and 'OTHER'. Usually attached to a button
		on a single-item page.
	*/
	buttons.runPlugin = function(event)
	{
		var target = event.currentTarget;
		var itemType = Data.get(target, 'item-type');
		var itemId = Data.int(target, 'item-id');
		var pluginType = Data.get(target, 'plugin-type');
		var extraUrl = Data.get(target, 'extra-url');
		var cmd = Data.get(target, 'cmd');
		Items.runPlugin(itemType, itemId, pluginType, extraUrl, cmd);
	}

	/**
		Event handler that opens the 'Run plug-in' dialog in
		a list context. The list id must be specified by the
		'data-table-id' attribute. The plug-in type to select 
		should be specified by the 'data-plugin-type' attribute. 
		Valid values are 'IMPORT' 'EXPORT', 'ANALYZE' and 'OTHER'. 
		Usually attached to a button on a list page.
	*/
	buttons.runListPlugin = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var pluginType = Data.get(target, 'plugin-type');
		var cmd = Data.get(target, 'cmd');
		Table.runPlugin(tableId, pluginType, cmd);
	}
	
	/**
		Event handler that opens the 'Configure columns' dialog
		for a list page. The list id must be specified by the
		'data-table-id' attribute. Usually attached to a button
		on a list page.
	*/
	buttons.configureColumns = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var settings = Data.get(target, 'settings');
		Table.configureColumns(tableId, settings);
	}
	
	/**
		Event handler that return the selected items in a list page
		to the parent window. This can be done in one of two ways:
		Using a callback method defined by the parent window (old method)
		Using an event handler defined by the parent window (new method)
		In both cases, the list id must be specified by the 'data-table-id' 
		attribute and the name of the callback method or event target must be
		stored in the 'callback' form element of the table. The current window
		is automatically closed when all selected items have been returned
		to the parent.
	*/
	buttons.returnSelected = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var regexp = Data.get(target, 'regexp');
		
		Table.returnSelectedItems(tableId, regexp);
		App.closeWindow();
	}
	
	return buttons;
}();

/**
	Functions that open common popup dialog windows.
*/
var Dialogs = function()
{
	var dialogs = {};
	var internal = {};

	/*
		Opens a popup window with the specified name and size and loads 
		the given url.
	
		@param url The url to load in the window
		@param name The name of the window
		@param width The width in pixels of the window
		@param height The height in pixels of the window
	*/
	dialogs.openPopup = function(url, name, width, height)
	{
		// Get last position of the window -- can be null
		var position = internal.getLastDialogPosition(name);

		if (position == null)
		{
			// If no last position is know, use the specified settings
			position = {};
			// Rescale according to user settings -- but not larger than current screen
			var scale = App.getScale();
			position.width = Math.ceil(Math.min(width * scale, screen.availWidth));
			position.height = Math.ceil(Math.min(height * scale, screen.availHeight));
			
			//	Try to position the popup window in the center of the parent window.
			var pos = App.getWindowPosition(window.top);
			var top = Math.ceil(pos.top+(pos.height-position.height) / 2);
			var left = Math.ceil(pos.left+(pos.width-position.width) / 2);
			position.top = (top < 0 && pos.top >= 0) ? 0 : top;
			position.left = (left < 0 && pos.left >= 0) ? 0 : left;
		}

		var options = "toolbar=no,location=no,directories=no,status=no,menubar=no,scrollbars=yes,resizable=yes";
		options += ",top="+position.top+",left="+position.left+",width="+position.width+",height="+position.height;
		// Check if another window with the same name already exists
		var oldWin = window.open('', name, options);
		try
		{
			if (oldWin && !oldWin.closed && oldWin.Popup && !oldWin.Popup.isReadOnly())
			{
				var msg = 'This window has already been opened. ';
				msg += 'Opening it again causes unsaved changes to be lost. Continue?';
				if (!confirm(msg))
				{
					oldWin.focus();
					return;
				}
				url += '&warnIfOpen=0';
			}
		}
		catch (err)
		{}
		
		internal.saveRequestedDialogPosition(name, position);

		var newWin;
		if (App.isTooLongUrl(url))
		{
			newWin = window.open(App.getRoot()+'blank.jsp', name, options);
			App.safeSetLocation(url, newWin, true);
		}
		else
		{
			if (url == '') url = App.getRoot()+'blank.jsp';
			newWin = window.open(url, name, options);
		}
		
		newWin.focus();
		return newWin;
	}
	
	internal.getLastDialogPosition = function(name)
	{
		if (!App.rememberDialogPositions()) return null;
		return JSON.parse(App.getLocal('dialog-position:'+name));
	}
	
	internal.saveRequestedDialogPosition = function(name, position)
	{
		if (!App.rememberDialogPositions()) return;
		App.setLocal('dialog-position-req:'+name, JSON.stringify(position));
	}
	
	dialogs.forgetDialogPositions = function()
	{
		var storage = App.localStorage();
		if (!storage) return;
		
		var prefix = App.getRoot()+':dialog-position:';
		
		for (var i = storage.length-1; i >= 0; i--)
		{
			var key = storage.key(i);
			if (key.indexOf(prefix) == 0) storage.removeItem(key);
		}
	}
	
	/**
		Get a reference to the popup dialog with the given name.
		Return null if no window with the given name exists.
	 */
	dialogs.getDialog = function(name)
	{
		var options = 'toolbar=no,location=no,directories=no,status=no,menubar=no,scrollbars=yes,resizable=yes';
		options += ',top=0,left=0,width=100,height=100';
		var win = window.open('', name, options);
		if (!win.closed && win.App)
		{
			return win;
		}
		else
		{
			win.close();
			return null;
		}
	}
	
	/**
		Open a popup window showing help information.
		@param helpId The ID of the help information
	*/
	dialogs.openHelp = function(helpId)
	{
		var url = App.getRoot()+'common/help/view_help.jsp?ID='+App.getSessionId();
		url += '&external_id='+helpId;
		Dialogs.openPopup(url, 'Help', 640, 550);
	}
	
	/**
		Open a popup window showing help information. The help section to
		display is given by the currently active tab on the given tab control.
		@param tabControl A tab control element or the id
		@param helpId Optional, a default help id if no help is specified by the tab control
	*/
	dialogs.openTabControlHelp = function(tabControl, helpId)	
	{
		var useHelpId = TabControl.getActiveHelpId(tabControl);
		if (!useHelpId) useHelpId = helpId;
		dialogs.openHelp(useHelpId);
	}
	
	/*
		Open a popup window with a large textarea field for editing text.
		@param textarea A form input/textarea element or the id of an element
		@param title The title that the popup dialog should have
	*/
	dialogs.openZoom = function(textarea, title, useToolbar)
	{
		textarea = Doc.element(textarea);
		if (!title) title = textarea.id;
		var url = App.getRoot()+'common/zoom.jsp?ID='+App.getSessionId();
		url += '&title='+encodeURIComponent(title)+'&textarea='+textarea.id;
		if (useToolbar) url+= '&useToolbar=1';
		Dialogs.openPopup(url, title.replace(/[^\w]/g, ''), 640, 600);
	}
	
	/*
		Open a popup window for selecting a date or datetime value.
		@param textarea A form input/textarea element or the id of an element
		@param title The title that the popup dialog should have
		@param dateFormat The dateformat to use when reading/writing the value, or null to use the logged in user's default format
		@param useTime 'true' to use date+time, 'false' to only use date
	*/
	dialogs.openCalendar = function(textarea, title, dateFormat, useTime)
	{
		textarea = Doc.element(textarea);
		if (!title) title = textarea.id;
		var url = App.getRoot()+'common/calendar.jsp?ID='+App.getSessionId();
		url += '&title='+encodeURIComponent(title)+'&textarea='+textarea.id;
		if (dateFormat) url += '&format='+encodeURIComponent(dateFormat);
		if (useTime) url += '&useTime=1';
		Dialogs.openPopup(url, title.replace(/[^\w]/, ''), 450, 300);
	}
	
	/**
		Open a popup window for selecting a color. When a color has
		been selected the hex value is saved to the given textarea
		element and a 'change' event is sent to it before the
		popup is closed.
		
		@param textarea A form input/textarea element or the id of an
			element
		@param title The title that the popup dialog should have
	*/
	dialogs.openColorChart = function(textarea, title)
	{
		textarea = Doc.element(textarea);
		if (!title) title = textarea.id;
		
		var url = App.getRoot()+'common/select_color.jsp?ID=' + App.getSessionId();
		url += '&title='+encodeURIComponent(title)+'&textarea='+textarea.id;
		Dialogs.openPopup(url, title.replace(/[^\w]/, ''), 450, 300);
	}
	
	dialogs.openExpressionBuilder = function(textarea, title, formulaType, rawDataType, channels, bioAssaySetId)
	{
		textarea = Doc.element(textarea);
		if (!title) title = textarea.id;

		var url = App.getRoot()+'common/expression_builder.jsp?ID='+App.getSessionId();
		url += '&title='+encodeURIComponent(title)+'&textarea='+textarea.id;
		url += '&formulatype='+formulaType;
		url += '&rawdatatype='+rawDataType;
		url += '&channels='+channels;
		url += '&restrictions=' + (formulaType == 'COLUMN_RESTRICTION' ? 1 : 0);
		if (bioAssaySetId) url += '&bioassayset_id='+bioAssaySetId;
		Dialogs.openPopup(url, title.replace(/[^\w]/, ''), 750, 500);
	}
	
	/**
		Open the 'Share item' dialog for a single item.
		@param itemType The type of the item
		@param itemId The id of the item
	*/
	dialogs.openShareItem = function(itemType, itemId)
	{
		var controller = Items.getController(itemType);
		var url = App.getRoot() + controller.url;
		url += '?ID='+App.getSessionId();
		url += '&cmd=ShareItem';
		url += '&item_id='+itemId;
		dialogs.openPopup(url, 'Share'+itemType, 600, 400);
	}

	/**
		Open the 'Set owner' dialog for a single item.
		@param itemType The type of the item
		@param itemId The id of the item
	*/
	dialogs.openSetOwner = function(itemType, itemId)
	{
		var controller = Items.getController(itemType);
		var url = App.getRoot() + controller.url;
		url += '?ID='+App.getSessionId();
		url += '&cmd=SetOwnerOfItem';
		url += '&item_id='+itemId;
		dialogs.openPopup(url, 'SetOwnerOfItem'+itemType, 450, 300);
	}

	/**
		Open a dialog for selecting one or more items of the given
		item type. 
		@param itemType The type of the item to select
		@param callback A document element or the id of an element that should receieve
			'base-selected' event for each selected item
		@param mode The selection mode (selectone, selectmultiple, selectfilter)
		@param extraUrl Extra parameters added to the URL that opens the popup
	*/
	dialogs.selectItem = function(itemType, callback, mode, extraUrl)
	{
		var width = 1050;
		var height = 700;
		var controller = Items.getController(itemType);
		var url = App.getRoot() + controller.url;
		url += '?ID='+App.getSessionId();
		if (itemType == 'FILE')
		{
			url += '&cmd=' + (mode ? 'SelectMultiple' : 'SelectOne');
		}
		else if (itemType == 'DIRECTORY')
		{
			if (!mode || !mode.indexOf || mode.indexOf('select') == -1)
			{
				mode = mode ? 'selectmultipledirectory' : 'selectonedirectory';
			}
			url += '&mode=' + mode;
			width = 350;
			height = 500;
		}
		else
		{
			if (!mode || !mode.indexOf || mode.indexOf('select') == -1)
			{
				mode = mode ? 'selectmultiple' : 'selectone';
			}
			url += '&cmd=UpdateContext';
			url += '&mode=' + mode;		
		}
		if (callback) 
		{
			callback = Doc.element(callback);
			if (!callback.id) callback.id = 'n'+(new Date()).getTime();
			url += '&callback='+callback.id;
		}
		if (extraUrl) url += extraUrl;
		
		dialogs.openPopup(url, 'Select'+itemType, width, height);
	}
	
	/**
		Help icon 'click' handler. The help id is found on the target element's
		'data-help-id' attribute or with the help of the active tab control on the
		'data-tabcontrol-id' attribute.
	*/
	internal.helpOnClick = function(event)
	{
		var hlp = event.currentTarget;
		var helpId = Data.get(hlp, 'help-id');
		var tabControlId = Data.get(hlp, 'tabcontrol-id');
		
		if (tabControlId)
		{
			dialogs.openTabControlHelp(tabControlId, helpId);
		}
		else if (helpId)
		{
			dialogs.openHelp(helpId);
		}
	}
	
	/**
		Zoom icon 'click' handler. The id of the textarea is found on the target
		element's 'data-textarea-id' attribute. The dialog title can be specified
		using the 'data-title' attribute, otherwise the textarea-id is used.
	 */
	internal.zoomOnClick = function(event)
	{
		var zoom = event.currentTarget;
		var textAreaId = Data.get(zoom, 'textarea-id');
		var dialogTitle = Data.get(zoom, 'title', textAreaId);
		var useToolbar = Data.int(zoom, 'use-toolbar', 0);
		dialogs.openZoom(textAreaId, dialogTitle, useToolbar);
	}
	
	/**
		Calendar icon 'click' handler. The id of the textarea is found on the target
		element's 'data-textarea-id' attribute. The dialog title can be specified
		using the 'data-title' attribute, otherwise the textarea-id is used.
		The date format to use must be specified by 'data-date-format' attribute and
		if time is included the 'data-use-time' must be set.
	 */
	internal.calendarOnClick = function(event)
	{
		var cal = event.currentTarget;
		var textAreaId = Data.get(cal, 'textarea-id');
		var dialogTitle = Data.get(cal, 'title', textAreaId);
		var dateFormat = Data.get(cal, 'date-format');
		var useTime = Data.get(cal, 'use-time', 0);
		dialogs.openCalendar(textAreaId, dialogTitle, dateFormat, useTime);
	}
	
	/**
		Event handler that sends a 'click' event to the element
		with id 'close' if the 'ESC' key is pressed. It is expected
		that the 'close' element actually closes the (popup) window,
		but it may perform other actions first.
	*/
	internal.closeOnEscape = function(event)
	{
		if (event.keyCode == 27)
		{
			Events.sendClickEvent('close', event);
		}
	}
	
	/**
		Initializer that add 'onclick' event handlers to common
		elements: 'help', 'zoom', 'calendar', etc.
	*/
	internal.addOnClickHandler = function(element, autoInit)
	{
		if (autoInit == 'help')
		{
			Buttons.addClickHandler(element, internal.helpOnClick);
		}
		else if (autoInit == 'zoom')
		{
			Buttons.addClickHandler(element, internal.zoomOnClick);
		}
		else if (autoInit == 'calendar')
		{
			Buttons.addClickHandler(element, internal.calendarOnClick);
		}
	}
	
	/**
		Initializer that add a keydown event handler to popup windows
		to detect 'ESC' which should close the window.
	*/
	internal.initPage = function()
	{
		if (window.opener && Doc.element('close'))
		{
			window.addEventListener('keydown', internal.closeOnEscape, false);
		}
	}
	
	Doc.addElementInitializer(internal.addOnClickHandler);
	Doc.onLoad(internal.initPage);
	
	return dialogs;
}();

var Items = function()
{
	var items = {};
	var internal = {};
	var controllers = null;
	
	/**
		Get controller information for a given item type.
	*/
	items.getController = function(itemType)
	{
		if (controllers == null) internal.initControllers();
		var controller = controllers[itemType];
		if (!controller)
		{
			var msg = 'Unhandled item: ' + itemType + '\n';
			msg += 'Please report this as a bug to the development team';
			alert(msg);
			return null;
		}
		return controller;
	}
	
	/**
		Get an array with all controllers.
	*/
	items.getAllControllers = function()
	{
		if (controllers == null) internal.initControllers();
		return controllers;
	}
	
	/**
		Go to the single-item view page for the given item.
		The page may open in a popup or in the main window.
	*/
	items.viewItem = function(itemType, itemId, extraUrl)
	{
		var controller = items.getController(itemType);
		if (controller.edit)
		{
			items.editItem(itemType, itemId, extraUrl);
			return;
		}
		var url = App.getRoot() + controller.url;
		url += '?ID='+App.getSessionId();
		url += '&cmd=ViewItem';
		if (itemId) url += '&item_id='+itemId;
		if (extraUrl) url += extraUrl;
		if (controller.popup)
		{
			Dialogs.openPopup(url, 'ViewItem'+itemType, controller.width, controller.height);
		}
		else
		{
			var theWindow = window;
			var i = 0;
			if (!controller.iframe)
			{
				while (theWindow.name != 'main' && i < 10)
				{
					theWindow = theWindow.parent;
					i++;
				}
			}
			theWindow.location.href = url;
		}
	}
	
	/**
		Open a dialog for creating a new item of the given type.
	*/
	items.newItem = function(itemType, extraUrl)
	{
		items.editItem(itemType, 0, extraUrl);
	}
	
	/**
		Open a dialog for editing the item of the given
		type and id.
	 */
	items.editItem = function(itemType, itemId, extraUrl)
	{
		var controller = items.getController(itemType);
		if (controller.edit == false)
		{
			items.viewItem(itemType, itemId, extraUrl);
			return;
		}
		var url = App.getRoot() + controller.url;
		url += '?ID='+App.getSessionId();
		var cmd = itemId ? 'EditItem' : 'NewItem';
		url += '&cmd='+cmd;
		if (itemId) url += '&item_id='+itemId;
		if (extraUrl) url += extraUrl;
		Dialogs.openPopup(url, 'Edit'+itemType, controller.width, controller.height);
	}
	
	/**
		Go to the list page for the given item type. Replaces the
		current page in the browser history.
	*/
	items.list = function(itemType, extraUrl)
	{
		internal.itemCommand('List', itemType, null, extraUrl);
	}
	
	/**
		Go to the list page for the given item type. Remember the
		last page in the browser history.
	 */
	items.listWithHistory = function(itemType, extraUrl)
	{
		internal.itemCommand('List', itemType, null, extraUrl, null, true);
	}
	
	/**
		Submit a request to mark the given item for removal.
		The current page is automatically reloaded.
	*/
	items.deleteItem = function(itemType, itemId, ask, extraUrl)
	{
		if (ask)
		{
			if (!confirm('Delete this item?')) return;
		}
		internal.itemCommand('DeleteItem', itemType, itemId, extraUrl);
	}
	
	/**
		Submit a request to unmark the given item for removal.
		The current page is automatically reloaded.
	*/
	items.restoreItem = function(itemType, itemId, extraUrl)
	{
		internal.itemCommand('RestoreItem', itemType, itemId, extraUrl);
	}
	
	/**
		Submit a request to delete the given item from the database.
		The current page is automatically changed to the list page.
	*/
	items.deleteItemPermanently = function(itemType, itemId, ask, notify, extraUrl)
	{
		if (ask)
		{
			if (!confirm('Delete this item permanently?')) return;
		}
		
		if (!notify)
		{
			// Create temporary div to receive redirect
			// notification from popup window
			var notifyDiv = document.createElement('div');
			notifyDiv.id = 'n'+(new Date()).getTime();
			Events.addEventHandler(notifyDiv, 'base-notify', 
				function() 
				{
					Items.list(itemType, extraUrl); 
				}
			);
			document.body.appendChild(notifyDiv);
			notify = notifyDiv.id;
		}
		
		var url = App.getRoot();
		url += 'views/trashcan/index.jsp';
		url += '?ID='+App.getSessionId();
		url += '&cmd=DeleteItem&popup=1';
		url += '&item_type='+itemType+'&item_id=' + itemId;
		url += '&notify='+notify;
		Dialogs.openPopup(url, 'Delete'+itemType, 300, 200);
	}

	
	/**
		Go to the trashcan page for the given item and list other
		items that are using it.
	*/
	items.showUsingItems = function(itemType, itemId)
	{
		var url = App.getRoot();
		url += 'views/trashcan/index.jsp';
		url += '?ID='+App.getSessionId();
		url += '&cmd=ViewUsingItems';
		url += '&item_type='+itemType+'&item_id='+itemId;
		location.href = url;
	}
	
	/**
		Open a popup window for selecting a plug-in that can run in the 
		current item context.
		@param itemType The current item type
		@param itemId The id of the current item
		@param pluginType One of: IMPORT, EXPORT, ANALYZE or OTHER (default)
	*/
	items.runPlugin = function(itemType, itemId, pluginType, extraUrl, cmd)
	{
		if (!cmd)
		{
			if (pluginType == 'EXPORT')
			{
				cmd = 'ExportItem';
			}
			else if (pluginType == 'IMPORT')
			{
				cmd = 'ImportItem';
			}
			else if (pluginType == 'ANALYZE')
			{
				cmd = 'RunAnalysisPlugin';
			}
			else
			{
				cmd = 'RunPlugin';
			}
		}
		internal.itemCommand(cmd, itemType, itemId, extraUrl, {width: 750, height: 500});
	}
	
	/**
		Generic callback for handling 'base-selected' event that is sent
		to the selection list when a single item has been selected.
	 */
	items.onItemSelected = function(event)
	{
		var list = event.currentTarget;
		if (!list.options) list = Doc.element(list.id+'.list');
		
		// If 'none' is first option, insert at index=1, otherwise at index=0
		var insertIndex = list.length > 0 && list[0].value == '0' ? 1 : 0;
		
		// Create the new option
		var option = new Option(event.detail.name, event.detail.id);
		option.title = option.text;
		
		var changed = true;
		if (list.length <= insertIndex)
		{
			// Add the new option to the end
			list[list.length] = option;
		}
		else if (list[insertIndex].value == 'NaN')
		{
			// Add a new option before the current
			list.add(option, list[insertIndex]);
		}
		else
		{
			// Replace the current option
			changed = list[list.selectedIndex].value != option.value;
			list[insertIndex] = option;
		}
		list.selectedIndex = insertIndex;
		if (changed) Events.sendChangeEvent(list);
	}
	
	internal.itemCommand = function(cmd, itemType, itemId, extraUrl, popup, noReplace)
	{
		var controller = items.getController(itemType);
		var url = App.getRoot() + controller.url;
		url += '?ID='+App.getSessionId();
		url += '&cmd='+cmd;
		if (itemId) url += '&item_id='+itemId;
		if (extraUrl) url += extraUrl;
		if (popup == true) popup = controller;
		if (popup)
		{
			Dialogs.openPopup(url, cmd+itemType, popup.width, popup.height);
		}
		else if (noReplace)
		{
			location.href = url;
		}
		else
		{
			location.replace(url);
		}
	}
	
	/**
		Event handler for clicking on item links. The item type and id should be stored
		as data-item-type and data-item-id attributes on the target element. If data-no-edit
		is set to 1 the edit dialog is disabled. If data-force-edit is set the edit
		dialog is always used. This event handler is automatically attached
		to auto-init elements with data-auto-init='item-link'.
	*/
	items.itemOnClick = function(event)
	{
		var element = event.currentTarget;
		var itemType = Data.get(element, 'item-type');
		var itemId = Data.int(element, 'item-id');
		var noEdit = Data.int(element, 'no-edit', 0);
		var forceEdit = Data.int(element, 'force-edit', 0);
		var extraUrl = Data.get(element, 'extra-url');
		var specialKey = event.altKey || event.ctrlKey || event.shiftKey;
		
		var editDialog = forceEdit || (specialKey && !noEdit);
		if (editDialog)
		{
			items.editItem(itemType, itemId, extraUrl);
		}
		else
		{
			items.viewItem(itemType, itemId, extraUrl);
		}
	}

	/**
		Initialize all item-link elements by adding a click event
		handler to them.
	 */
	internal.initializeItemLinks = function(element, autoInit)
	{
		if (autoInit != 'item-link') return;
		Events.addEventHandler(element, 'click', Items.itemOnClick);
	}
	Doc.addElementInitializer(internal.initializeItemLinks);


	/**
		Controller JSP page for all items. Each entry contains:
		
		url: The JSP that is the controller for the item type
		width, height: Size of popup for edit dialog
		edit: If true, always open in edit mode, if false, never open in edit mode
		popup: If set, the view page is opened in a popup
		noAnyToAny: If true, it is not possible to link AnyToAny items
	 */
	internal.initControllers = function()
	{
		controllers = [];
		controllers['ANNOTATION'] = { url:'common/annotations/index.jsp', width:750, height:500, edit:true, popup:true, noAnyToAny:true };
		controllers['ANNOTATIONTYPE'] = { url:'admin/annotationtypes/index.jsp', width:600, height:400 };
		controllers['ANNOTATIONTYPECATEGORY'] = { url:'admin/annotationtypecategories/index.jsp', width:600, height:400 };
		controllers['ANYTOANY'] = { url:'common/anytoany/index.jsp', width:600, height:400, popup:true, noAnyToAny:true };
		controllers['ARRAYBATCH'] = { url:'lims/arraybatches/index.jsp', width:750, height:500 };
		controllers['ARRAYDESIGN'] = { url:'lims/arraydesigns/index.jsp', width:750, height:500 };
		controllers['ARRAYSLIDE'] = { url:'lims/arrayslides/index.jsp', width:750, height:500 };
		controllers['BIOASSAY'] = { url:'views/experiments/bioassays/index.jsp', width:600, height:400, noAnyToAny:true };
		controllers['BIOASSAYSET'] = { url:'views/experiments/bioassaysets/index.jsp', width:750, height:500, noAnyToAny:true };
		controllers['BIOMATERIALEVENT'] = { url:'biomaterials/events/index.jsp', width:600, height:400, noAnyToAny:true };
		controllers['BIOMATERIALLIST'] = { url:'biomaterials/lists/index.jsp', width:750, height:500 };
		controllers['BIOPLATE'] = { url:'biomaterials/bioplates/index.jsp', width:750, height:500 };
		controllers['BIOPLATEEVENT'] = { url:'biomaterials/bioplates/events/index.jsp', width:600, height:400, noAnyToAny:true };
		controllers['BIOPLATEEVENTTYPE'] = { url:'biomaterials/bioplateeventtypes/index.jsp', width:600, height:400 };
		controllers['BIOPLATETYPE'] = { url:'biomaterials/bioplatetypes/index.jsp', width:600, height:400 };
		controllers['BIOSOURCE'] = { url:'biomaterials/biosources/index.jsp', width:750, height:500 };
		controllers['BIOWELL'] = { url:'biomaterials/bioplates/wells/index.jsp', width:450, height:300, edit: true, noAnyToAny:true };
		controllers['CLIENT'] = { url:'admin/clients/index.jsp', width:450, height:300 };
		controllers['CHANGEHISTORY'] = { url:'common/history/index.jsp', width:600, height:400, edit:false, popup:true, noAnyToAny:true };
		controllers['DATAFILETYPE'] = { url:'admin/datafiletypes/index.jsp', width:600, height:400 };
		controllers['DERIVEDBIOASSAY'] = { url:'views/derivedbioassays/index.jsp', width:750, height:500 };
		controllers['DIRECTORY'] = { url:'filemanager/directories/index.jsp', width:600, height:400, popup:true };
		controllers['EXPERIMENT'] = { url:'views/experiments/index.jsp', width:750, height:500 };
		controllers['EXTRACT'] = { url:'biomaterials/extracts/index.jsp', width:750, height:500 };
		controllers['EXTRAVALUE'] = { url:'views/experiments/extravalues/index.jsp', width:600, height:400, edit:false };
		controllers['EXTRAVALUETYPE'] = { url:'admin/extravaluetypes/index.jsp', width:600, height:400 };
		controllers['FEATURE'] = { url:'lims/arraydesigns/features/index.jsp', width:750, height:500, popup:true, edit:false, noAnyToAny:true };
		controllers['FILE'] = { url:'filemanager/index.jsp', iframe:true, width:750, height:500 };
		controllers['FILESERVER'] = { url:'filemanager/fileservers/index.jsp', width:750, height:500 };
		controllers['FORMULA'] = { url:'views/formulas/index.jsp', width:900, height:600 };
		controllers['GROUP'] = { url:'admin/groups/index.jsp', width:600, height:400 };
		controllers['HARDWARE'] = { url:'admin/hardware/index.jsp', width:800, height:500 };
		controllers['HELP'] = { url:'admin/clients/help/index.jsp', width:600, height:400, noAnyToAny:true };
		controllers['ITEMLIST'] = { url:'views/itemlists/index.jsp', width:750, height:500 };
		controllers['SYNCFILTER'] = { url:'views/itemlists/syncfilter/index.jsp', width:750, height:500, popup:true, noAnyToAny:true };
		controllers['ITEMSUBTYPE'] = { url:'admin/itemsubtypes/index.jsp', width:690, height:460 };
		controllers['JOB'] = { url:'views/jobs/index.jsp', width:750, height:500, popup:true, edit:false };
		controllers['JOBAGENT'] = { url:'admin/jobagents/index.jsp', width:750, height:500 };
		controllers['KIT'] = { url:'biomaterials/kits/index.jsp', width:750, height:500 };
		controllers['MESSAGE'] = { url:'my_base/messages/index.jsp', width:600, height:400, popup:true, edit:false };
		controllers['MIMETYPE'] = { url:'admin/mimetypes/index.jsp', width:600, height:400 };
		controllers['NEWS'] = { url:'admin/news/index.jsp', width:600, height:400 };
		controllers['PERMISSIONTEMPLATE'] = { url:'views/permissiontemplates/index.jsp', width:450, height:300 };
		controllers['PHYSICALBIOASSAY'] = { url:'views/physicalbioassays/index.jsp', width:750, height:500 };
		controllers['PLATE'] = { url:'lims/plates/index.jsp', width:750, height:500 };
		controllers['PLATEEVENT'] = { url:'lims/plates/events/index.jsp', width:600, height:400, noAnyToAny:true };
		controllers['PLATEEVENTTYPE'] = { url:'lims/platetypes/eventtypes/index.jsp', width:450, height:300 };
		controllers['PLATEGEOMETRY'] = { url:'lims/geometries/index.jsp', width:450, height:300 };
		controllers['PLATEMAPPING'] = { url:'lims/platemappings/index.jsp', width:600, height:400 };
		controllers['PLATETYPE'] = { url:'lims/platetypes/index.jsp', width:450, height:300 };
		controllers['PLATFORM'] = { url:'admin/platforms/index.jsp', width:600, height:400 };
		controllers['PLATFORMVARIANT'] = { url:'admin/platforms/variants/index.jsp', width:600, height:400 };
		controllers['PLUGINCONFIGURATION'] = { url:'admin/pluginconfigurations/index.jsp', width:600, height:400 };
		controllers['PLUGINDEFINITION'] = { url:'admin/plugindefinitions/index.jsp', width:750, height:500 };
		controllers['PLUGINTYPE'] = { url:'admin/plugintypes/index.jsp', width:600, height:400 };
		controllers['PROJECT'] = { url:'my_base/projects/index.jsp', width:750, height:500 };
		controllers['PROTOCOL'] = { url:'admin/protocols/index.jsp', width:750, height:500 };
		controllers['QUANTITY'] = { url:'admin/quantities/index.jsp', width:450, height:300 };
		controllers['QUOTA'] = { url:'admin/quota/index.jsp', width:750, height:500 };
		controllers['QUOTATYPE'] = { url:'admin/quotatypes/index.jsp', width:450, height:300 };
		controllers['RAWBIOASSAY'] = { url:'views/rawbioassays/index.jsp', width:750, height:500 };
		controllers['ROOTRAWBIOASSAY'] = { url:'views/experiments/rootrawbioassays/index.jsp', width:750, height:500, noAnyToAny: true };
		controllers['RAWDATA'] = { url:'views/rawbioassays/rawdata/index.jsp', width:750, height:500, popup: true, edit:false, noAnyToAny:true };
		controllers['REPORTER'] = { url:'views/reporters/index.jsp', width:600, height:400, noAnyToAny:true };
		controllers['REPORTERCLONETEMPLATE'] = { url:'admin/reporterclonetemplates/index.jsp', width:750, height:500 };
		controllers['REPORTERLIST'] = { url:'views/reporterlists/index.jsp', width:600, height:400 };
		controllers['REPORTERSCORE'] = controllers['REPORTER'];
		controllers['REPORTERTYPE'] = { url:'admin/reportertypes/index.jsp', width:450, height:300 };
		controllers['ROLE'] = { url:'admin/roles/index.jsp', width:600, height:400 };
		controllers['SAMPLE'] = { url:'biomaterials/samples/index.jsp', width:750, height:500 };
		controllers['SESSION'] = { url:'views/sessions/index.jsp', width:450, height:300, popup:true, edit:false, noAnyToAny:true };
		controllers['SOFTWARE'] = { url:'admin/software/index.jsp', width:800, height:500 };
		controllers['TAG'] = { url:'biomaterials/tags/index.jsp', width:800, height:500 };
		controllers['TRANSFORMATION'] = { url:'views/experiments/transformations/index.jsp', width:450, height:300, noAnyToAny:true };
		controllers['UNIT'] = { url:'admin/quantities/units/index.jsp', width:600, height:400, noAnyToAny:true };
		controllers['USER'] = { url:'admin/users/index.jsp', width:750, height:500 };
		controllers['USERDEVICE'] = { url:'views/devices/index.jsp', width:450, height:300, noAnyToAny:true };
		controllers['WELL'] = { url:'lims/plates/wells/index.jsp', width:750, height:500, noAnyToAny:true };
	}

	
	return items;
}();

var Files = function()
{
	var files = {};
	var internal = {};

	/**
		Open a new browser window for viewing the file contents of
		the file with the given id. If no window name is specified a
		new window is always created.
	*/
	files.viewFile = function(fileId, windowName)
	{
		var controller = Items.getController('FILE');
		var url = App.getRoot() + controller.url;
		url += '?ID='+App.getSessionId();
		url += '&cmd=ViewFile&item_id='+fileId;
		if (!windowName) windowName = '_blank';
		window.open(url, windowName);
	}
	
	/**
		Open a popup window for downloading the file contents of
		the file with the given id. The download should start 
		automatically.
	*/
	files.downloadFile = function(fileId)
	{
		var controller = Items.getController('FILE');
		var url = App.getRoot() + controller.url;
		url += '?ID='+App.getSessionId();
		url += '&cmd=DownloadFile&item_id='+fileId;
		Dialogs.openPopup(url, 'DownloadFile', 450, 300);
	}
	
	/**
		Event handler that call 'viewFile' when clicking
		on the attached target. The file id should be stored
		in the 'data-file-id' attribute.
	*/
	files.viewFileOnClick = function(event)
	{
		files.viewFile(Data.get(event.currentTarget, 'file-id'));
	}
	
	/**
		Event handler that call 'downloadFile' when clicking
		on the attached target. The file id should be stored
		in the 'data-file-id' attribute.
	*/
	files.downloadFileOnClick = function(event)
	{
		files.downloadFile(Data.get(event.currentTarget, 'file-id'));
	}
	
	internal.initializeFileLinks = function(element, autoInit)
	{
		if (autoInit == 'view-file')
		{
			Events.addEventHandler(element, 'click', Files.viewFileOnClick);
		}
		else if (autoInit == 'download-file')
		{
			Events.addEventHandler(element, 'click', Files.downloadFileOnClick);
		}
	}
	Doc.addElementInitializer(internal.initializeFileLinks);
	
	return files;
}();

var Forms = function()
{
	var forms = {};
	var internal = {};
	
	/**
		Display a notification that is related to a form element.
		Typically, an error message that explains what need to be
		filled in in an invalid element. 
		
		@param element A document element or the id of an element, if no
			element is found the message is displayed with an alert dialog
			instead
		@param message The message to display
		@param subclass Optional, a subclass that is applied to the notification 'div' container
		@param pointerClass Optional, if not given the function will automatically decide if
			the pointer is above or below the notification. Accept values: pointer-above, pointer-below, pointer-left
		@param pointerAlign 'left' or 'right', if not specified auto-select depending on location on page
	*/
	forms.showNotification = function(element, message, subclass, pointerClass, pointerAlign)
	{
		element = Doc.element(element);
		if (!element)
		{
			alert(message);
		}
		else
		{
			// Create the notification div element
			var notifyDiv = document.createElement('div');
			notifyDiv.className = 'notify';
			if (subclass) notifyDiv.className += ' ' + subclass;

			// Decide how to position the notification depending on the element's
			// position in the window
			var pos = Doc.getElementPosition(element);
			var winPos = App.getWindowPosition();
			if (!pointerClass)
			{
				// Automatic detection if pointer is above or below
				if (pos.top+pos.height+100 > winPos.height)
				{
					// The element is too close to the bottom so we need to
					// position the notification above it
					pointerClass = 'pointer-below';
				}
				else
				{
					// Position the notification below the element
					pointerClass = 'pointer-above';
				}
			}
			
			var pointerPos; // style for the pointer position
			if (pointerClass == 'pointer-below')
			{
				notifyDiv.style.bottom = (winPos.height-pos.top+8)+'px';
				notifyDiv.style.top = 'auto';
			}
			else if (pointerClass == 'pointer-left')
			{
				notifyDiv.style.top = (pos.top-2) + 'px';
				notifyDiv.style.left = (pos.left+pos.width+8)+'px';
				pointerPos = 'top: 0.6em;';
			}
			else // pointer-above
			{
				notifyDiv.style.top = (pos.top+pos.height+8)+'px';
			}
			
			if (!pointerPos)
			{
				if (pos.left > winPos.width / 2 && pointerAlign != 'left')
				{
					// If we are on the right side of the window, 
					// align the notification to the right
					notifyDiv.style.right = (winPos.width-(pos.left+pos.width)) + 'px';
					pointerPos = pos.width < 30 ? 'right: 0.6em; left: auto;' : 'right: 1.2em; left: auto;';
				}
				else
				{
					// otherwise, align it to the left
					notifyDiv.style.left = (pos.left+3)+'px';
					pointerPos = pos.width < 30 ? 'left: 0.6em;' : 'left: 1.2em;';
				}
			}
			notifyDiv.innerHTML = '<div class="'+pointerClass+'" style="'+pointerPos+'"><div></div></div><div class="notify-message">'+message+'</div>';

			// Display the notification
			document.body.appendChild(notifyDiv);
			if (element.tabIndex < 0) element.tabIndex = 0;
			if (element.focus) element.focus();
			if (element.select) element.select();

			// Remember the div element and add event handlers that hide it when the element 
			// loses focus or a key is pressed
			element.notifyDiv = notifyDiv;
			notifyDiv.notifyElement = element;
			element.addEventListener('blur', internal.hideNotification, false);
			element.addEventListener('keypress', internal.hideNotification, false);
			notifyDiv.addEventListener('click', internal.hideNotification, false);
		}
	}
	
	/**
		Event handler that hide the notification message displayed with
		Forms.showNotification
	*/
	internal.hideNotification = function(event)
	{
		// The target of the event can be either the form element or the notification div
		var target = event.currentTarget; 
		var element;
		var notifyDiv;
		if (target.notifyDiv)
		{
			element = target;
			notifyDiv = element.notifyDiv;
		}
		else if (target.notifyElement)
		{
			element = target.notifyElement;
			notifyDiv = target;
		}
		
		// Remove the notification
		if (notifyDiv)
		{
			document.body.removeChild(notifyDiv);
		}
		// Remove the event handlers from the element
		if (element)
		{
			element.removeEventListener('keypress', internal.hideNotification);
			element.removeEventListener('blur', internal.hideNotification);
		}
	}

	
	/**
		Create a hidden input element in a form.
		@param frm The form object
		@param name The name of the hidden element
		@param value The value of the hidden element
	*/
	forms.addHidden = function(frm, name, value)
	{
		frm = Doc.form(frm);
		var hidden = frm.ownerDocument.createElement('input');
		hidden.setAttribute('type', 'hidden');
		hidden.setAttribute('name', name);
		hidden.setAttribute('value', value);
		frm.appendChild(hidden);
	}
	
	// Kept for backwards compatibility
	// @Deprecated
	forms.createHidden = function(frm, name, value, doc)
	{
		App.deprecatedMethod('Forms.createHidden()', '3.5');
		Forms.addHidden(frm, name, value);
	}

	
	/*
		Convert an URL to a HTML form using POST. This method extract 
		all query parameters from the URL (except ID) and generate a 
		<form> tag with hidden input fields. 
		The method returns an unattached Form object. To post it, 
		do: document.body.appendChild(form); form.submit(); If there
		is no query part in the URL, null is returned.
		@param url The url to convert
		@param targetWin The window the form should post to, if no window is given
			or if it doesn't have a name the form is posted to the current
			window.
	*/
	forms.convertUrlToForm = function(url, targetWin)
	{
		// Split the URL at '?' and extract main part and query part
		// If there is no query part return null.
		var i = url.indexOf('?');
		if (i < 0) return null;
		var query = url.substring(i+1);
		if (!query) return null;
		
		url = url.substring(0, i)+'?ID='+App.getSessionId();
	
		// Create new <form> tag
		var frm = document.createElement('form');
		frm.setAttribute('action', url);
		frm.setAttribute('method', 'post');
		if (targetWin && targetWin.name) 
		{
			frm.setAttribute('target', targetWin.name);
		}
		
		// Add <input type="hidden" ...> elements to the form
		var options = query.split('&');
		for (var i = 0; i < options.length; i++)
		{
			var kv = options[i].split('=');
			var name = kv[0];
			var value = decodeURIComponent(kv[1]);
			if (name && name != 'ID')
			{
				forms.addHidden(frm, name, value, document);
			}
		}
		return frm;
	}
	
	/**
		Clone the given form and add all elements in it as
		hidden elements in the cloned form. Note that unselected
		checkboxes and radio buttons are not cloned. Selection lists
		with multiple selections are cloned as multiple hidden fields.
	*/
	forms.cloneAsHidden = function(frm)
	{
		// Duplicate in hidden form
		var cloned = document.createElement('form');
		cloned.setAttribute('action', frm.action);
		cloned.setAttribute('method', frm.method);
		if (frm.target)
		{
			cloned.setAttribute('target', frm.target);
		}
		
		for (var i = 0; i < frm.elements.length; i++)
		{
			var e = frm.elements[i];
			var shouldClone = e.type == 'text' ||  e.type == 'textarea' || e.type == 'select-one' || e.type == 'password' || e.type == 'hidden';
			shouldClone |= (e.type == 'radio' || e.type == 'checkbox' ) && e.checked; 
			if (shouldClone)
			{
				Forms.addHidden(cloned, e.name, e.value);
			}
			else if (e.type == 'select-multiple')
			{
				for (var j = 0; j < e.length; j++)
				{
					if (e[j].selected) Forms.addHidden(cloned, e.name, e[j].value);
				}
			}
		}
		return cloned;
	}
	
	/**
		Event handler that check/uncheck all checkboxes in a form. The form
		should be specified in the target element's data-form attribute. The 
		data-regexp attribute can be set to force a specific match on the name 
		attribute of checkboxes, otherwise only numeric names are matched.
	*/
	forms.checkUncheckOnClick = function(event)
	{
		var element = event.currentTarget;
		var form = Data.get(element, 'form');
		var nameRegexp = Data.get(element, 'regexp');
		var specialKey = event.altKey || event.ctrlKey || event.shiftKey;
		if (specialKey)
		{
			Forms.checkUncheck(form, regexp);
		}
		else
		{
			Forms.toggleCheckboxes(form, regexp);
		}
	}
	
	/**
		Check or unchecks all enabled checkboxes in a form. If the first checkbox that
		is found is checked they are unchecked, otherwise they are checked.
		The method only considers checkboxes having a name consisting only
		of digits.
		@param frm A form object or the name or id of a form object
		@param nameRegexp A regular expression that must match the name of the checkbox,
			default matches names with digits
	*/
	forms.checkUncheck = function(frm, nameRegexp)
	{
		frm = Doc.form(frm);
		var first = null;
		if (!nameRegexp) nameRegexp = /\d+/;
		for (var i=0; i < frm.elements.length; i++)
		{
			var element = frm.elements[i];
			if (element.type == 'checkbox' && !element.disabled && element.name.match(nameRegexp))
			{
				var was = element.checked;
				if (first == null)
				{
					first = element;
					element.checked = !element.checked;
				}
				else
				{
					element.checked = first.checked;
				}
				if (was != element.checked) Events.sendChangeEvent(element);
			}
		}
	}
	
	/**
		Toggle checkboxes in a form. The method only considers checkboxes 
		having a name consisting only of digits (unless a regular expression is given).
		@param frm A form object or the name or id of a form object
		@param nameRegexp A regular expression that must match the name of the checkbox,
			default matches names with digits
	*/
	forms.toggleCheckboxes = function(frm, nameRegexp)
	{
		frm = Doc.form(frm);
		if (!nameRegexp) nameRegexp = /\d+/;
		for (var i=0; i < frm.elements.length; i++)
		{
			var element = frm.elements[i];
			if (element.type == 'checkbox' && !element.disabled && element.name.match(nameRegexp))
			{
				element.checked = !element.checked;
				Events.sendChangeEvent(element);
			}
		}
	}

	/**
		Calculate the number of checked checkboxes in a form. The method only
		considers checkboxes having a name consisting only
		of digits. As a special case, if no checkboxes are found, then
		it checks if a radio button group with the name item_id exists and
		if one of those is checked.
		@param frm A form object or the name or id of a form object
		@param nameRegexp A regular expression that must match the name of the checkbox,
			default matches names with digits
	*/
	forms.numChecked = function(frm, nameRegexp)
	{
		frm = Doc.form(frm);
		var checked = 0;
		if (!nameRegexp) nameRegexp = /\d+/;
		for (var i=0; i < frm.elements.length; i++) 
		{
			var element = frm.elements[i];
			if (element.type == 'checkbox' && element.name.match(nameRegexp) && element.checked)
			{
				checked++;
			}
		}
		if (checked == 0 && frm.item_id)
		{
			checked += Forms.getCheckedRadio(frm.item_id) ? 1 : 0;
		}
		return checked;
	}

	
	/**
		Adds a list option to a selection list. The option will be placed 
		before the option specified by the index value. Use -1 to insert the 
		option at the end.
		@param list The ID or Select object
		@param index The index of the new option
		@param option The Option object
	*/
	forms.addListOption = function(list, index, option)
	{
		list = Doc.element(list);
		if (option.hasAttribute('text') && !option.hasAttribute('title'))
		{
			option.title = option.text;
		}
		if (index < 0 || index >= list.length)
		{
			// Add to end of list
			list[list.length] = option;
		}
		else
		{
			list.add(option, list[index]);
		}
	}
	
	/**
		Switch place of two options in a list
		@param list The list object
		@param index1 The index of the first option
		@param index2 The index of the second option
	*/
	forms.switchListOptions = function(list, index1, index2)
	{
		list = Doc.element(list);
		var opt1 = list.options[index1];
		var opt2 = list.options[index2];
		list.options[index1] = new Option();
		list.options[index2] = new Option();
		list.options[index1] = opt2;
		list.options[index2] = opt1;
	}

	
	/**
		Move all selected options in the list up or down. Returns the number
		of options that were selected (and moved).
		@param list The list as an Select object
		@param down A boolean value (true to move down, false to move up)
	*/
	forms.moveListOptions = function(list, down)
	{
		list = Doc.element(list);
		var moved = 0;
		if (list.options.length > 0)
		{
			var index = 1;
			var end_condition = list.options.length;
			var delta = -1;
			if (down)
			{
				index = list.options.length - 2;
				end_condition = -1;
				delta = 1;
			}
			while (index != end_condition)
			{
				if (!list.options[index+delta].selected && list.options[index].selected)
				{
					moved++;
					forms.switchListOptions(list, index, index+delta);
				}
				index = index-delta;
			}
		}
		return moved;
	}

	
	/**
		Select the option in a list with the specified value. If
		the value is not found among the options the selection is left
		unchanged.
		@param list The list object
		@param value The value to look for
	*/
	forms.selectListOption = function(list, value)
	{
		list = Doc.element(list);
		for (var i = 0; i < list.length; i++)
		{
			if (list[i].value == value)
			{
				list.selectedIndex = i;
				return;
			}
		}
	}

	
	/**
		Select all options in a list which has a value among one of the specified 
		values. Values not found are deselected.
		@param list The list object
		@param values An array with the values
		@param checkTextIfNoValues If set and none of the values was found, the text for each option is also checked
		@return The number of selected values in the list
	*/
	forms.selectListOptions = function(list, values, checkTextIfNoValues)
	{
		list = Doc.element(list);
		var numSelected = 0;
		for (var i = 0; i < list.length; i++)
		{
			var selected = false;
			for (var j = 0; j < values.length; j++)
			{
				if (list[i].value == values[j]) 
				{
					selected = true;
					j = values.length;
					numSelected++;
				}
			}
			list[i].selected = selected;
		}
		if (numSelected == 0 && checkTextIfNoValues)
		{
			for (var i = 0; i < list.length; i++)
			{
				var selected = false;
				for (var j = 0; j < values.length; j++)
				{
					if (list[i].text == values[j]) 
					{
						selected = true;
						j = values.length;
						numSelected++;
					}
				}
				list[i].selected = selected;
			}			
		}
		return numSelected;
	}

	
	/**
		Assuming that a list is sorted, get the index of
		the entry point were an option with the given text 
		would be inserted.
		@param list The list to search
		@param text The value to compare against the text
			on the list options
	*/
	forms.getInsertIndexOfSortedList = function(list, text)
	{
		list = Doc.element(list);
		for (var i = 0; i < list.length; i++)
		{
			if (list[i].text > text) return i;
		}
		return list.length;
	}
	
	/**
	 	Move all selected options from list1 to list2. Return
	 	the number of moved options. If the 'to' list has
	 	set the option 'data-is-sorted' the options selected
	 	options are inserted into their sorted positions, otherwise
	 	at the end of the list. Options with the 'data-no-remove' attribute
	 	set are not moved.
	*/
	forms.moveSelected = function(from, to)
	{
		from = Doc.element(from);
		to = Doc.element(to);
		var sorted = Data.int(to, 'is-sorted');
		
		var moved = 0;
		for (var i=0; i < from.length; i++)
		{
			var option = from[i];
			if (option.selected && !Data.int(option, 'no-remove', 0))
			{
				moved++;
				var movedOption = new Option(option.text, option.value, false, true);
				var insertIndex = sorted ? forms.getInsertIndexOfSortedList(to, movedOption.text) : to.length;
				forms.addListOption(to, insertIndex, movedOption);
				from[i] = null;
				i--;
			}
		}
		return moved;
	}
	
	/**
		Event handler for moving selected options from one list to
		another. The event handler may be attached to any element
		which must specify 'data-from-id' and 'data-to-id' with
		ID values for the two lists. Or, it can be attached to the
		'from' list (typcially as dblclick event), in which case 
		only a 'data-to-id' must be specified.
	*/
	forms.moveSelectedOnClick = function(event)
	{
		var from = Data.get(event.currentTarget, 'from-id') || event.currentTarget;
		var to = Data.get(event.currentTarget, 'to-id');
		forms.moveSelected(from, to);
	}
	
	/**
		Get the checked radio button element in a group of radio buttons. Returns
		null if no radio button is selected.
		@param radio The radio buttons group object
	*/
	forms.getCheckedRadio = function(radio)
	{
		if (radio.checked)
		{
			// If there is only one radio button
			return radio;
		}
		for (var i = 0;  i < radio.length; i++)
		{
			if (radio[i].checked) return radio[i];
		}
		return null;
	}

	/**
		Check the radio button with the specified value. If no
		radio button has a matching value all buttons are unchecked. Returns
		the index of the checked radio button. Radio buttons that have
		been disabled are never checked.
		@param radio The radio buttons group object
		@param value The value of the radio button that should be checked
	*/
	forms.checkRadio = function(radio, value)
	{
		var returnValue = -1;
		if (radio.length)
		{
			// More than one item in the list
			for (var i = 0; i < radio.length; i++)
			{
				radio[i].checked = (radio[i].value == value) && !radio[i].disabled;
				returnValue = radio[i].checked ? i : returnValue;
			}
		}
		else
		{
			// Only a single item in the list
			radio.checked = radio.value == value && !radio.disabled;
			returnValue = 0;
		}
		return returnValue;
	}

	/**
		Check all checkboxes having a value matching one of the
		specified values. Buttons that don't matches any value
		are unchecked.
		@param checkBoxes The check box group
		@param value An array of values
		@return The number of checked boxes
	*/
	forms.checkCheckBoxes = function(checkBoxes, values)
	{
		var numChecked = 0;
		for (var i = 0; i < checkBoxes.length; i++)
		{
			checkBoxes[i].checked = false;
			for (var j = 0; j < values.length; j++)
			{
				if (checkBoxes[i].value == values[j])
				{
					checkBoxes[i].checked = true;
					numChecked++;
					j = values.length;
				}
			}
		}
		return numChecked;
	}

	
	/**
		Event handler that submit the form of the element it is attached to
		when the event is triggered. 
	*/
	forms.submit = function(event)
	{
		var element = event.currentTarget;
		if (element.form && element.form.submit) element.form.submit(); 
	}
	
	/**
		Initializer that set the focus on the element
		that has data-auto-init="focus" or data-auto-init="focus-select"
	*/
	internal.autoFocus = function(element, autoInit)
	{
		if (autoInit == 'focus' || autoInit == 'focus-select')
		{
			// Defer setting the focus until since the element may not yet be visible
			Doc.addFinalizer(function()
			{
				setTimeout(function()
					{
						element.focus();
						if (autoInit == 'focus-select') element.select();
					}, 250);
			});
		}
	}
	Doc.addElementInitializer(internal.autoFocus);
	
	/**
		Initializer that attach event handler to input elements
		that restricts typing to either integer or numbers.
		Note that this initializer is not active by default, but
		has to be added with Doc.addElementInitializer(Forms.numberOnlyInitializer)
		on pages that need it.
	*/
	forms.numberOnlyInitializer = function(element, autoInit)
	{
		if (autoInit == 'integer-only')
		{
			Events.addEventHandler(element, 'keypress', Events.integerOnly);
		}
		else if (autoInit == 'number-only')
		{
			Events.addEventHandler(element, 'keypress', Events.numberOnly);
		}
	}
	
	
	/**
		Find all <label> tags and create a reverse link between it and
		the checkbox/radiobutton it controls so that we can disable the
		label when the checkbox/radiobutton is disabled
	*/
	forms.linkCheckboxesWithLabels = function(element)
	{
		if (!element) element = document;
		var labels = element.getElementsByTagName('label');
		if (labels.length > 0)
		{
			if (!window.MutationObserver) return;
			// We listen on attribute changes on the checkboxes/radiobuttons
			var observer = new MutationObserver(internal.checkBoxModified);
			var config = { attributes: true };
			
			for (var i = 0; i < labels.length; i++)
			{
				var lbl = labels[i];
				// Check the 'for' attribute first
				var refId = lbl.getAttribute('for');
				var ref;
				if (refId)
				{
					ref = Doc.element(refId);
				}
				else
				{
					// Check child elements for <input>
					var inputs = lbl.getElementsByTagName('input');
					if (inputs.length == 1) ref = inputs[0];
				}
				if (ref)
				{
					// Set initial 'disabled' status on <label>
					Doc.addOrRemoveClass(lbl, 'disabled', ref.disabled);
					// ...and get notified  when checkbox/radiobutton is modified
					ref.labelRef = lbl;
					observer.observe(ref, config);
				}
			}
		}
	}
	
	/**
		Callback when attributes have changed on list of checkboxes/radiobuttons
	*/
	internal.checkBoxModified = function(records)
	{
		for (var i = 0; i < records.length; i++)
		{
			var target = records[i].target;
			if (target.labelRef)
			{
				Doc.addOrRemoveClass(target.labelRef, 'disabled', target.disabled);
			}
		}
	}
	
	Doc.addFinalizer(forms.linkCheckboxesWithLabels);
		
	return forms;
}();

var Strings = function()
{
	var strings = {};
	
	/**
		Trim leading and trailing whitespace from a string.
		@param value The string to trim
		@return The trimmed string
	*/
	strings.trim = function(value)
	{
		if (!value) return '';
		value = value.replace(/^\s+/g, '');
		value = value.replace(/\s+$/g, '');
		return value;
	}
	
	/**
		Cut a string if it is longer than a specified length. If cut,
		three dots are appended to the string. The returned string is never
		longer than the specified max length.
	
		@param value The string to cut
		@param maxLength The maximum length
	*/
	strings.cut = function(value, maxLength)
	{
		if (!value) return '';
		if (value.length > maxLength) 
		{
			value = value.substr(0, maxLength-3)+'...';
		}
		return value;
	}
	
	/**
		Encode HTML ampersand, start and end brackets to
		make sure value is displayed as-is (not
		as HTML).
	*/
	strings.encodeTags = function(value)
	{
		if (!value) return '';
		var encoded = value;
		encoded = encoded.replace(/\&/g, '&amp;');
		encoded = encoded.replace(/\</g, '&lt;');
		encoded = encoded.replace(/\>/g, '&gt;');
		return encoded;
	}

	/**
		Decode HTML amptersand, start and end brackets.
	*/
	strings.decodeTags = function(value)
	{
		if (!value) return '';
		var unencoded = value;
		unencoded = unencoded.replace(/\&lt\;/g, '<');
		unencoded = unencoded.replace(/\&gt\;/g, '>');
		unencoded = unencoded.replace(/\&amp\;/g, '&');
		return unencoded;
	}
	
	/**
		Compare two strings ignoring case.
		returns -1 if a<b, +1 if a>b, 0 if a==b
	*/
	strings.compareIgnoreCase = function(a, b)
	{
		var lcA = a.toLowerCase();
		var lcB = b.toLowerCase();
		if (lcA < lcB) return -1;
		if (lcA > lcB) return +1;
		return 0;
	}

	return strings;
}();


var Numbers = function()
{
	var numbers = {};
	var GB = 1073741824;
	var MB = 1048576;
	var kB = 1024;

	/**
		Checks if the supplied value is an integer.
	*/
	numbers.isInteger = function(value)
	{
		return (value.search(/^\-?\d+$/) != -1);
	}

	/**
		Format a number with the specified number of decimals
	*/
	numbers.formatNumber = function(value, numDecimals, unit)
	{
		if (typeof value != 'number') return '';
		var result = '';
		if (numDecimals < 0)
		{
			result = value;
		}
		else
		{
			result = value.toFixed(numDecimals);
		}
		if (unit) result += unit;
		return result;
	}

	numbers.formatBytes = function(value)
	{
		if (typeof value != 'number') return '';
		if (value > GB)
		{
			return numbers.formatNumber(value / GB, 1, ' GB');
		}
		else if (value > MB)
		{
			return numbers.formatNumber(value / MB, 1, ' MB');
		}
		else if (value > kB)
		{
			return numbers.formatNumber(value / kB, 1, ' kB');
		}
		else
		{
			return value + ' bytes';
		}
	}

	numbers.formatSeconds = function(seconds)
	{
		if (!seconds) return '';
		var result = '';
		var mustAppend = false;
		if (seconds >= 86400)
		{
			var days = Math.floor(seconds / 86400);
			seconds = seconds - 86400*days;
			result += days;
			result += days == 1 ? " day " : " days ";
			mustAppend = true;
		}
		if (seconds >= 3600 || mustAppend)
		{
			var hours = Math.floor(seconds / 3600);
			seconds = seconds - 3600*hours;
			result += hours;
			result += hours == 1 ? " hour " : " hours ";
			mustAppend = true;
		}
		if (seconds >= 60 || mustAppend)
		{
			var minutes = Math.floor(seconds / 60);
			seconds = seconds - 60*minutes;
			result += minutes;
			result += minutes == 1 ? " minute " : " minutes ";
			mustAppend = true;
		}
		result += seconds;
		result += seconds == 1 ? " second" : " seconds";
		return result;
	}

	// Kept for backwards compatibility
	// @Deprecated
	numbers.numberOnly = function(event)
	{
		App.deprecatedMethod('Numbers.numberOnly()', '3.5');
		Events.numberOnly(event);
	}
	
	// Kept for backwards compatibility
	// @Deprecated
	numbers.integerOnly = function(event)
	{
		App.deprecatedMethod('Numbers.integerOnly()', '3.5');
		Events.integerOnly(event);
	}
	
	return numbers;
}();


var Dates = function()
{
	//Functions related to handling date values
	//Implementation based on code provided by:
	//===================================================================
	//Author: Matt Kruse <matt@mattkruse.com>
	//WWW: http://www.mattkruse.com/
	//
	//NOTICE: You may use this code for any purpose, commercial or
	//private, without any further permission from the author. You may
	//remove this notice from your final code if you wish, however it is
	//appreciated by the author if at least my web site address is kept.
	//
	//You may *NOT* re-distribute this code in any way except through its
	//use. That means, you can include it in your product, or your web
	//site, or any other form where the code is actually being used. You
	//may not put the plain javascript up on your site for download or
	//include it in your javascript libraries for download. 
	//If you wish to share this code with others, please just point them
	//to the URL instead.
	//Please DO NOT link directly to my .js files from your site. Copy
	//the files to your server and use them there. Thank you.
	//===================================================================
	//HISTORY
	//------------------------------------------------------------------
	//May 17, 2003: Fixed bug in parseDate() for dates <1970
	//March 11, 2003: Added parseDate() function
	//March 11, 2003: Added "NNN" formatting option. Doesn't match up
	//	              perfectly with SimpleDateFormat formats, but 
	//	              backwards-compatability was required.

	//------------------------------------------------------------------
	//These functions use the same 'format' strings as the 
	//java.text.SimpleDateFormat class, with minor exceptions.
	//The format string consists of the following abbreviations:
	//
	//Field        | Full Form          | Short Form
	//-------------+--------------------+-----------------------
	//Year         | yyyy (4 digits)    | yy (2 digits), y (2 or 4 digits)
	//Month        | MMMM(name or abbr.)| MM (2 digits), M (1 or 2 digits)
	//	           | NNN (abbr.)        |
	//Day of Month | dd (2 digits)      | d (1 or 2 digits)
	//Day of Week  | EE (name)          | E (abbr)
	//Hour (1-12)  | hh (2 digits)      | h (1 or 2 digits)
	//Hour (0-23)  | HH (2 digits)      | H (1 or 2 digits)
	//Hour (0-11)  | KK (2 digits)      | K (1 or 2 digits)
	//Hour (1-24)  | kk (2 digits)      | k (1 or 2 digits)
	//Minute       | mm (2 digits)      | m (1 or 2 digits)
	//Second       | ss (2 digits)      | s (1 or 2 digits)
	//AM/PM        | a                  |
	//
	//NOTE THE DIFFERENCE BETWEEN MM and mm! Month=MM, not mm!
	//Examples:
	//"MMM d, y" matches: January 01, 2000
	//	                   Dec 1, 1900
	//	                   Nov 20, 00
	//"M/d/yy"   matches: 01/20/00
	//	                   9/2/00
	//"MMM dd, yyyy hh:mm:ssa" matches: "January 01, 2000 12:30:45AM"
	//------------------------------------------------------------------
	
	var MONTH_NAMES = ['January','February','March','April','May','June','July','August','September','October','November','December','Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec'];
	var DAY_NAMES = ['Sunday','Monday','Tuesday','Wednesday','Thursday','Friday','Saturday','Sun','Mon','Tue','Wed','Thu','Fri','Sat'];

	
	var dates = {};
	var internal = {};
	
	// convert a date to a string:
	// default format is yyyy-MM-dd
	dates.formatDate = function(date, format)
	{
		if (!format) format = "yyyy-MM-dd";
		var result="";
		var i_format=0;
		var c="";
		var token="";
		var y=date.getYear()+"";
		var M=date.getMonth()+1;
		var d=date.getDate();
		var E=date.getDay();
		var H=date.getHours();
		var m=date.getMinutes();
		var s=date.getSeconds();
		var yyyy,yy,MMM,MM,dd,hh,h,mm,ss,ampm,HH,H,KK,K,kk,k;
		
		// Convert real date parts into formatted versions
		var value=new Object();
		if (y.length < 4) 
		{
			y=""+(y-0+1900);
		}
		value["y"]=""+y;
		value["yyyy"]=y;
		value["yy"]=y.substring(2,4);
		value["M"]=M;
		value["MM"]=internal.LZ(M);
		value["MMM"]=MONTH_NAMES[M+11];
		value["MMMM"]=MONTH_NAMES[M-1];
		value["NNN"]=MONTH_NAMES[M+11];
		value["d"]=d;
		value["dd"]=internal.LZ(d);
		value["E"]=DAY_NAMES[E+7];
		value["EE"]=DAY_NAMES[E];
		value["H"]=H;
		value["HH"]=internal.LZ(H);
		if (H==0) 
		{
			value["h"]=12;
		}
		else if (H>12)
		{
			value["h"]=H-12;
		}
		else 
		{
			value["h"]=H;
		}
		value["hh"]=internal.LZ(value["h"]);
		if (H>11)
		{
			value["K"]=H-12;
		}
		else 
		{
			value["K"]=H;
		}
		value["k"]=H+1;
		value["KK"]=internal.LZ(value["K"]);
		value["kk"]=internal.LZ(value["k"]);
		if (H > 11) 
		{
			value["a"]="PM"; 
		}
		else 
		{
			value["a"]="AM"; 
		}
		value["m"]=m;
		value["mm"]=internal.LZ(m);
		value["s"]=s;
		value["ss"]=internal.LZ(s);
		while (i_format < format.length) 
		{
			c=format.charAt(i_format);
			token="";
			while ((format.charAt(i_format)==c) && (i_format < format.length)) 
			{
				token += format.charAt(i_format++);
			}
			if (value[token] != null) 
			{
				result=result + value[token]; 
			}
			else 
			{
				result=result + token; 
			}
		}
		return result;
	}

	/*
		Convert a year value to a string with four digits, ie. pad with zeroes
		as needed
	*/
	dates.getFourDigitYear = function(year)
	{
		if (year < 10) 
		{
			year = '000'+year;
		}
		else if (year < 100) 
		{
			year = '00'+year;
		}
		else if (year < 1000) 
		{
			year = '0'+year;
		}
		return year;
	}
	
	/*
		Create a new date object with the specified year, month and day
	*/
	dates.newDate = function(year, month, day)
	{
		var date = new Date();
		date.setFullYear(year, month, day);
		return date;
	}
	
	// ------------------------------------------------------------------
	// isDate ( date_string, format_string )
	// Returns true if date string matches format of format string and
	// is a valid date. Else returns false.
	// It is recommended that you trim whitespace around the value before
	// passing it to this function, as whitespace is NOT ignored!
	// ------------------------------------------------------------------
	dates.isDate = function(val, format) 
	{
		return dates.parseString(val, format) != null;
	}

	// ------------------------------------------------------------------
	// parseString( date_string , format_string )
	//
	// This function takes a date string and a format string. It matches
	// If the date string matches the format string, it returns the 
	// the date object. If it does not match, it returns null.
	// ------------------------------------------------------------------
	dates.parseString = function(val, format)
	{
		if (!format) format = 'yyyy-MM-dd';
		var i_val=0;
		var i_format=0;
		var c="";
		var token="";
		var token2="";
		var x,y;
		var now=new Date();
		var year=now.getYear();
		var month=now.getMonth()+1;
		var date=1;
		var hh=now.getHours();
		var mm=now.getMinutes();
		var ss=now.getSeconds();
		var ampm="";
		
		while (i_format < format.length) 
		{
			// Get next token from format string
			c=format.charAt(i_format);
			token="";
			while ((format.charAt(i_format)==c) && (i_format < format.length)) 
			{
				token += format.charAt(i_format++);
			}
			// Extract contents of value based on format token
			if (token=="yyyy" || token=="yy" || token=="y") 
			{
				if (token=="yyyy") { x=4;y=4; }
				if (token=="yy")   { x=2;y=2; }
				if (token=="y")    { x=2;y=4; }
				year=internal.getInt(val,i_val,x,y);
				if (year==null) { return null; }
				i_val += year.length;
				if (year.length==2) 
				{
					if (year > 70) 
					{ 
						year=1900+(year-0); 
					}
					else 
					{ 
						year=2000+(year-0); 
					}
				}
			}
			else if (token=='MMMM' || token=="MMM" || token=="NNN")
			{
				month=0;
				for (var i=0; i<MONTH_NAMES.length; i++) 
				{
					var month_name=MONTH_NAMES[i];
					if (val.substring(i_val,i_val+month_name.length).toLowerCase()==month_name.toLowerCase()) 
					{
						month=i+1;
						if (month>12) { month -= 12; }
						i_val += month_name.length;
						break;
					}
				}
				if ((month < 1)||(month>12)) { return null; }
			}
			else if (token=="EE"||token=="E")
			{
				for (var i=0; i<this.DAY_NAMES.length; i++) 
				{
					var day_name=DAY_NAMES[i];
					if (val.substring(i_val,i_val+day_name.length).toLowerCase()==day_name.toLowerCase()) 
					{
						i_val += day_name.length;
						break;
					}
				}
			}
			else if (token=="MM"||token=="M") 
			{
				month=internal.getInt(val,i_val,token.length,2);
				if(month==null||(month<1)||(month>12)){return null;}
				i_val+=month.length;
			}
			else if (token=="dd"||token=="d") 
			{
				date=internal.getInt(val,i_val,token.length,2);
				if(date==null||(date<1)||(date>31)){return null;}
				i_val+=date.length;
			}
			else if (token=="hh"||token=="h") 
			{
				hh=internal.getInt(val,i_val,token.length,2);
				if(hh==null||(hh<1)||(hh>12)){return null;}
				i_val+=hh.length;
			}
			else if (token=="HH"||token=="H") 
			{
				hh=internal.getInt(val,i_val,token.length,2);
				if(hh==null||(hh<0)||(hh>23)){return null;}
				i_val+=hh.length;
			}
			else if (token=="KK"||token=="K") 
			{
				hh=internal.getInt(val,i_val,token.length,2);
				if(hh==null||(hh<0)||(hh>11)){return null;}
				i_val+=hh.length;
			}
			else if (token=="kk"||token=="k") 
			{
				hh=internal.getInt(val,i_val,token.length,2);
				if(hh==null||(hh<1)||(hh>24)){return null;}
				i_val+=hh.length;hh--;
			}
			else if (token=="mm"||token=="m") 
			{
				mm=internal.getInt(val,i_val,token.length,2);
				if(mm==null||(mm<0)||(mm>59)){return null;}
				i_val+=mm.length;
			}
			else if (token=="ss"||token=="s") 
			{
				ss=internal.getInt(val,i_val,token.length,2);
				if(ss==null||(ss<0)||(ss>59)){return null;}
				i_val+=ss.length;
			}
			else if (token=="a") 
			{
				if (val.substring(i_val,i_val+2).toLowerCase()=="am") {ampm="AM";}
				else if (val.substring(i_val,i_val+2).toLowerCase()=="pm") {ampm="PM";}
				else {return null;}
				i_val+=2;
			}
			else 
			{
				if (val.substring(i_val,i_val+token.length)!=token) {return null;}
				else {i_val+=token.length;}
			}
		}
		
		// If there are any trailing characters left in the value, it doesn't match
		if (i_val != val.length) { return null; }
		
		// Is date valid for month?
		if (month==2) 
		{
			// Check for leap year
			if ( ( (year%4==0)&&(year%100 != 0) ) || (year%400==0) ) 
			{ // leap year
				if (date > 29){ return null; }
			}
			else 
			{ 
				if (date > 28) { return null; } 
			}
		}
		if ((month==4)||(month==6)||(month==9)||(month==11)) 
		{
			if (date > 30) { return null; }
		}
		
		// Correct hours value
		if (hh<12 && ampm=="PM") 
		{ 
			hh=hh-0+12; 
		}
		else if (hh>11 && ampm=="AM") 
		{ 
			hh-=12; 
		}
		return new Date(year,month-1,date,hh,mm,ss);
	}

	internal.isInteger = function(val) 
	{
		var digits="1234567890";
		for (var i=0; i < val.length; i++) 
		{
			if (digits.indexOf(val.charAt(i))==-1) { return false; }
		}
		return true;
	}
	
	internal.getInt = function(str,i,minlength,maxlength) 
	{
		for (var x=maxlength; x>=minlength; x--)
		{
			var token=str.substring(i,i+x);
			if (token.length < minlength) { return null; }
			if (internal.isInteger(token)) { return token; }
		}
		return null;
	}
	
	internal.LZ = function(x) 
	{
		return(x<0||x>9?"":"0")+x;
	}

	/*
		Calculate the number of days in the specified month
		Months are numbered from 1 to 12. Returns 0 if the 
		specified month is invalid.
	*/
	dates.daysInMonth = function(year, month)
	{
		switch (month)
		{
			// Months with 31 days
			case 1:
			case 3:
			case 5:
			case 7:
			case 8:
			case 10:
			case 12:
				return 31;
				break;
			// Months with 30 days
			case 4:
			case 6:
			case 9:
			case 11:
				return 30;
				break;
			
			// February, taking leap years into account
			case 2:
				if (((year % 4 == 0) && (year % 100 != 0)) || (year % 400 == 0))
				{
					return 29;
				}
				else
				{
					return 28;
				}
				break;
	
			default:
				return 0;
		}
		return 0;
	}

	/*
		Opens a popup window allowing a user to select a date graphically. The 
		current date is taken from the specified input field of the specified form and 
		is saved back to it when the window is closed.
	
		@param title The title of the window
		@param form The name of the form where the input field is located
		@param input The name of the input field that holds the date value
		@param callback Deprecated, onchange event is sent instead
	*/
	// @Deprecated
	dates.selectDate = function(title, form, input, callback, format)
	{
		App.deprecatedMethod('Dates.selectDate()', '3.5');
		var textarea = document.forms[form][input];
		if (!textarea.id) textarea.id = 't'+(new Date()).getTime();
		Dialogs.openCalendar(textarea, title, format, false);
	}
	
	/*
		Opens a popup window allowing a user to select a date and time graphically. The 
		current value is taken from the specified input field of the specified form and 
		is saved back to it when the window is closed.
	
		@param title The title of the window
		@param form The name of the form where the input field is located
		@param input The name of the input field that holds the date value
		@param callback Deprecated, onchange event is sent instead
	*/
	// @Deprecated
	dates.selectDateTime = function(title, form, input, callback, format)
	{
		App.deprecatedMethod('Dates.selectDateTime()', '3.5');
		var textarea = document.forms[form][input];
		if (!textarea.id) textarea.id = 't'+(new Date()).getTime();
		Dialogs.openCalendar(textarea, title, format, true);
	}

	return dates;
}();


var MultiSelect = function()
{
	var select = {};
	var internal = {};
	
	// The currently active multi-select element
	var activeSelect = null;
	// The last active multi-select element
	var lastSelect = null;
	// Flag to indicate that automatic hiding of the options list 
	// should be disabled due to the user is interacting with the list
	var disableHide = false;
	// The option that is currently higlighted
	var highlightedOption = null;
	
	/**
		Get the number of options that are checked in the 
		given multi-select element.
		@param element The ID or element object for a multi-select control
	*/
	select.getNumChecked = function(element)
	{
		element = Doc.element(element);
		if (!element.options) return 0;
		var numChecked = 0;
		for (var i = 0; i < element.options.length; i++)
		{
			if (element.options[i].checked)
			{
				numChecked++;
			}
		}
		return numChecked;
	}
	
	/**
		Check the given option. The filter display is
		automatically updated.
		@param option The option element to check
	*/
	select.checkOption = function(option)
	{
		if (!Data.int(activeSelect, 'multiple'))
		{
			for (var i = 0; i < activeSelect.options.length; i++)
			{
				activeSelect.options[i].checked = false;
				Doc.removeClass(activeSelect.options[i].div, 'checked');
			}
		}
		option.option.checked = true;
		Doc.addClass(option, 'checked');
		select.updateFilter(activeSelect);
	}
	
	/**
		Uncheck the given option. The filter display is
		automatically updated.
		@param option The option element to uncheck
	*/
	select.uncheckOption = function(option)
	{
		option.option.checked = false;
		Doc.removeClass(option, 'checked');
		select.updateFilter(activeSelect);
	}
	
	/**
		Toggle the given option. The filter display is
		automatically updated.
		@param option The option element to check
	*/
	select.toggleOption = function(option)
	{
		if (!option.option.checked)
		{
			select.checkOption(option);
		}
		else
		{
			select.uncheckOption(option);
		}
	}

	/**
		Event handler when clicking on the check/uncheck all icon.
	*/
	select.checkUncheckAllOnClick = function(event)
	{
		var specialKey = event.altKey || event.ctrlKey || event.shiftKey;
		if (specialKey)
		{
			select.checkUncheckAll();
		}
		else
		{
			select.toggleAll();
		}
	}

	
	/*
		Check or uncheck all options. If forceCheck is true/false
		this value is used, otherwise the status of the first
		option is reversed and all other options are set to the
		same value.
	*/
	select.checkUncheckAll = function(check)
	{
		if (!activeSelect) return;
		
		var options = activeSelect.options;
		if (check != false && check != true) check = !options[0].checked;
		for (var i = 0; i < options.length; i++)
		{
			options[i].checked = check;
			Doc.addOrRemoveClass(options[i].div, 'checked', check);
		}
		select.updateFilter(activeSelect);
	}

	/*
		Toggle all options.
	*/
	select.toggleAll = function()
	{
		if (!activeSelect) return;
		
		var options = activeSelect.options;
		for (var i = 0; i < options.length; i++)
		{
			options[i].checked = !options[i].checked;
			Doc.addOrRemoveClass(options[i].div, 'checked', options[i].checked);
		}
		select.updateFilter(activeSelect);
	}

	
	/**
		Update the visible and hidden filter values for
		the given multi-select element.
		@param element The ID or element object for a multi-select control
	*/
	select.updateFilter = function(element)
	{
		element = Doc.element(element);
		if (!element.options) return;
		
		var keys = [];
		var values = [];
		
		for (var i = 0; i < element.options.length; i++)
		{
			var option = element.options[i];
			if (option.checked)
			{
				keys[keys.length] = option.key;
				var ii = option.value.indexOf('<');
				values[values.length] = Strings.decodeTags(ii < 0 ? option.value : option.value.substring(0, ii).trim());
			}
		}
		if (element.infoDiv)
		{
			element.infoDiv.innerHTML = values.length + ' selected';
		}
		
		var prefix = element.notEqual && values.length > 0 ? '<>' : '';
		element.displayField.value = prefix+values.join('|');
		element.hiddenField.value = prefix+keys.join('|');
		// Special case when an empty option is selected
		if (values.length > 0 && element.hiddenField.value == '')
		{
			element.hiddenField.value = '=';
		}
	}
	
	/**
		Highlight the given option as if the mouse is over it
		or if the keyboard arrows have been used.
	*/
	select.highlightOption = function(option, autoScroll)
	{
		if (highlightedOption)
		{
			Doc.removeClass(highlightedOption, 'active');
		}
		highlightedOption = option;
		if (highlightedOption)
		{
			Doc.addClass(highlightedOption, 'active');
			if (autoScroll) select.scrollIntoView(highlightedOption);
		}
	}
	
	/**
		Scroll the list so the given option is visible. 
		'alignFraction' can be set between 0 and 1 and is the fraction
		between top/bottom alignment. If not set, the alignment is
		auto to top or bottom.
	*/
	select.scrollIntoView = function(option, alignFraction)
	{
		var parent = option.parentNode;
		var parentTop = parent.scrollTop;
		var optionTop = option.offsetTop-parent.offsetTop;
		var optionBottom = optionTop + option.offsetHeight;
		var parentBottom = parentTop + parent.offsetHeight;
		
		var scrollToCenter = alignFraction != undefined ? optionTop - alignFraction*(parent.offsetHeight-option.offsetHeight) : 0;
		if (optionTop < parentTop)
		{
			parent.scrollTop = scrollToCenter || optionTop;
		}
		else if (optionBottom > parentBottom)
		{
			parent.scrollTop = scrollToCenter || (optionBottom - parent.offsetHeight);
		}
	}
	
	/**
		Set the given multi-select element as the active element.
		If the currently active element is a different element, it
		is automatically hidden.
	*/
	select.setActive = function(element)
	{
		element = Doc.element(element);
		if (element == activeSelect) return;
		
		// Get the dropdown div element
		var dropdownDiv = element.dropdownDiv;
		if (!dropdownDiv)
		{
			dropdownDiv = internal.createDropdownDiv(element);
		}

		// Position the dropdown
		var pos = select.getFixedDropDownPosition(element);
		dropdownDiv.style.left = pos.left;
		dropdownDiv.style.right = pos.right;
		dropdownDiv.style.width = pos.width;
		dropdownDiv.style.minWidth = pos.minWidth;
		dropdownDiv.style.maxWidth = pos.maxWidth;

		// Show it
		Doc.addClass(element, 'active');
		dropdownDiv.style.display = 'block';
		activeSelect = element;
		element.displayField.focus();
		// Need timeout since positions seems to not be correct until the dropdown has been draw on the screen
		setTimeout(function(){select.scrollToFirstChecked(element)}, 125);
	}
	
	/**
		Calculate left and width for the dropdown 
		given the filter element.
	*/
	select.getFixedDropDownPosition = function(element)
	{
		// Position the dropdown div relative the  main element
		var fpos = Doc.getElementPosition(element);

		var dpos = {};

		// The default is to make the drop-down the same size as the filter
		dpos.left = (fpos.left+1)+'px';
		dpos.width = (fpos.width)+'px';
		dpos.right = 'auto';
		dpos.minWidth = (Math.max(fpos.width, 150))+'px';
		dpos.maxWidth = 'auto';
		
		if (fpos.width < 300)
		{
			// If the width is smaller than 300 we should open up either
			// to the left or to the right, we need to check how close we are to
			// the right side
			try
			{
				// Now we must check if we end up outside the right side of
				// the browser window, find the parent <div class="data"> element
				// or document.body
				var p = element.parentNode;
				while (p != null && p.tagName != 'DIV' && p.className != 'data')
				{
					p = p.parentNode;
				}
				if (p==null) p = document.body;

				// Get position data for the parent
				// +16 should be safe to account for a scrollbar
				var ppos = Doc.getElementPosition(p);
				var scrollbar = p.scrollHeight > p.clientHeight ? 16 : 0;
				var maxRight = ppos.width + p.scrollLeft - scrollbar;

				dpos.width = 'auto';
				dpos.maxWidth = '300px';
				if (fpos.left + 300 > maxRight)
				{
					// Align to right
					dpos.left = 'auto';
					dpos.right = (ppos.right-fpos.right-scrollbar)+'px';
					if (fpos.right > maxRight)
					{
						// And adjust scroll position if needed
						p.scrollLeft += fpos.right - maxRight;
					}
				}
			}
			catch (err)
			{
				var msg = err;
				if (err.toString)
				{
					msg = err.toString() + '\n' + err.stack;
				}
				console.error(msg);
			}
		}
		return dpos;
	}
	
	/**
		Scroll the selection list so that the first selected element is
		visible and aligned about 1/4 between top and bottom.
	*/
	select.scrollToFirstChecked = function(element)
	{
		for (var i = 0; i < element.options.length; i++)
		{
			var opt = element.options[i];
			if (opt.checked) 
			{
				select.scrollIntoView(opt.div, 0.25);
				return;
			}
		}
	}
	
	/**
		Hide the currently displayed dropdown (if any)
	*/
	select.hideDropdown = function()
	{
		if (!activeSelect) return;
		select.highlightOption(null);
		Doc.removeClass(activeSelect, 'active');
		activeSelect.dropdownDiv.style.display = 'none';
		activeSelect = null;
	}
	
	/**
		Submit the form.
	*/
	select.submit = function()
	{
		if (!activeSelect) return;
		activeSelect.displayField.form.submit();
	}
	
	/**
		Clear this filter.
	*/
	select.clear = function()
	{
		if (!activeSelect) return;
		select.checkUncheckAll(false);
		select.submit();
	}
	
	/**
		Initialize a multi-select enabled field.
	 */
	internal.initializeMultiSelect = function(element, autoInit)
	{
		if (autoInit != 'multiselect') return;
		
		// Get the child elements that are part of the multi-select control
		var hidden = element.getElementsByClassName("multiselect-hidden")[0];
		var display = element.getElementsByClassName("multiselect-display")[0];
		var dropdown = element.getElementsByClassName("dropdown")[0];

		Doc.addClass(element, 'interactable');
		
		// Store cross-references between the main and child elements
		element.hiddenField = hidden;
		element.displayField = display;
		element.dropdown = dropdown;
		hidden.multiSelect = element;
		display.multiSelect = element;
		dropdown.multiSelect = element;
		// Disable auto-complete on the input field
		display.setAttribute('autocomplete', 'off');
		
		// Load the key-value options from the data-options attribute
		element.options = Data.json(element, 'options');
		// 
		element.notEqual = hidden.value.indexOf('<>') == 0;
		// Show the current filter
		select.updateFilter(element);
		
		// Add event handlers to the 'display' input field
		Events.addEventHandler(display, 'focus', internal.onFocus);
		Events.addEventHandler(display, 'blur', internal.onBlur);
		Events.addEventHandler(display, 'keydown', internal.onKeyDown);
		
		// Add event handler to the 'dropdown' icon
		Events.addEventHandler(dropdown, 'click', internal.dropdownOnClick);
		
		// Ensure that the options list stay visible when the 
		// mouse is inside the active multi-select control
		Events.addEventHandler(element, 'mouseenter', internal.setHideFlag);
		Events.addEventHandler(element, 'mouseleave', internal.setHideFlag);
	}
	Doc.addElementInitializer(internal.initializeMultiSelect);
	
	/**
		Create the dropdown div with all options for the given
		multi-select element.
	*/
	internal.createDropdownDiv = function(element)
	{
		var imgRoot = App.getRoot()+'images/';
		
		// Create the control panel
		var controlDiv = document.createElement('div');
		controlDiv.className = 'multioptioncontrol bg-filled-100';

		// The 'submit' icon
		var submitIcon = document.createElement('span');
		submitIcon.className = 'link submit';
		submitIcon.innerHTML = '<img src="' + imgRoot + 'ok.png">';
		submitIcon.title = 'Submit filter and update table';
		controlDiv.appendChild(submitIcon);
		Events.addEventHandler(submitIcon, 'click', select.submit);

		// The 'clear' icon
		var clearIcon = document.createElement('span');
		clearIcon.className = 'link clear';
		clearIcon.innerHTML = '<img src="' + imgRoot + 'remove.png">';
		clearIcon.title = 'Clear filter and update table';
		controlDiv.appendChild(clearIcon);
		Events.addEventHandler(clearIcon, 'click', select.clear);

		// The '=/≠' icon for toggling between equals/not-equals comparison
		var neqDiv = document.createElement('img');
		neqDiv.className = 'link equal-or-not-equal';
		neqDiv.src = imgRoot+(element.notEqual ? 'not-equal.png' : 'equal.png');
		neqDiv.title = 'Toggle between equals and not-equals comparison';
		controlDiv.appendChild(neqDiv);
		Events.addEventHandler(neqDiv, 'click', internal.notEqualOnClick);

		var multiple = Data.int(element, 'multiple');		
		if (multiple)
		{
			// The 'info' section
			var infoDiv = document.createElement('span');
			infoDiv.className = 'info link';
			infoDiv.innerHTML = select.getNumChecked(element)+' selected';
			infoDiv.title = 'Toggle selection (use CTRL, ALT or SHIFT to select/deselect all)';
			controlDiv.appendChild(infoDiv);
			Events.addEventHandler(infoDiv, 'click', select.checkUncheckAllOnClick);
			element.infoDiv = infoDiv;
		}

		// Create the list with options
		var options = element.options;
		var allOptionsDiv = document.createElement('div');
		allOptionsDiv.className = 'multioptionoptions';
		
		for (var i = 0; i < options.length; i++)
		{
			var optionDiv = document.createElement('div');
			optionDiv.className = 'multioption interactable' + (options[i].checked ? ' checked' : '');
			optionDiv.option = options[i];
			optionDiv.innerHTML = options[i].value;
			options[i].div = optionDiv;
			optionDiv.title = options[i].value;
			allOptionsDiv.appendChild(optionDiv);
			Events.addEventHandler(optionDiv, 'click', internal.optionOnClick);
			Events.addEventHandler(optionDiv, 'mouseover', internal.optionHighlight);
		}
		
		// Main div
		var mainDiv = document.createElement('div');
		mainDiv.className = 'multioptions ' + (multiple ? 'multiple' : 'single');
		mainDiv.appendChild(controlDiv);
		mainDiv.appendChild(allOptionsDiv);
		
		Events.addEventHandler(mainDiv, 'click', internal.ensureFocused);
		
		element.dropdownDiv = mainDiv;
		element.appendChild(mainDiv);
		return mainDiv;
	}
	
	/**
		Event handler for the 'dropdown' icon. This should toggle the options
		list.
	*/
	internal.dropdownOnClick = function(event)
	{
		var multiSelect = event.currentTarget.multiSelect;
		if (multiSelect == activeSelect)
		{
			// Hide the options
			select.hideDropdown();
		}
		else
		{
			// Show the new options
			select.setActive(multiSelect);
			disableHide = true;
		}
	}
	
	/**
		Event handler used when a multi-select enabled field receives
		the focus. This should set the current field to the active
		field and display the options dropdown.
	*/
	internal.onFocus = function(event)
	{
		select.setActive(event.target.multiSelect);
		disableHide = true;
	}
	
	/*
		Event handler used when a multi-select enabled field
		loses the focus. This should hide the options and set
		the current field to null. We must use the delayed version 
		since the blur may occur due to a user clicking on an option
		in the dropdown in which case the field should regain focus and
		stay displayed.
	*/
	internal.onBlur = function(event)
	{
		if (disableHide) return;
		select.hideDropdown();
	}

	/*
		Called when a key is pressed in a multi-select enabled field.
		This function detects:
		* up/down arrow: moves the selected option
		* space: toggles the selected status of the current option
		* enter: toggles the current option and submits the form
		* tab: move to the next focusable field
	*/
	internal.onKeyDown = function(event)
	{
		var keyCode = event.keyCode;
		var preventDefault = true;
		if (keyCode == 38) // up
		{
			if (highlightedOption && highlightedOption.previousSibling)
			{
				// highlight the previous option
				select.highlightOption(highlightedOption.previousSibling, true);
			}
		}
		else if (keyCode == 40) // down
		{
			if (!highlightedOption)
			{
				// highlight the first option
				select.highlightOption(event.target.multiSelect.options[0].div, true);
			}
			else if (highlightedOption.nextSibling)
			{
				// highlight the next option
				select.highlightOption(highlightedOption.nextSibling, true);
			}
		}
		else if (keyCode == 32) // space
		{
			// SPACE key should toggle the highlighted option
			if (highlightedOption) 
			{
				select.toggleOption(highlightedOption);
			}
		}
		else if (keyCode == 13) // enter
		{
			// ENTER key should toggle the highlighted option
			// and submit the form
			if (highlightedOption) 
			{
				select.toggleOption(highlightedOption);
			}
			select.submit();
		}
		else if (keyCode == 9) // tab
		{
			// Enable TAB key to move to next field
			disableHide = false;
			preventDefault = false;
		}
		if (preventDefault)
		{
			event.preventDefault();
		}
	}
	
	internal.notEqualOnClick = function(event)
	{
		if (!activeSelect) return;
		
		var imgRoot = App.getRoot()+'images/';
		var current = activeSelect.notEqual || false;
		activeSelect.notEqual = !current;
		event.currentTarget.src = imgRoot+(activeSelect.notEqual ? 'not-equal.png' : 'equal.png');
		select.updateFilter(activeSelect);
	}
	
	/**
		Event handler that react to click events on individual options
	*/
	internal.optionOnClick = function(event)
	{
		var option = event.currentTarget;
		select.toggleOption(option);
	}
	
	/**
		Highlight the target option when the mouse is over the
		option and remove the highlight when the mouse is outside.
	*/
	internal.optionHighlight = function(event)
	{
		var option = event.currentTarget;
		select.highlightOption(option);
	}
	
	/**
		Ensure that the active multi-select's display field has the focus.
	*/
	internal.ensureFocused = function(event)
	{
		if (!activeSelect) return;
		activeSelect.displayField.focus();
	}
	
	/**
		Disable automatic hiding of the options list when the mouse
		is entering a multi-select control.
	*/
	internal.setHideFlag = function(event)
	{
		disableHide = event.type == 'mouseenter' && (event.currentTarget == activeSelect || !activeSelect);
	}

	internal.initMultiselect = function()
	{
		window.addEventListener('resize', internal.repositionActiveElement);
	}
	
	/**
		Re-position active element when the window has been resized.
	*/
	internal.repositionActiveElement = function()
	{
		if (!activeSelect) return;
		var dropdownDiv = activeSelect.dropdownDiv;
		if (!dropdownDiv) return;

		// Position the dropdown
		var pos = select.getFixedDropDownPosition(activeSelect);
		dropdownDiv.style.left = pos.left;
		dropdownDiv.style.right = pos.right;
		dropdownDiv.style.width = pos.width;
		dropdownDiv.style.minWidth = pos.minWidth;
		dropdownDiv.style.maxWidth = pos.maxWidth;
	}
	
	Doc.onLoad(internal.initMultiselect);
	
	return select;
}();

var SmartEnum = function()
{
	var smartenum = {};
	var internal = {};

	// The currently active smart-enum element
	var activeEnum = null;
	// Flag to indicate that automatic hiding of the options list 
	// should be disabled due to the user is interacting with the list
	var disableHide = false;
	// The option that is currently higlighted
	var highlightedOption = null;

	
	/**
		Set the given smart-enum element as the active element.
		If the currently active element is a different element, it
		is automatically hidden.
	*/
	smartenum.setActive = function(element)
	{
		element = Doc.element(element);
		if (element == activeEnum) return;
		
		// Get the dropdown div element
		var dropdownDiv = element.dropdownDiv;
		if (!dropdownDiv)
		{
			dropdownDiv = internal.createDropdownDiv(element);
			// Position the dropdown div relative the  main element
			var pos = Doc.getElementPosition(element);
			dropdownDiv.style.width = ((pos.width < 150 ? 150 : pos.width))+'px';
		}
	
		// Position the dropdown
		var pos = MultiSelect.getFixedDropDownPosition(element);
		dropdownDiv.style.left = pos.left;
		dropdownDiv.style.right = pos.right;
		dropdownDiv.style.width = pos.width;
		dropdownDiv.style.minWidth = pos.minWidth;
		dropdownDiv.style.maxWidth = pos.maxWidth;
		
		// Show it
		Doc.addClass(element, 'active');
		dropdownDiv.style.display = 'block';
		activeEnum = element;
		element.displayField.focus();
	}

	
	/**
		Highlight the given option as if the mouse is over it
		or if the keyboard arrows have been used.
	*/
	smartenum.highlightOption = function(option)
	{
		if (!activeEnum) return;
		if (highlightedOption)
		{
			Doc.removeClass(highlightedOption, 'active');
		}
		highlightedOption = option;
		if (highlightedOption)
		{
			Doc.addClass(highlightedOption, 'active');
			// Scroll the highlighted option into view
			var parent = option.parentNode;
			var parentTop = parent.scrollTop;
			var optionTop = option.offsetTop;
			var optionBottom = optionTop + option.offsetHeight;
			var parentBottom = parentTop + parent.offsetHeight;
			if (optionTop < parentTop)
			{
				parent.scrollTop = optionTop;
			}
			else if (optionBottom > parentBottom)
			{
				parent.scrollTop = optionBottom - parent.offsetHeight + 2;
			}
		}
	}

	/**
		Higlight the option that is the next visible one below the given
		option. If there is no such option, the currently highlighted option
		is kept.
	*/
	smartenum.highlightNext = function(option)
	{
		if (!activeEnum) return;
		option = option != null ? option.nextSibling : activeEnum.dropdownDiv.firstChild;
		while (option && option.style.display == 'none')
		{
			option = option.nextSibling;
		}
		if (option) smartenum.highlightOption(option);
	}
	
	/**
		Higlight the option that is the previous visible one above the given
		option. If there is no such option, the currently highlighted option
		is kept.
	*/
	smartenum.highlightPrevious = function(option)
	{
		if (!activeEnum) return;
		option = option != null ? option.previousSibling : activeEnum.dropdownDiv.lastChild;
		while (option && option.style.display == 'none')
		{
			option = option.previousSibling;
		}
		if (option) smartenum.highlightOption(option);
	}

	/**
		Filter the options in the drop-down list depending on the
		text entered in the display field. Operator prefixes (eg. =, >, ...)
		are ignored.
	*/
	smartenum.filterOptions = function()
	{
		if (!activeEnum) return;
		if (activeEnum.displayField.value == activeEnum.displayField.lastValue) 
		{
			// If the value has not change, there is no need to re-filter the options
			return;
		}
		activeEnum.displayField.lastValue = activeEnum.displayField.value;
		// Strip away operator prefix and split at '%'
		var text = activeEnum.displayField.value.replace(/^[=!<>]*/, '').split('%');
		for (var i = 0; i < activeEnum.dropdownDiv.childNodes.length; i++)
		{
			var opt = activeEnum.dropdownDiv.childNodes[i];
			var optText = opt.innerHTML;
			var startIndex = 0;
			var match = true;
			// Check that each text part is matching...
			for (var j = 0; j < text.length && match; j++)
			{
				// ... and each piece must match AFTER the other match
				var index = optText.indexOf(text[j], startIndex);
				if (index == -1) match = false;
				startIndex = index+text[j].length+1;
			}
			opt.style.display = match ? 'block' : 'none';
		}
		
		if (highlightedOption != null && highlightedOption.style.display == 'none')
		{
			smartenum.highlightOption(null);
		}
	}
	
	/**
		Hide the currently displayed dropdown (if any)
	*/
	smartenum.hideDropdown = function()
	{
		if (!activeEnum) return;
		Doc.removeClass(activeEnum, 'active');
		smartenum.highlightOption(null);
		activeEnum.dropdownDiv.style.display = 'none';
		activeEnum = null;
	}

	/**
		Set the given option as the filter text and then submit the form.
	*/
	smartenum.submitOption = function(option)
	{
		if (!activeEnum) return;
		var field = activeEnum.displayField;
		var text = field.value;
		if (option)
		{
			field.value = text.replace(/^([=!<>]*).*/, '$1' + option.innerHTML);
		}
		field.form.submit();
	}
	
	/**
		Initialize a smart-enum enabled field.
	 */
	internal.initializeSmartEnum = function(element, autoInit)
	{
		if (autoInit != 'smartenum') return;

		// Get the child elements that are part of the multi-select control
		var display = element.getElementsByClassName("smartenum-display")[0];
		var dropdown = element.getElementsByClassName("dropdown")[0];
	
		Doc.addClass(element, 'interactable');
		
		// Store cross-references between the main and child elements
		element.displayField = display;
		element.dropdown = dropdown;
		display.smartEnum = element;
		dropdown.smartEnum = element;
		// Disable auto-complete on the input field
		display.setAttribute('autocomplete', 'off');
		
		// Load the options array from the data-options attribute
		element.options = Data.json(element, 'options').sort(Strings.compareIgnoreCase);
		
		// Add event handlers to the 'display' input field
		Events.addEventHandler(display, 'focus', internal.onFocus);
		Events.addEventHandler(display, 'blur', internal.onBlur);
		Events.addEventHandler(display, 'keydown', internal.onKeyDown);
		Events.addEventHandler(display, 'keyup', internal.onKeyUp);
		
		// Add event handler to the 'dropdown' icon
		Events.addEventHandler(dropdown, 'click', internal.dropdownOnClick);
		
		// Ensure that the options list stay visible when the 
		// mouse is inside the active smart-enum control
		Events.addEventHandler(element, 'mouseenter', internal.setHideFlag);
		Events.addEventHandler(element, 'mouseleave', internal.setHideFlag);
	}
	Doc.addElementInitializer(internal.initializeSmartEnum);

	
	/**
		Event handler for the 'dropdown' icon. This should toggle the options
		list.
	*/
	internal.dropdownOnClick = function(event)
	{
		var smartEnum = event.currentTarget.smartEnum;
		if (smartEnum == activeEnum)
		{
			// Hide the options
			smartenum.hideDropdown();
		}
		else
		{
			// Show the new options
			smartenum.setActive(smartEnum);
			disableHide = true;
		}
	}

	/**
		Create the dropdown div with all options for the given
		smart-enum element.
	*/
	internal.createDropdownDiv = function(element)
	{
		// Create the list with options
		var options = element.options;

		// Main div
		var mainDiv = document.createElement('div');
		mainDiv.className = 'smartenum';

		for (var i = 0; i < options.length; i++)
		{
			var optionDiv = document.createElement('div');
			optionDiv.className = 'smartoption interactable';
			optionDiv.optionIndex = i;
			optionDiv.innerHTML = options[i];
			mainDiv.appendChild(optionDiv);
		}
		
		Events.addEventHandler(mainDiv, 'click', internal.ensureFocused);
		Events.addEventHandler(mainDiv, 'click', internal.optionOnClick);
		Events.addEventHandler(mainDiv, 'mouseover', internal.optionHighlight);
		
		element.dropdownDiv = mainDiv;
		element.appendChild(mainDiv);
		
		return mainDiv;
	}

	/**
		Event handler used when a smart-enum enabled field receives
		the focus. This should set the current field to the active
		field.
	*/
	internal.onFocus = function(event)
	{
		smartenum.setActive(event.target.smartEnum);
		disableHide = true;
	}
	
	/**
		Event handler used when a multi-select enabled field
		loses the focus. This should hide the options and set
		the current field to null.
	*/
	internal.onBlur = function(event)
	{
		if (disableHide) return;
		smartenum.hideDropdown();
	}

	/**
		Called when a key is pressed in a smart-enum enabled field.
		This function detects:
		* up/down arrow: moves the highlighted option
		* enter: submits the form
	*/
	internal.onKeyDown = function(event)
	{
		if (!activeEnum) smartenum.setActive(event.target.smartEnum);
		var keyCode = event.keyCode	
		if (keyCode == 38) // up
		{
			smartenum.highlightPrevious(highlightedOption);
			event.preventDefault();
		}
		else if (keyCode == 40) // down
		{
			smartenum.highlightNext(highlightedOption);
			event.preventDefault();
		}
		else if (keyCode == 13) // enter
		{
			// ENTER key should select the highlighted option and submit the form
			smartenum.submitOption(highlightedOption);
		}
		else if (keyCode == 27) // escape
		{
			smartenum.hideDropdown();
		}

	}

	
	/**
		Called when a key is released in a smart-enum enabled field. 
		This function detects changes to the text and displays matching
		options.
	*/
	internal.onKeyUp = function(event)
	{
		smartenum.filterOptions();
	}

	
	/**
		Highlight the target option when the mouse is over the
		option and remove the highlight when the mouse is outside.
	*/
	internal.optionHighlight = function(event)
	{
		var option = event.target;
		if (option.optionIndex >= 0)
		{
			smartenum.highlightOption(option);
		}
	}

	/**
		Event handler that react to click events on individual options
	*/
	internal.optionOnClick = function(event)
	{
		var option = event.target;
		if (option.optionIndex >= 0)
		{
			smartenum.submitOption(option);
		}
	}
	
	/**
		Disable automatic hiding of the options list when the mouse
		is entering a smart-enum control.
	*/
	internal.setHideFlag = function(event)
	{
		disableHide = event.type == 'mouseenter' && (event.currentTarget == activeEnum || !activeEnum);
	}

	internal.initEnum = function()
	{
		window.addEventListener('resize', internal.repositionActiveElement);
	}
	
	/**
		Re-position active element when the window has been resized.
	*/
	internal.repositionActiveElement = function()
	{
		if (!activeEnum) return;
		var dropdownDiv = activeEnum.dropdownDiv;
		if (!dropdownDiv) return;

		// Position the dropdown
		var pos = MultiSelect.getFixedDropDownPosition(activeEnum);
		dropdownDiv.style.left = pos.left;
		dropdownDiv.style.right = pos.right;
		dropdownDiv.style.width = pos.width;
		dropdownDiv.style.minWidth = pos.minWidth;
		dropdownDiv.style.maxWidth = pos.maxWidth;
	}
	
	Doc.onLoad(internal.initEnum);

	
	return smartenum;
}();

/**
	Handles div elements with longs texts that the user may want to 
	only display a part of until clicking with the mouse to reveal the
	complete contents. To begin with the div elements are marked with
	the class name 'constrained autoshow'. By itself, this trigger automatic
	showing/hiding on mouse hover. A scan and check is required to see if
	there is an actual text overflow and instead attach showing/hiding
	to mouse click events. 
*/
var TextOverflow = function()
{
	var overflow = {};
	var internal = {};
	
	/**
		Get the text overflow method currently in use. There are
		three possible values:
		* display (the default): overflowed text is displayed
		* hover: overflowed text is displayed on mouse hover
		* click: overflowed text is displayed when the mouse is clicked
	*/
	overflow.getOverflowMethod = function()
	{
		return Data.get(document.body, 'text-overflow', 'display');
	}
	
	/**
		Check all tags with class name 'constrained' if they 
		have overflowed text. If so, attach a mouse click
		handler to the element that show/hide the overflow text.
	*/
	overflow.checkAndFixAll = function()
	{
		var tags = document.getElementsByClassName('constrained');
		for (var i = 0; i < tags.length; i++)
		{
			overflow.checkTag(tags[i]);
		}
		for (var i = 0; i < tags.length; i++)
		{
			overflow.fixTag(tags[i]);
		}
	}

	/**
		Check the given tag for overflowed text and attach a
		mouse click handler if needed.
	*/
	overflow.checkAndFixTag = function(tag)
	{
		overflow.checkTag(tag);
		overflow.fixTag(tag);
	}
	
	/**
		Check if the given tag has overflowed text of not.
		The result is stored as 3 properties on the tag element
		* tag.isWidthOverflowed
		* tag.isHeightOverflowed
		* tag.isOverflowed
	*/
	overflow.checkTag = function(tag)
	{
		var isWidthOverflowed = tag.clientWidth < tag.scrollWidth;
		var isHeightOverflowed = tag.clientHeight + 5 < tag.scrollHeight;
		tag.isWidthOverflowed = isWidthOverflowed;
		tag.isHeightOverflowed = isHeightOverflowed;
		tag.isOverflowed = isWidthOverflowed || isHeightOverflowed;
	}
	
	/**
		Fix the (possible) overflowed tag and attach a mouse click
		handler to it if needed.
	*/
	overflow.fixTag = function(tag)
	{
		var className = 'constrained';
		if (tag.isOverflowed)
		{
			className += ' overflowed manualshow';
			Events.addEventHandler(tag, 'click', overflow.overflowedOnClick);
			tag.title = 'Click to show/hide the complete text!';
		}
		tag.className = className;
	}

	/**
		Event handler that toggle the visiblity status of the overflowed
		text.
	*/
	overflow.overflowedOnClick = function(event)
	{
		var tag = event.currentTarget;
		if (tag.isWidthOverflowed)
		{
			Doc.addOrRemoveClass(tag, 'constrained');
		}
		else
		{
			Doc.addOrRemoveClass(tag, 'shown');
		}
	}

	internal.initCheck = function()
	{
		if (overflow.getOverflowMethod() == 'click')
		{
			setTimeout(overflow.checkAndFixAll, 200);
		}
	};
	Doc.addFinalizer(internal.initCheck);
	
	return overflow;
}();

var HideableSection = function()
{
	var section = {};
	var internal = {};

	// Show/hide a section 
	section.toggle = function(sectionDiv)
	{
		sectionDiv = Doc.element(sectionDiv);
		var icon = Doc.element(sectionDiv.id+'.icon');
		var content = Doc.element(sectionDiv.id+'.content');
		var showClass = Data.get(sectionDiv, 'showclass');
		var hideClass = Data.get(sectionDiv, 'hideclass');
		
		var visible = content.style.display != 'none';
		if (visible)
		{
			// Hide the section
			content.style.display = 'none';
			icon.src = App.getRoot() + 'images/show_section.png';
			visible = false;
			if (showClass) Doc.removeClass(main, showClass);
			if (hideClass) Doc.addClass(main, hideClass);
		}
		else
		{
			// Show the section
			content.style.display = 'block';
			icon.src = App.getRoot() + 'images/hide_section.png';
			visible = true;
			if (showClass) Doc.addClass(main, showClass);
			if (hideClass) Doc.removeClass(main, hideClass);
		}
		return visible;
	}
	
	/**
		Show/hide a section and send an Ajax request to the server with
		the new status. If the browser doesn't support Ajax, this method is
		identical to toggle()
	*/
	section.toggleAndSendWithAjax = function(sectionDiv, itemType, subcontext)
	{
		var visible = section.toggle(sectionDiv);
		var request = Ajax.getXmlHttpRequest();
		if (request != null)
		{
			var url = App.getRoot()+'common/store_show_hide_section.jsp?ID='+App.getSessionId();
			url += '&sectionId='+sectionDiv.id;
			url += '&itemType='+itemType;
			url += '&subcontext='+subcontext;
			url += '&visible='+(visible ? '1' : '0');
			request.open("GET", url, true);
			request.send(null);
		}
	}

	/**
		Add click event handlers to the sections.
	*/
	internal.initHideableSection = function(element, autoInit)
	{
		if (autoInit != 'hideable-section') return;
		Events.addEventHandler(element, 'click', internal.onClick);
	}
	Doc.addElementInitializer(internal.initHideableSection);
	
	/**
		Click event handler for a section. It will toggle
		the visibility of the content part and, if a context has
		been specified, save the current setting.
	*/
	internal.onClick = function(event)
	{
		var target = event.currentTarget;
		var sectionDiv = Doc.element(target.id.substring(0, target.id.indexOf('.')));
		var context = Data.get(sectionDiv, 'context');
		var subcontext = Data.get(sectionDiv, 'subcontext', '');
		if (context)
		{
			section.toggleAndSendWithAjax(sectionDiv, context, subcontext);
		}
		else
		{
			section.toggle(sectionDiv);
		}
	}
	
	return section;
}();

var Ajax = function()
{
	var ajax = {};
	var internal = {};
	
	ajax.getXmlHttpRequest = function()
	{
		return new XMLHttpRequest();
	}
	
	/**
		Register callback functions for a successful of failed request.
		When the request has been completed, the onSuccessHandler or
		onFailHandler will be called with the request as a parameter:
		onSuccessHandler(request);
	*/
	ajax.setReadyStateHandler = function(request, onSuccessHandler, onFailHandler)
	{
		request.addEventListener('readystatechange', function()
			{
				internal.onReadyStateChange(request, onSuccessHandler, onFailHandler);
			}
		);
	}

	internal.onReadyStateChange = function(request, onSuccessHandler, onFailHandler)
	{
		if (request.readyState == 4)
		{
			if (request.status == 200 && onSuccessHandler)
			{
				onSuccessHandler.call(null, request);
			}
			else if (onFailHandler)
			{
				onFailHandler.call(null, request);
			}
		}
	}
	
	return ajax;
}();

// Place for storing temporary objects
window.pageValues = [];

