/* $Id: menu.js 7419 2017-11-03 08:57:18Z nicklas $
	------------------------------------------------------------------
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
	------------------------------------------------------------------

	JavaScript functions for the Menu taglib.

	@author Nicklas
	@version 2.0
*/
'use strict';

var Menu = function()
{
	var menu = {};
	var internal = {};
	var visibleMenus = [];
	
	/**
		Show a top-level menu at the given position.
		@param menuDiv A menu object or ID
		@param x,y The pixel coordinates for the top left/right corner of the menu
		@param alignRight If set the x coordinate is counted from the right side of
			the window and the right side of the menu is aligned to that pixel
	*/
	menu.showTopMenu = function(menuDiv, x, y, alignRight)
	{
		menu.hideAll();
		menuDiv = Doc.element(menuDiv);
		if (alignRight)
		{
			menuDiv.style.right = (x)+'px';
			menuDiv.style.left = 'auto';
		}
		else
		{
			menuDiv.style.left = (x)+'px';
			menuDiv.style.right = 'auto';
		}
		menuDiv.style.top = (y)+'px';
		menuDiv.style.display = 'block';
		menuDiv.style.zIndex = 999;
		menuDiv.canHide = true;
		internal.registerVisible(menuDiv);
	}
	
	/**
		Toggle a top-level menu. If the menu is not visible it is shown,
		otherwise it is hidden.
		@param menuDiv A menu object or ID
		@param x,y The pixel coordinates for the top left corner of the menu
	*/
	menu.toggleTopMenu = function(menuDiv, x, y, alignRight)
	{
		menuDiv = Doc.element(menuDiv);
		if (menuDiv.style.display == 'none')
		{
			menu.showTopMenu(menuDiv, x, y, alignRight);
		}
		else
		{
			menu.hideSubMenus(menuDiv, true);
		}
	}

	/**
		Hide all visible menus and their submenus.
	*/
	menu.hideAll = function()
	{
		var i = 0;
		while (visibleMenus.length > 0 && i < 10)
		{
			menu.hideSubMenus(visibleMenus[0], true);
			i++;
		}
	}
	
	/**
		Get the menu that the given menu-item belongs to.
	*/
	menu.getMenuOfMenuItem = function(menuItem)
	{
		menuItem = Doc.element(menuItem);
		if (!menuItem.menu)
		{
			var parent = menuItem.parentNode;
			while (parent.tagName.toLowerCase() != 'div')
			{
				parent = parent.parentNode;
			}
			menuItem.menu = parent;
		}
		return menuItem.menu;
	}
	
	/**
		Show a submenu.
		@param menuItem The ID or document element of the menu-item that is linked with the submenu
		@param subMenu The submenu object or ID to display
	*/
	menu.showSubMenu = function(menuItem, subMenu)
	{
		menuItem = Doc.element(menuItem);
		subMenu = Doc.element(subMenu);
		var menuDiv = menu.getMenuOfMenuItem(menuItem);
		menu.hideSubMenus(menuDiv);
		var pos = Doc.getElementPosition(menuItem);
		var vertical = Data.int(menuDiv, 'vertical');
		var dx = vertical ? pos.width : -1; // place submenu to the right of the parent menu
		var dy = vertical ? 0 : pos.height; // place submenu below the parent menu
		subMenu.style.left = (pos.left+dx)+'px';
		subMenu.style.right = 'auto';
		subMenu.style.top = (pos.top+dy)+'px';
		subMenu.style.display = 'block';
		subMenu.style.zIndex = menuDiv.style.zIndex+1;
		subMenu.canHide = true;
		menuDiv.activeSubMenu = subMenu;
		internal.registerVisible(menuDiv);
	}

	
	/**
		Hide submenus that have been opened from the given
		menu.
		@param menuDiv A menu object or the ID
		@param hideTopMenu If set, the top, level menu is hidden as well
	*/
	menu.hideSubMenus = function(menuDiv, hideTopMenu)
	{
		menuDiv = Doc.element(menuDiv);
		var activeSubMenu = menuDiv.activeSubMenu;
		if (activeSubMenu)
		{
			menu.hideSubMenus(activeSubMenu, false);
			activeSubMenu.style.display = 'none';
			menuDiv.activeSubMenu = null;
			internal.unregisterVisible(menuDiv);
		}
		if (menuDiv.canHide && hideTopMenu) 
		{
			menuDiv.style.display = 'none';
			internal.unregisterVisible(menuDiv);
		}
	}

	/**
		Toggle the visibility of a submenu. If the given
		submenu is the currently active submenu, it is hidden,
		otherwise it is shown.
		@param menuItem The ID or document element of the menu-item that is linked with the submenu
		@param subMenu The submenu object or ID to display
	*/
	menu.toggleSubMenu = function(menuItem, subMenu)
	{
		menuItem = Doc.element(menuItem);
		subMenu = Doc.element(subMenu);
		var menuDiv = menu.getMenuOfMenuItem(menuItem);
		if (menuDiv.activeSubMenu == subMenu)
		{
			menu.hideSubMenus(menuDiv);
		}
		else
		{
			menu.showSubMenu(menuItem, subMenu);
		}
	}
	
	/**
		Open the given URL. 
	*/
	menu.openUrl = function(url, target)
	{
		var targetWindow = (target == null ? window : window.top.frames[target]);
		if (targetWindow != null)
		{
			targetWindow.location = url;
		}
		else
		{
			window.open(url, target);
		}
	}

	/**
		Add a click event handler to the menuitem unless it is disabled.
		The menuItem is either the ID or a document element
	*/
	menu.addClickHandler = function(menuItem, handler)
	{
		menuItem = Doc.element(menuItem);
		if (!menuItem) return;
		Events.addEventHandler(menuItem, 'click', handler);
	}
	
	/**
		Add the given menu to the list of visible menus
	*/
	internal.registerVisible = function(menuDiv)
	{
		visibleMenus[visibleMenus.length] = menuDiv;
	}
	
	/**
		Remove the given menu from the list of visible menus.
	*/
	internal.unregisterVisible = function(menuDiv)
	{
		for (var i = visibleMenus.length-1; i >= 0; i--)
		{
			if (visibleMenus[i] == menuDiv) 
			{
				visibleMenus.splice(i, 1);
				return;
			}
		}
	}

	/**
		Initialize menu items by adding event listeners to the menu items, etc.
		This method handle the following auto-initializable elements:
		* submenu: A menu-item that opens up a new submenu
		* menuitem: A menu-item in a menu that has either a target url or an onclick (deprecated) handler
		* menu-anchor: An element that open a new top-level menu
	*/
	internal.initMenuItems = function(element, autoInit)
	{
		if (autoInit == 'submenu')
		{
			Events.addEventHandler(element, 'click', internal.submenuOnClick);
			Events.addEventHandler(element, 'mouseover', internal.submenuOnMouseOver);
		}
		else if (autoInit == 'menuitem')
		{
			var url = Data.get(element, 'url');
			if (url)
			{
				element.addEventListener('click', Events.stopIfDisabled, true);
				Events.addEventHandler(element, 'click', internal.menuItemWithUrlOnClick);
			}
		}
		else if (autoInit == 'menu')
		{
			Events.addEventHandler(element, 'mouseover', internal.menuOnMouseOver);
		}
		else if (autoInit == 'menu-anchor')
		{
			Events.addEventHandler(element, 'click', internal.menuAnchorOnClick);
		}
	}
	Doc.addElementInitializer(internal.initMenuItems);
	
	/**
		Click handler for menu items that has an 'data-url' attribute specified. 
		By default, the url is opened in the main window. If a 'data-target'
		attribute is specified, the url is opened in the window named by the target.
		If a 'data-popup' is specified, the url is opened in a popup window.
	*/
	internal.menuItemWithUrlOnClick = function(event)
	{
		var menuItem = event.currentTarget;
		var url = Data.get(menuItem, 'url');
		var popup = Data.get(menuItem, 'popup');
		url = url.replace(/\$SESSION\-ID\$/, App.getSessionId());
		if (popup)
		{
			var options = popup.split(/[,\s]+/);
			Dialogs.openPopup(url, options[0], options[1], options[2]);
		}
		else
		{
			Menu.openUrl(url, Data.get(menuItem, 'target'));
		}
	}
	
	/**
		Event handler for mouse over events which hide all
		submenus for the current menu.
	*/
	internal.menuItemOnMouseOver = function(event)
	{
		var menuItem = event.currentTarget;
		var menuDiv = menu.getMenuOfMenuItem(menuItem);
		menu.hideSubMenus(menuDiv);
	}
	
	/**
		Event handler for mouse over events which hide all
		submenus for the current menu.
	*/
	internal.menuOnMouseOver = function(event)
	{
		if (!event.target.classList.contains('menuitem')) return;
		menu.hideSubMenus(event.currentTarget);
	}
	
	/**
		Event handler for a submenu-item that opens a submenu
		automatically when the mouse is over the submenu-item.
	*/
	internal.submenuOnMouseOver = function(event)
	{
		var menuItem = event.currentTarget;
		var submenuId = Data.get(menuItem, 'submenu-id');
		var menuDiv = menu.getMenuOfMenuItem(menuItem);
		var open = Data.get(menuDiv, 'open');
		if (open == 'enter')
		{
			menu.showSubMenu(menuItem, submenuId);
		}
		else
		{
			var subMenu = Doc.element(submenuId);
			if (menuDiv.activeSubMenu && (menuDiv.activeSubMenu != subMenu))
			{
				menu.showSubMenu(menuItem, subMenu);
			}
		}
		event.stopPropagation();
	}
	
	/**
		Click handler for submenu-items that toggle the visibility of
		the submenu.
	*/
	internal.submenuOnClick = function(event)
	{
		var menuItem = event.currentTarget;
		var submenuId = Data.get(menuItem, 'submenu-id');
		menu.toggleSubMenu(menuItem, submenuId);
		event.stopPropagation();
	}
	
	/**
		Event handler for an gui element that, when clicked, open a new
		top-level menu. The menu id should be specified by the 'data-menu-id'
		attribute. If 'data-align-right' is set, the right border of the menu
		is aligned with the right side of the target element, otherwise the
		left sides are aligned (the default).
	 */
	internal.menuAnchorOnClick = function(event)
	{
		var target = event.currentTarget;
		var menuId = Data.get(target, 'menu-id');
		var alignRight = Data.int(target, 'align-right');
		
		var pos = Doc.getElementPosition(target);
		var x = pos.left;
		var y = pos.top+pos.height;
		if (alignRight)
		{
			var winPos = App.getWindowPosition();
			x = winPos.width - pos.left - pos.width;
		}
		menu.toggleTopMenu(menuId, x, y, alignRight);
		event.stopPropagation();
	}
	
	return menu;
}();

Events.addEventHandler(document, 'click', Menu.hideAll);